/*
 * IRC - Internet Relay Chat, badwords.c
 * Copyleft (C) 2000 meow <csi@gnu.net>
 *
 * Modified for RageIRCd by Alasdair McWilliam. <th3dud3@turbo-irc.net>
 * 
 * Provides functions, for loading and filtering unwanted words out of
 * a string, or in this case part of a message.  Please note that this
 * is flawed because when mode_strip is set, mode_stripbadwords is not
 * active.  This is due to the structure of m_message(), and therefore
 * will not change until I (or someone else) revamps the code.
 *
 * You can redistribute and/or modify this under the terms of the GNU
 * General Public License as published by the Free Software Foundation.
 *
 * Disclaimer: You have no rights.  Use at your own risk.  Don't drink
 * too much pepsi.
 */

#include <unistd.h>
#include "config.h"
#include "struct.h"
#include "common.h"
#include "h.h"
#include "badwords.h"
#include "numeric.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "setup.h"
#ifdef HAVE_REGEX
#include <regex.h>
#else
#include "../include/regex/regex.h"
#endif

aBadword	*badwords = NULL;

/*
 * Returns a string which has been filtered by the words loaded via
 * the load_badwords() function. It's primary use is to filter swear
 * words from both private and public messages.
 */
char	*strip_badwords(str)
	char		*str;
{
	regmatch_t	pmatch[1];
	regex_t		pcomp;
	static	char	cleanstr[4096];
	char		buf[4096], *ptr;
	int		errorcode, matchlen, stringlen;
	aBadword	*ab;

	if (!badwords)
	{
		return str;
	}

	strncpy(cleanstr, str, sizeof(cleanstr) - 1);
	memset(&pmatch, 0, sizeof(pmatch));

	stringlen = strlen(cleanstr);
	matchlen = 0;
	buf[0] = '\0';

	for (ab = badwords; ab; ab = ab->next)
	{
		if ((errorcode = regcomp(&pcomp, ab->word, REG_ICASE)) > 0)
		{
			regfree(&pcomp);
			return cleanstr;
		}

		ptr = cleanstr;
		while (regexec(&pcomp, ptr, 1, pmatch, 0) != REG_NOMATCH)
		{
			if (pmatch[0].rm_so == -1)
			{
				break;
			}
			matchlen += pmatch[0].rm_eo - pmatch[0].rm_so;
			strncat(buf, ptr, pmatch[0].rm_so);
			strcat(buf, ab->replace);
			ptr += pmatch[0].rm_eo;
			memset(&pmatch, 0, sizeof(pmatch));
		}

		strcat(buf, ptr);
		strncpy(cleanstr, buf, sizeof(cleanstr));
		memset(buf, 0, sizeof(buf));
		regfree(&pcomp);

		if (matchlen == stringlen)
		{
			break;
		}
	}
	return (cleanstr);
}

/*
 * Loads bad words from a file, into a structure and then links them into
 * a linked list. The words are then called by strip_badwords, to filter
 * any unwanted (swear) words.
 *
 * Rewritten for RageIRCd by Alasdair McWilliam. <th3dud3@turbo-irc.net>
 */
int	load_badwords(mode)
	int		mode;
{
	FILE		*fd;
	char		*word, *replace, *x, buf[1024];
	int		badrep;

	fd = fopen(BADWORDS_FILE, "r");
	if (!fd)
	{
		return -1;
	}

	while (fgets(buf, 1024, fd))
	{
		if ((buf[0] == '#') || (buf[0] == '/') || (buf[0] == '\0'))
		{
			continue;
		}
		strip(buf);
		if ((buf[0] == '#') || (buf[0] == '/') || (buf[0] == '\0'))
		{
			continue;
		}

		x = (char *)strtok((char *)buf, " ");
		if (strcmp("badword", x) == 0)
		{
			word = (char *)strtok((char *)NULL, " ");
			if (!word)
			{
				continue;
			}
			if (check_word(word) == -1)
			{
				if (mode == 0)
				{
					fprintf(stderr, "* WARNING! Invalid badword '%s'\n", word);
				}
				else
				{
					sendto_ops("Warning: Invalid badword '%s'\n", word);
				}
				continue;
			}

			badrep = 0;
			replace = (char *)strtok((char *)NULL, " ");
			if (!replace)
			{
				if (mode == 0)
				{
					fprintf(stderr, "* WARNING! No replacement for badword '%s' (using default)\n", word);
				}
				else
				{
					sendto_ops("Warning: No replacement for badword '%s' (using default)", word);
				}
				badrep = 1;
			}
			if (check_word(replace) == -1 && badrep == 0)
			{
				if (mode == 0)
				{
					fprintf(stderr, "* WARNING! Invalid replacement '%s' for badword '%s' (using default)\n", replace, word);
				}
				else
				{
					sendto_ops("Warning: Invalid replacement '%s' for badword '%s' (using default)", replace, word);
				}
				badrep = 1;
			}

			if (badrep == 1)
			{
				add_badword(word, REPLACEWORD);
			}
			else
			{
				add_badword(word, replace);
			}
		}
	}
	fclose(fd);
	return 0;
}

int	check_word(word)
	char		*word;
{
	char		*p;
	int		stars = 0, badchars = 0, i = 0;

	p = word;
	if (p)
	{
		while (*p)
		{
			if (*p == '*')
			{
				stars++;
			}
			if (!isallowed(*p) && *p != '*')
			{
				badchars = 1;
			}
			p++;
			i++;
		}
	}
	if (stars >= i)
	{
		return -1;
	}
	if (badchars == 1)
	{
		return -1;
	}
}


int	add_badword(word, replace)
	char		*word, *replace;
{
	aBadword	*ab;

	ab = (aBadword *)MyMalloc(sizeof(aBadword));
	ab->word = strdup(word);
	ab->replace = strdup(replace);

	ab->next = badwords;
	ab->prev = NULL;
	if (badwords)
	{
		badwords->prev = ab;
	}
	badwords = ab;
}

aBadword	*del_badword(ab)
	aBadword		*ab;
{
	aBadword		*x, *z;

	for (x = badwords; x; x = x->next)
	{
		if (x == ab)
		{
			z = x->next;
			MyFree((char *)(ab->word));
			MyFree((char *)(ab->replace));

			if (x->prev)
			{
				x->prev->next = x->next;
			}
			else
			{
				badwords = x->next;
			}

			if (x->next)
			{
				x->next->prev = x->prev;
			}

			MyFree((aBadword *) ab);
			return z;
		}
	}
	return NULL;
}

void	rehash_badwords(mode)
	int		mode;
{
	aBadword	*ab, x;

	for (ab = badwords; ab; ab = ab->next)
	{
		x.next = del_badword(ab);
		ab = &x;
	}
	load_badwords(mode);
}

void	list_badwords(sptr)
	aClient		*sptr;
{
	aBadword	*ab;

	for (ab = badwords; ab; ab = ab->next)
	{
		sendto_one(sptr, ":%s %i %s :BW %s (%s)", me.name, RPL_TEXT, sptr->name, ab->word, ab->replace);
	}
}
