/***************************************************************************
 * IRC - Internet Relay Chat, src/channel.c
 * Copyright (C) 1990 Jarkko Oikarinen and University of Oulu, Co Center
 *
 * Contains code from DarkFyre IRCd - Copyright (C) binary
 * Contains code from Unreal IRCd - Copyright (C) The UnrealIRCd Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <string.h>
#include "ircstats.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "channel.h"
#include "msg.h"		/* For TOK_*** and MSG_*** strings  */
#include "hash.h"		/* For CHANNELHASHSIZE */
#include "h.h"
#include "dynconf2.h"
#include "sjoin.h"

extern	ircstats	IRCstats;
static	int		opermode = 0;

aChannel	*channel = NullChn;

/*
 * New mode system - based on DarkFyre/Twilight IRCd by binary.
 */
static	void	set_mode PROTO((aChannel *, aClient *, int, char **, u_int *, char [MAXMODEPARAMS][MODEBUFLEN+3], int));
static  void	make_mode_str PROTO((aChannel *, u_int, u_int, char [LINKLEN+1], int, char [MAXMODEPARAMS][MODEBUFLEN+3], char *, char *, char));
static	int	do_mode_char PROTO((aChannel *, u_int, char, char *, u_int, aClient *, u_int *, char [MAXMODEPARAMS][MODEBUFLEN+3], char));
static	void	do_mode PROTO((aChannel *, aClient *, aClient *, int, char **, int));
static	void	bounce_mode PROTO((aChannel *, aClient *, int, char **));

static void add_invite PROTO((aClient *, aChannel *));
static int can_join PROTO((aClient *, aClient *, aChannel *, char *, char **));
static int channel_link PROTO((aClient *, aClient *, int, char **));
void channel_modes PROTO((aClient *, char *, char *, aChannel *));
static int check_channelmask PROTO((aClient *, aClient *, char *));

/* Bans */
static int add_banid PROTO((aClient *, aChannel *, char *));
static int del_banid PROTO((aChannel *, char *));
static int find_banid PROTO((aChannel *, char *));

/* Exception Bans */
static int add_exbanid PROTO((aClient *, aChannel *, char *));
static int del_exbanid PROTO((aChannel *, char *));
static int find_exbanid PROTO((aChannel *, char *));

/* Forwards */
static int add_forwardid PROTO((aClient *, aChannel *, char *, char *));
static int del_forwardid PROTO((aChannel *, char *, char *));
static int find_forwardid PROTO((aChannel *, char *, char *));

static int have_ops PROTO((aChannel *));
static int number_of_zombies PROTO((aChannel *));
static void sub1_from_channel PROTO((aChannel *));

static	aChannel	*get_channel PROTO((aClient *, char *, int));

void clean_channelname PROTO((char *));
void del_invite PROTO((aClient *, aChannel *));
int check_for_target_limit(aClient *, void *, const char *);	/* from s_user.c */
int sendmodeto_one(aClient *, char *, char *, char *, char *, time_t);

static char *PartFmt = ":%s PART %s";
static char *PartFmt2 = ":%s PART %s :%s";
/*
 * some buffers for rebuilding channel/nick lists with ,'s
 */
static char nickbuf[BUFSIZE], buf[BUFSIZE];
static char modebuf[MODEBUFLEN], parabuf[MODEBUFLEN];

/*
 * channel link stuff
 */
#define MAXBOUNCE 5
static	int	bouncedtimes = 0;
typedef	struct	ChLink	aChLink;
struct ChLink {
	aChannel	*chan;
};


/* ***************************************************** modes ********/

typedef struct {
	u_int		mode;
	char		flag;
	int		halfop_mode;
	int		param;
} aCtab;

aCtab cFlagTab[] = {
	{MODE_CHANADMIN,		'a',	0,	1},
	{MODE_ADMINSONLY,		'A',	0,	0},
	{MODE_BAN,			'b',	0,	1},
	{MODE_OPERBAN,			'B',	0,	0},
	{MODE_NOCOLORS,			'c',	1,	0},
	{MODE_NOCTCP,			'C',	1,	0},
	{MODE_NONICKCH,			'd',	0,	0},
	{MODE_EXBAN,			'e',	0,	1},
	{MODE_FORWARD,			'F',	0,	1},
	{MODE_FLOODLIMIT,		'f',	0,	1},
	{MODE_HALFOP,			'h',	1,	1},
	{MODE_INVITEONLY,		'i',	0,	0},
	{MODE_PEACE,			'j',	0,	0},
	{MODE_KEY,			'k',	0,	1},
	{MODE_NOKNOCKS,			'K',	0,	0},
	{MODE_LIMIT,			'l',	0,	1},
	{MODE_LINK,			'L',	0,	0},
	{MODE_MODERATED,		'm',	0,	0},
	{MODE_NOPRIVMSGS,		'n',	1,	0},
	{MODE_NETADMINSONLY,		'N',	0,	0},
	{MODE_CHANOP,			'o',	0,	1},
	{MODE_IRCOPSONLY,		'O',	0,	0},
	{MODE_PRIVATE,			'p',	0,	0},
	{MODE_RGSTR,			'r',	0,	0},
	{MODE_RGSTRONLY,		'R',	0,	0},
	{MODE_SECRET,			's',	0,	0},
	{MODE_STRIPCOLORS,		'S',	1,	0},
	{MODE_TOPICLIMIT,		't',	1,	0},
	{MODE_NOINVITES,		'T',	0,	0},
	{MODE_VOICE,			'v',	1,	1},
	{MODE_STRIPBADWORDS,		'V',	1,	0},
	{MODE_STRIPFORMAT,		'x',	1,	0},
	{0x0,				0x0,	0x0,	0x0}
};

char	cmode_string[512];
void	make_cmode_string(void)
{
	char	*p = &cmode_string[0];
	aCtab	*tab = &cFlagTab[0];

	while (tab->mode != 0x0)
	{
		*p = tab->flag;
		p++;
		tab++;
	}
	p = '\0';
}

int	check_halfop_mode(u_int modetype)
{
	aCtab		*tab = &cFlagTab[0];

	while (tab->mode != 0x0)
	{
		if (tab->mode == modetype)
		{
			return ((tab->halfop_mode == 1) ? TRUE : FALSE);
		}
		tab++;
	}
}

/*
 * return the length (>=0) of a chain of links.
 */
static int list_length(lp)
    Reg1 Link *lp;
{
    Reg2 int count = 0;

    for (; lp; lp = lp->next)
	count++;
    return count;
}

/*
   ** find_chasing
   **   Find the client structure for a nick name (user) using history
   **   mechanism if necessary. If the client is not found, an error
   **   message (NO SUCH NICK) is generated. If the client was found
   **   through the history, chasing will be 1 and otherwise 0.
 */
static aClient *find_chasing(sptr, user, chasing)
    aClient *sptr;
    char *user;
    Reg1 int *chasing;
{
    Reg2 aClient *who = find_client(user, (aClient *) NULL);

    if (chasing)
	*chasing = 0;
    if (who)
	return who;
    if (!(who = get_history(user, (long) KILLCHASETIMELIMIT))) {
	sendto_one(sptr, err_str(ERR_NOSUCHNICK),
		   me.name, sptr->name, user);
	return NULL;
    }
    if (chasing)
	*chasing = 1;
    return who;
}

/*
 * Ban functions to work with mode +b
 */
/* add_banid - add an id to be banned to the channel  (belongs to cptr) */

static int add_banid(cptr, chptr, banid)
    aClient *cptr;
    aChannel *chptr;
    char *banid;
{
    Reg1 Ban *ban;
    Reg2 int cnt = 0, len = 0;

    if (MyClient(cptr))
	(void) collapse(banid);
    for (ban = chptr->banlist; ban; ban = ban->next) {
	len += strlen(ban->banstr);
	if (MyClient(cptr)) {
	    if ((len > MAXBANLENGTH) || (++cnt >= MAXBANS)) {
		sendto_one(cptr, err_str(ERR_BANLISTFULL),
			   me.name, cptr->name,
			   chptr->chname, banid);
		return -1;
	    } else {
		if (!match(ban->banstr, banid) ||
		    !match(banid, ban->banstr))
		    return -1;
	    }
	} else if (!mycmp(ban->banstr, banid))
	    return -1;

    }
    ban = make_ban();
    bzero((char *) ban, sizeof(Ban));
/*  ban->flags = CHFL_BAN;          They're all bans!! */
    ban->next = chptr->banlist;
    ban->banstr = (char *) MyMalloc(strlen(banid) + 1);
    (void) strcpy(ban->banstr, banid);
    ban->who = (char *) MyMalloc(strlen(cptr->name) + 1);
    (void) strcpy(ban->who, cptr->name);
    ban->when = TStime();
    chptr->banlist = ban;
    return 0;
}
/*
 * del_banid - delete an id belonging to cptr
 */
static int del_banid(chptr, banid)
    aChannel *chptr;
    char *banid;
{
    Reg1 Ban **ban;
    Reg2 Ban *tmp;

    if (!banid)
	return -1;
    for (ban = &(chptr->banlist); *ban; ban = &((*ban)->next))
	if (mycmp(banid, (*ban)->banstr) == 0) {
	    tmp = *ban;
	    *ban = tmp->next;
	    MyFree(tmp->banstr);
	    MyFree(tmp->who);
	    free_ban(tmp);
	    return 0;
	}
    return -1;
}

/*
 * find_banid - Find an exact match for a ban
 */
static int find_banid(chptr, banid)
    aChannel *chptr;
    char *banid;
{
    Reg1 Ban **ban;
/*  Reg2 Ban *tmp; Compiler says this is unused */

    if (!banid)
	return -1;
    for (ban = &(chptr->banlist); *ban; ban = &((*ban)->next))
	if (!mycmp(banid, (*ban)->banstr))
	    return 1;
    return 0;
}

/*
 *  Exception Ban Functions (+e)
 *  -Ax0n
 */
static int add_exbanid(cptr, chptr, banid)
    aClient *cptr;
    aChannel *chptr;
    char *banid;
{
    Reg1 Ban *ban;
    Reg2 int cnt = 0, len = 0;

    if (MyClient(cptr))
	(void) collapse(banid);
    for (ban = chptr->exbanlist; ban; ban = ban->next) {
	len += strlen(ban->banstr);
	if (MyClient(cptr)) {
	    if ((len > MAXBANLENGTH) || (++cnt >= MAXBANS)) {
		sendto_one(cptr, err_str(ERR_BANLISTFULL),
			   me.name, cptr->name,
			   chptr->chname, banid);
		return -1;
	    } else {
		if (!match(ban->banstr, banid) ||
		    !match(banid, ban->banstr))
		    return -1;
	    }
	} else if (!mycmp(ban->banstr, banid))
	    return -1;

    }
    ban = make_ban();
    bzero((char *) ban, sizeof(Ban));
/*  ban->flags = CHFL_BAN;          They're all bans!! */
    ban->next = chptr->exbanlist;
    ban->banstr = (char *) MyMalloc(strlen(banid) + 1);
    (void) strcpy(ban->banstr, banid);
    ban->who = (char *) MyMalloc(strlen(cptr->name) + 1);
    (void) strcpy(ban->who, cptr->name);
    ban->when = TStime();
    chptr->exbanlist = ban;
    return 0;
}

static int del_exbanid(chptr, banid)
    aChannel *chptr;
    char *banid;
{
    Reg1 Ban **ban;
    Reg2 Ban *tmp;

    if (!banid)
	return -1;
    for (ban = &(chptr->exbanlist); *ban; ban = &((*ban)->next))
	if (mycmp(banid, (*ban)->banstr) == 0) {
	    tmp = *ban;
	    *ban = tmp->next;
	    MyFree(tmp->banstr);
	    MyFree(tmp->who);
	    free_ban(tmp);
	    return 0;
	}
    return -1;
}

static int find_exbanid(chptr, banid)
    aChannel *chptr;
    char *banid;
{
    Reg1 Ban **ban;
/*  Reg2 Ban *tmp; Compiler says this is unused */

    if (!banid)
	return -1;
    for (ban = &(chptr->exbanlist); *ban; ban = &((*ban)->next))
	if (!mycmp(banid, (*ban)->banstr))
	    return 1;
    return 0;
}

/*
 * add_forwardid - adds a +F to forwardlist.
 */
static	int	add_forwardid(cptr, chptr, hostname, channel)
	aClient		*cptr;
	aChannel	*chptr;
	char		*hostname, *channel;
{
	aForward	*af;
	int		cnt = 0, len = 0;

	if (MyClient(cptr))
		(void)collapse(hostname);
		(void)collapse(channel);

	for (af = chptr->forwardlist; af; af = af->next)
	{
		len += (strlen(af->hostname));
		len += (len + strlen(af->channel));

		if (MyClient(cptr))
		{
			if ((len > MAXFORWARDLEN) || (++cnt >= MAXBANS))
			{
				sendto_one(cptr, err_str(ERR_FORWARDLISTFULL), me.name, cptr->name, chptr->chname, hostname, channel);
				return -1;
			}
			else
			{
				if ((!match(af->hostname, hostname) || !match(hostname, af->hostname))  && (!match(af->channel, channel)
					|| !match(channel, af->channel)))
				{
					return -1;
				}
			}
		}
		else if (!mycmp(af->hostname, hostname) && !mycmp(af->channel, channel))
		{
			return -1;
		}
	}

	af = make_forward();
	bzero((char *) af, sizeof(aForward));
	af->next = chptr->forwardlist;
	af->hostname = (char *) MyMalloc(strlen(hostname) + 1);
	(void)strcpy(af->hostname, hostname);
	af->channel = (char *) MyMalloc(strlen(channel) + 1);
	(void)strcpy(af->channel, channel);
	af->who = (char *) MyMalloc(strlen(cptr->name) + 1);
	(void)strcpy(af->who, cptr->name);
	af->when = TStime();
	chptr->forwardlist = af;

	return 0;
}

/*
 * del_forwardid - removes a +F from the forwardlist
 */
static	int	del_forwardid(chptr, hostname, channel)
	aChannel	*chptr;
	char		*hostname, *channel;
{
	aForward	**af;
	aForward	*tmp;

	if (!(hostname || channel))
	{
		return -1;
	}

	for (af = &(chptr->forwardlist); *af; af =  &((*af)->next))
	{
		if ((mycmp(hostname, (*af)->hostname) == 0) && (mycmp(channel, (*af)->channel) == 0))
		{
			tmp = *af;
			*af = tmp->next;
			MyFree(tmp->hostname);
			MyFree(tmp->channel);
			MyFree(tmp->who);
			free_forward(tmp);
			return 0;
		}
	}
	return -1;
}

/*
 * find_forwardid - finds a +F in the forwardlist
 */
static	int	find_forwardid(chptr, hostname, channel)
	aChannel	*chptr;
	char		*hostname, *channel;
{
	aForward	**af;

	if (!(hostname || channel))
	{
		return -1;
	}
	for (af = &(chptr->forwardlist); *af; af = &((*af)->next))
	{
		if (!mycmp(hostname, (*af)->hostname) && !mycmp(channel, (*af)->channel))
		{
			return 1;
		}
	}
	return 0;
}

/*
 * IsMember - returns 1 if a person is joined and not a zombie
 */
int IsMember(cptr, chptr)
    aClient *cptr;
    aChannel *chptr;
{
    Link *lp;
    /* Zombies have been removed */
    return (((lp = find_user_link(chptr->members, cptr))	/* &&
								   !(lp->flags & CHFL_ZOMBIE) */ )? 1 : 0);
}

/*
 * is_banned - returns a pointer to the ban structure if banned else NULL
 */
extern Ban *is_banned(cptr, chptr)
    aClient *cptr;
    aChannel *chptr;
{
    Ban *tmp;
    char *s;
    char host[NICKLEN+USERLEN+HOSTLEN+6], maskedhost[NICKLEN+USERLEN+HOSTLEN+6];
    int testem = 0;

    if (!IsPerson(cptr)) {
	return NULL;
    }
    if (strcmp(cptr->user->host, cptr->user->maskedhost)) {
	testem = 1;
    }

    s = make_nick_user_host(cptr->name, cptr->user->username, cptr->user->host);
    strcpy(host, s);

    s = make_nick_user_host(cptr->name, cptr->user->username, cptr->user->maskedhost);
    strcpy(maskedhost, s);

    for (tmp = chptr->banlist; tmp; tmp = tmp->next) {
	if ((match(tmp->banstr, host) == 0) || (testem && (match(tmp->banstr, maskedhost) == 0))) {
	    break;
	}
    }
    return (tmp);

}

/*
 * is_banexception - see is_banned(), same diff.
 */
extern Ban *is_banexception(cptr, chptr)
    aClient *cptr;
    aChannel *chptr;
{
    Ban *tmp;
    char *s;
    char host[NICKLEN+USERLEN+HOSTLEN+6], maskedhost[NICKLEN+USERLEN+HOSTLEN+6];
    int testem = 0;

    if (!IsPerson(cptr)) {
	return NULL;
    }
    if (strcmp(cptr->user->host, cptr->user->maskedhost)) {
	testem = 1;
    }

    s = make_nick_user_host(cptr->name, cptr->user->username, cptr->user->host);
    strcpy(host, s);

    s = make_nick_user_host(cptr->name, cptr->user->username, cptr->user->maskedhost);
    strcpy(maskedhost, s);

    for (tmp = chptr->exbanlist; tmp; tmp = tmp->next) {
	if ((match(tmp->banstr, host) == 0) || (testem && (match(tmp->banstr, maskedhost) == 0))) {
	    break;
	}
    }
    return (tmp);
}

/*
 * is_forwarded - identifies if user matches a +F
 */
extern	aForward	*is_forwarded(cptr, chptr)
	aClient		*cptr;
	aChannel	*chptr;
{
	aForward	*tmp;
	char		*s;
	char		host[NICKLEN+USERLEN+HOSTLEN+6], maskedhost[NICKLEN+USERLEN+HOSTLEN+6];
	int		testem = 0;

	if (!IsPerson(cptr))
		return NULL;
	if (strcmp(cptr->user->host, cptr->user->maskedhost))
		testem = 1;

	s = make_nick_user_host(cptr->name, cptr->user->username, cptr->user->host);
	strcpy(host, s);
	s = make_nick_user_host(cptr->name, cptr->user->username, cptr->user->maskedhost);
	strcpy(maskedhost, s);

	for (tmp = chptr->forwardlist; tmp; tmp = tmp->next)
	{
		if ((match(tmp->hostname, host) == 0) || (testem && (match(tmp->hostname, maskedhost) == 0)))
		{
			break;
		}
	}
	return (tmp);
}

/*
 * adds a user to a channel by adding another link to the channels member
 * chain.
 */
static void add_user_to_channel(chptr, who, flags)
    aChannel *chptr;
    aClient *who;
    int flags;
{
    Reg1 Link *ptr;

    if (who->user) {
	ptr = make_link();
	ptr->value.cptr = who;
	ptr->flags = flags;

	/* flood */
	ptr->flood = (aFloodOpt *) MyMalloc(sizeof(aFloodOpt));
	ptr->flood->nmsg = 0;
	ptr->flood->lastmsg = 0;

	ptr->next = chptr->members;
	chptr->members = ptr;
	chptr->users++;

	ptr = make_link();
	ptr->value.chptr = chptr;
	ptr->next = who->user->channel;
	who->user->channel = ptr;
	who->user->joined++;
    }
}

void remove_user_from_channel(sptr, chptr)
    aClient *sptr;
    aChannel *chptr;
{
    Reg1 Link **curr;
    Reg2 Link *tmp;
    Reg3 Link *lp = chptr->members;

    for (; lp && (lp->flags & CHFL_ZOMBIE || lp->value.cptr == sptr);
	 lp = lp->next);
    for (;;) {
	for (curr = &chptr->members; (tmp = *curr); curr = &tmp->next)
	    if (tmp->value.cptr == sptr)
	    {
		MyFree((aFloodOpt *) tmp->flood);
		*curr = tmp->next;
		free_link(tmp);
		break;
	    }
	for (curr = &sptr->user->channel; (tmp = *curr); curr = &tmp->next)
	    if (tmp->value.chptr == chptr) {
		*curr = tmp->next;
		free_link(tmp);
		break;
	    }
	sptr->user->joined--;
	if (lp)
	    break;
	if (chptr->members)
	    sptr = chptr->members->value.cptr;
	else
	    break;
	sub1_from_channel(chptr);
    }
    sub1_from_channel(chptr);
}

static int have_ops(chptr)
    aChannel *chptr;
{
    Reg1 Link *lp;

    if (chptr) {
	lp = chptr->members;
	while (lp) {
	    if (lp->flags & CHFL_CHANOP)
		return (1);
	    lp = lp->next;
	}
    }
    return 0;
}

int is_chan_op(cptr, chptr)
    aClient *cptr;
    aChannel *chptr;
{
    Reg1 Link *lp;

    if (chptr)
	if ((lp = find_user_link(chptr->members, cptr)))
	    return (lp->flags & CHFL_CHANOP);

    return 0;
}

int is_half_op(cptr, chptr)
	aClient		*cptr;
	aChannel	*chptr;
{
	Reg1	Link	*lp;

	if (chptr)
	    if ((lp = find_user_link(chptr->members, cptr)))
		return (lp->flags & CHFL_HALFOP);

	return 0;
}

int is_chanadmin(cptr, chptr)
	aClient		*cptr;
	aChannel	*chptr;
{
	Reg1	Link	*lp;

	if (chptr)
	    if ((lp = find_user_link(chptr->members, cptr)))
		return (lp->flags & CHFL_CHANADMIN);

	return 0;
}

int has_voice(cptr, chptr)
    aClient *cptr;
    aChannel *chptr;
{
    Reg1 Link *lp;

    if (chptr)
	if ((lp = find_user_link(chptr->members, cptr)) &&
	    !(lp->flags & CHFL_ZOMBIE))
	    return (lp->flags & CHFL_VOICE);

    return 0;
}

#define CANNOT_SEND_MODERATED 1
#define CANNOT_SEND_NOPRIVMSGS 2
#define CANNOT_SEND_NOCOLORS 3
#define CANNOT_SEND_BAN 4
#define CANNOT_SEND_NOCTCP 5

int	can_send(cptr, chptr, msgtext)
	aClient		*cptr;
	aChannel	*chptr;
	char		*msgtext;
{
	Link		*lp;

	if (IsULine(cptr, cptr) || IsServer(cptr))
	{
		return 0;
	}

	lp = find_user_link(chptr->members, cptr);

	if ((chptr->mode.mode & MODE_MODERATED) && !IsOper(cptr))
	{
		if (!lp || !(lp->flags & (CHFL_CHANADMIN | CHFL_CHANOP | CHFL_HALFOP | CHFL_VOICE)) || (lp->flags & CHFL_ZOMBIE))
		{
			return (CANNOT_SEND_MODERATED);
		}
	}

	if (chptr->mode.mode & MODE_NOPRIVMSGS)
	{
		if (!IsMember(cptr, chptr))
		{
			return (CANNOT_SEND_NOPRIVMSGS);
		}
	}

	if (chptr->mode.mode & MODE_NOCOLORS)
	{
		if (!lp || (strchr((char *)msgtext, 3) || strchr((char *)msgtext, 27)))
		{
			return (CANNOT_SEND_NOCOLORS);
		}
	}

	if (is_banned(cptr, chptr) && MyClient(cptr))
	{
		if (IsOper(cptr) || is_banexception(cptr, chptr))
		{
			return 0;
		}
		if (!lp || !(lp->flags & (CHFL_CHANADMIN | CHFL_CHANOP | CHFL_VOICE | CHFL_HALFOP)) || (lp->flags & CHFL_ZOMBIE))
		{
			return (CANNOT_SEND_BAN);
		}
	}

	if (chptr->mode.mode & MODE_NOCTCP)
	{
		if (!lp || ((msgtext[0] == 1) && (strncmp(&msgtext[1], "ACTION ", 7))))
		{
			return (CANNOT_SEND_NOCTCP);
		}
	}

	return 0;
}

aChannel	*find_channel(chname, chptr)
	char		*chname;
	aChannel	*chptr;
{
	return hash_find_channel(chname, chptr);
}

/*
 * write the "simple" list of channel modes for channel chptr onto buffer mbuf
 * with the parameters in pbuf.
 *
 * fixed to support new SJOIN (RTS-1)
 */
void channel_modes(cptr, mbuf, pbuf, chptr)
	aClient			*cptr;
	char			*mbuf, *pbuf;
	aChannel		*chptr;
{
	u_int			zode;
	aCtab			*tab = &cFlagTab[0];
	char			tmpbuf[1024];

	*mbuf++ = '+';

	while (tab->mode != 0x0)
	{
		if ((chptr->mode.mode & tab->mode))
		{
			zode = chptr->mode.mode;

			if (!(zode & (MODE_LIMIT | MODE_KEY | MODE_LINK  | MODE_FLOODLIMIT)))
				if (!(zode & (MODE_VOICE | MODE_HALFOP | MODE_CHANOP | MODE_CHANADMIN)))
					if (!(zode & (MODE_BAN | MODE_EXBAN | MODE_FORWARD)))
						*mbuf++ = tab->flag;
		}
		tab++;
	}

	if (chptr->mode.limit)
	{
		*mbuf++ = 'l';
		if (IsMember(cptr, chptr) || IsServer(cptr) || IsULine(cptr, cptr))
		{
			(void)ircsprintf(tmpbuf, "%i ", chptr->mode.limit);
			(void)strcat(pbuf, tmpbuf);
		}
	}

	if (*chptr->mode.key)
	{
		*mbuf++ = 'k';
		if (IsMember(cptr, chptr) || IsServer(cptr) || IsULine(cptr, cptr))
		{
			(void)ircsprintf(tmpbuf, "%s ", chptr->mode.key);
			(void)strcat(pbuf, tmpbuf);
		}
	}

	if (*chptr->mode.link)
	{
		*mbuf++ = 'L';
		if (IsMember(cptr, chptr) || IsServer(cptr) || IsULine(cptr, cptr))
		{
			(void)ircsprintf(tmpbuf, "%s ", chptr->mode.link);
			(void)strcat(pbuf, tmpbuf);
		}
	}

	if (chptr->mode.secs)
	{
		*mbuf++ = 'f';
		if (IsMember(cptr, chptr) || IsServer(cptr) || IsULine(cptr, cptr))
		{
			(void)ircsprintf(tmpbuf, "%s%i:%i ", ((chptr->mode.fban == 1) ? "*" : ""), chptr->mode.msgs, chptr->mode.secs);
			(void)strcat(pbuf, tmpbuf);
		}
	}

	*mbuf++ = '\0';
	return;
}

static int send_mode_list(cptr, chname, creationtime, top, mask, flag)
    aClient *cptr;
    Link *top;
    int mask;
    char flag, *chname;
    time_t creationtime;
{
    Reg1 Link *lp;
    Reg2 char *cp, *name;
    int count = 0, send = 0, sent = 0;

    cp = modebuf + strlen(modebuf);
    if (*parabuf)		/* mode +l or +k xx */
	count = 1;
    for (lp = top; lp; lp = lp->next) {
	/* 
	 * Okay, since ban's are stored in their own linked
	 * list, we won't even bother to check if CHFL_BAN
	 * is set in the flags. This should work as long
	 * as only ban-lists are feed in with CHFL_BAN mask.
	 * However, we still need to typecast... -Donwulff 
	 */
	if (mask == CHFL_BAN) {
/*          if (!(((Ban *)lp)->flags & mask)) continue; */
	    name = ((Ban *) lp)->banstr;
	} else {
	    if (!(lp->flags & mask))
		continue;
	    name = lp->value.cptr->name;
	}
	if (strlen(parabuf) + strlen(name) + 11 < (size_t) MODEBUFLEN) {
	    if (*parabuf)
		(void) strcat(parabuf, " ");
	    (void) strcat(parabuf, name);
	    count++;
	    *cp++ = flag;
	    *cp = '\0';
	} else if (*parabuf)
	    send = 1;
	if (count == RESYNCMODES)
	    send = 1;
	if (send) {
	    /* cptr is always a server! So we send creationtimes */
	    sendmodeto_one(cptr, me.name, chname, modebuf,
			   parabuf, creationtime);
	    sent = 1;
	    send = 0;
	    *parabuf = '\0';
	    cp = modebuf;
	    *cp++ = '+';
	    if (count != RESYNCMODES) {
		(void) strcpy(parabuf, name);
		*cp++ = flag;
	    }
	    count = 0;
	    *cp = '\0';
	}
    }
    return sent;
}

/*
 * send_ban_list - returns a list of bans and exbans
 *
 * recoded to include bans and exbans in one function
 * based on UnrealIRCd's send_ban_list
 *
 * Modified to support Channel Forwards (+F)
 *
 *                - TD (December 2000)
 */
static int send_ban_list(cptr, chname, creationtime, channel)
	aClient 	*cptr;
	char		*chname;
	time_t		creationtime;
	aChannel	*channel;
{
	Ban		*top;
	Ban		*lp;
	aForward	*ftop;
	aForward	*af;
	char		*cp, *name;
	char		tmpbuf[MAXFORWARDLEN+1];
	int		count = 0, send = 0, sent = 0;

	cp = modebuf + strlen(modebuf);
	if (*parabuf)
		count = 1;

	top = channel->banlist;
	for (lp = top; lp; lp = lp->next)
	{
		name = ((Ban *) lp)->banstr;

		if ((strlen(parabuf) + strlen(name) + 11) < (size_t)MODEBUFLEN)
		{
			if (*parabuf)
				(void)strcat(parabuf, " ");
			(void)strcat(parabuf, name);
			count++;
			*cp++ = 'b';
			*cp = '\0';
		}
		else if (*parabuf)
			send = 1;
		if (count == MODEPARAMS)
			send = 1;
		if (send)
		{
			sendto_one(cptr, "%s %s %s %s %lu", (IsToken(cptr) ? TOK_MODE : MSG_MODE),
				chname, modebuf, parabuf, creationtime);
			sent = 1;
			send = 0;
			*parabuf = '\0';
			cp = modebuf;
			*cp++ = '+';

			if (count != MODEPARAMS)
			{
				(void)strcpy(parabuf, name);
				*cp++ = 'b';
			}
			count = 0;
			*cp = '\0';
		}
	}
	top = channel->exbanlist;
	for (lp = top; lp; lp = lp->next)
	{
		name = ((Ban *)lp)->banstr;

		if ((strlen(parabuf) + strlen(name) + 11) < (size_t)MODEBUFLEN)
		{
			if (*parabuf)
				(void)strcat(parabuf, " ");
			(void)strcat(parabuf, name);
			count++;
			*cp++ = 'e';
			*cp = '\0';
		}
		else if (*parabuf)
			send = 1;
		if (count == MODEPARAMS)
			send = 1;
		if (send)
		{
			sendto_one(cptr, "%s %s %s %s %lu", (IsToken(cptr) ? TOK_MODE : MSG_MODE),
				modebuf, parabuf, creationtime);
			sent = 1;
			send = 0;
			*parabuf = '\0';
			cp = modebuf;
			*cp++ = '+';
			if (count != MODEPARAMS)
			{
				(void)strcpy(parabuf, name);
				*cp++ = 'e';
			}
			count = 0;
			*cp = '\0';
		}
	}
	ftop = channel->forwardlist;
	for (af = ftop; af; af = af->next)
	{
		(void)ircsprintf(tmpbuf, "%s:%s", af->hostname, af->channel);
		name = tmpbuf;

		if ((strlen(parabuf) + strlen(name) + 11) < (size_t)MODEBUFLEN)
		{
			if (*parabuf)
				(void)strcat(parabuf, " ");
			(void)strcat(parabuf, name);
			count++;
			*cp++ = 'F';
			*cp = '\0';
		}
		else if (*parabuf)
			send = 1;
		if (count == MODEPARAMS)
			send = 1;

		if (send)
		{
			sendto_one(cptr, "%s %s %s %s %lu", (IsToken(cptr) ? TOK_MODE : MSG_MODE), modebuf, parabuf, creationtime);
			sent = 1;
			send = 0;
			*parabuf = '\0';
			cp = modebuf;
			*cp++ = '+';
			if (count != MODEPARAMS)
			{
				(void)strcpy(parabuf, name);
				*cp++ = 'L';
			}
			count = 0;
			*cp = '\0';
		}
	}
	return sent;
}

/*
 * send_channel_modes_sjoin
 *
 * Recoded September 5th 2000 by Th3Dud3, based on send_channel_modes_sts (StarChat)
 * Modified for the new RTS protocol SJOIN. -TD (Feb. 2001)
 */
void	send_channel_modes_sjoin(cptr, chptr)
	aClient		*cptr;
	aChannel	*chptr;
{
	Link		*lp;
	char		*name, *bufptr;
	int		n = 0;

	*modebuf = '\0';
	*parabuf = '\0';
	channel_modes(cptr, modebuf, parabuf, chptr);

	if (*parabuf == '\0')
	{
		strcat(parabuf, "<none>");
	}

	(void)ircsprintf(buf, ":%s SJOIN %s %ld %s %s :", me.name, chptr->chname, chptr->creationtime, modebuf, parabuf);
	bufptr = (buf + strlen(buf));

	for (lp = chptr->members; lp; lp = lp->next)
	{
		if (lp->flags & MODE_CHANADMIN)
		{
			*bufptr++ = '*';
		}
		if (lp->flags & MODE_CHANOP)
		{
			*bufptr++ = '@';
		}
		if (lp->flags & MODE_HALFOP)
		{
			*bufptr++ = '%';
		}
		if (lp->flags & MODE_VOICE)
		{
			*bufptr++ = '+';
		}

		name = lp->value.cptr->name;
		strcpy(bufptr, name);

		bufptr += strlen(bufptr);
		*bufptr++ = ' ';
		n++;

		if ((bufptr - buf) > (BUFSIZE - 80))
		{
			*bufptr++ = '\0';
			if (bufptr[-1] == ' ')
			{
				bufptr[-1] = '\0';
			}

			sendto_one(cptr, "%s", buf);
			n = 0;
			bufptr = (buf + strlen(buf));
		}
	}

	if (n)
	{
		*bufptr++ = '\0';
		if (bufptr[-1] == ' ')
		{
			bufptr[-1] = '\0';
		}
		sendto_one(cptr, "%s", buf);
	}

	*parabuf = '\0';
	*modebuf = '+';
	modebuf[1] = '\0';
	send_ban_list(cptr, chptr->chname, chptr->creationtime, chptr);
	if (modebuf[1] || *parabuf)
	{
		sendto_one(cptr, ":%s MODE %s %s %s", me.name, chptr->chname, modebuf, parabuf);
	}
	return;
}

typedef	struct	Parv	aParv;
struct  Parv
{
        int             parc;
        char            *parv[256];
};

aParv pparv;
aParv	*mp2parv(mbuf, pbuf)
	char	*mbuf, *pbuf; 
{
	char	*p, *s;
	int	c;

	pparv.parv[0] = mbuf;
	c = 1;

	for (s = (char *)strtoken(&p, pbuf, " "); s; s = (char *)strtoken(&p, NULL, " "))
	{
		pparv.parv[c] = s;
		c++;
	}
	pparv.parv[c] = NULL;
	return (&pparv);
}

/*
 * m_sjoin - synchs channels and channel users using the new RTS protocol, based
 *           upon the StarChat STS1 protocol. Code based on original SJOIN code
 *           from StarChat IRCd v5.10.fx4 and UnrealIRCd v3.1.1.
 *
 *              parv[0] - sender prefix (server)
 *              parv[1] - channel name
 *              parv[2] - channel TS
 *              parv[3] - channel modes
 *  parv[4 to parc - 2] - channel parameters
 *       parv[parc - 1] - channel users + statuses
 *
 */

#define Max(a, b)	(((a) > (b)) ? (a) : (b))

int	m_sjoin(cptr, sptr, parc, parv)
	aClient		*cptr, *sptr;
	int		parc;
	char		*parv[];
{
	char		nick[NICKLEN + 1];
	Link		*lp, *lp2;
	aChannel	*chptr;
	aClient		*acptr;
	aParv		*ap;
	Mode		oldmode;
	aCtab		*tab;
	char		pvar[MAXMODEPARAMS][MODEBUFLEN + 3];
	char		tmpbuf[1024], parabuf2[1024];
	int		nomodes, noparams, wipeours, wipetheirs, pcount, j;
	int		ts, oldts;
	int		copy, merge;
	char		*ns, *p = NULL, *s = NULL;
	int		i, joinflags, modeflags;
	int		x, y, z, keyerror, linkerror, e;

	if (!IsServer(sptr) || (parc < 6) || IsClient(sptr))
		return 0;
	if (!IsChannelName(parv[1]))
		return 0;

	nomodes = 0;
	if (parv[3][1] == '\0')
	{
		nomodes = 1;
	}

	noparams = 0;
	if (!strncmp(parv[4], "<none>", 6))
	{
		noparams = 1;
	}

	chptr = get_channel(cptr, parv[1], CREATE);
	ts = oldts = 0;
	ts = atol(parv[2]);

	wipeours = wipetheirs = copy = merge = 0;

	if (chptr->creationtime > ts)
	{
		wipeours = copy = 1;
		oldts = chptr->creationtime;
		chptr->creationtime = ts;
	}
	else if ((chptr->creationtime < ts) && (chptr->creationtime != 0))
	{
		wipetheirs = 1;
	}
	else if (chptr->creationtime == ts)
	{
		merge = 1;
	}
	if (chptr->creationtime == 0)
	{
		oldts = -1;
		chptr->creationtime = ts;
	}
	else
	{
		oldts = chptr->creationtime;
	}

	if (ts < 750000)
	{
		if (ts != 0)
		{
			sendto_ops("WARNING! Possible desynch: SJOIN for %s has a fishy timestamp (%ld < 750000)", chptr->chname, ts);
		}
	}

	/*
	 * Clear all channel modes and user statuses.
	 */
	modebuf[0] = parabuf[0] = '\0';
	channel_modes(cptr, modebuf, parabuf, chptr);

	if (wipeours)
	{
		modebuf[0] = '-';

		if (modebuf[1] != '\0')
		{
			ap = mp2parv(modebuf, parabuf);
			set_mode(chptr, cptr, ap->parc, ap->parv, &pcount, pvar, 0);

			sendto_serv_butone(cptr, ":%s MODE %s %s %s %lu", me.name, chptr->chname, modebuf, parabuf, chptr->creationtime);
			sendto_channel_butserv(chptr, cptr, ":%s MODE %s %s %s", me.name, chptr->chname, modebuf, parabuf);
		}

		for (lp = chptr->members; lp; lp = lp->next)
		{
			parabuf[0] = '\0';
			j = 1;

			if (lp->flags & MODE_CHANADMIN)
			{
				lp->flags &= ~MODE_CHANADMIN;

				modebuf[j] = 'a';
				strcat(parabuf, lp->value.cptr->name);
				strcat(parabuf, " ");
				j++;
			}

			if (lp->flags & MODE_CHANOP)
			{
				lp->flags &= ~MODE_CHANOP;

				modebuf[j] = 'o';
				strcat(parabuf, lp->value.cptr->name);
				strcat(parabuf, " ");
				j++;
			}

			if (lp->flags & MODE_HALFOP)
			{
				lp->flags &= ~MODE_HALFOP;

				modebuf[j] = 'h';
				strcat(parabuf, lp->value.cptr->name);
				strcat(parabuf, " ");
				j++;
			}
			if (lp->flags & MODE_VOICE)
			{
				lp->flags &= ~MODE_VOICE;

				modebuf[j] = 'v';
				strcat(parabuf, lp->value.cptr->name);
				strcat(parabuf, " ");
				j++;
			}

			if (j > 1)
			{
				modebuf[j] = '\0';

				sendto_serv_butone(chptr, cptr, ":%s MODE %s %s %s %lu", me.name, chptr->chname, modebuf, parabuf,
					chptr->creationtime);
				sendto_channel_butserv(chptr, sptr, ":%s MODE %s %s %s", me.name, chptr->chname, modebuf, parabuf);
			}
		}
	}

	/*
	 * Make users join the correct channels, and set their user statuses
	 */
	*modebuf = '+';
	modebuf[1] = '\0';
	*parabuf = '\0';

	strcpy(tmpbuf, parv[parc - 1]);

	for (p = strtoken(&s, tmpbuf, " "); p; p = strtoken(&s, (char *)NULL, " "))
	{
		i = joinflags = 0;
		ns = p;

		while ((*ns == '*') || (*ns == '@') || (*ns == '%') || (*ns == '+'))
		{
			switch (*(ns++))
			{
				case '*':
					joinflags |= JF_CHANADMIN;
					break;
				case '@':
					joinflags |= JF_CHANOP;
					break;
				case '%':
					joinflags |= JF_HALFOP;
					break;
				case '+':
					joinflags |= JF_VOICE;
					break;
			}
		}

		while ((*ns != '\0') && (*ns != ' '))
		{
			nick[i++] = *(ns++);
		}
		nick[i] = '\0';

		if ((nick[0] == '\0') || (nick[0] == ' '))
			continue;

		if (!(acptr = find_client(nick, NULL)))
		{
			sendto_ops("WARNING! Possible desynch: missing user %s in SJOIN for %s", nick, chptr->chname);
			continue;
		}

		modeflags = 0;
		if (wipetheirs)
		{
			joinflags = 0;
		}

		if (joinflags & JF_CHANADMIN)
		{
			modeflags |= CHFL_CHANADMIN;
		}
		if (joinflags & JF_CHANOP)
		{
			modeflags |= CHFL_CHANOP;
		}
		if (joinflags & JF_HALFOP)
		{
			modeflags |= CHFL_HALFOP;
		}
		if (joinflags & JF_VOICE)
		{
			modeflags |= CHFL_VOICE;
		}

		if (!IsMember(acptr, chptr))
		{
			add_user_to_channel(chptr, acptr, modeflags);
			sendto_serv_butone(cptr, ":%s JOIN :%s", nick, chptr->chname);
			sendto_channel_butserv(chptr, acptr, ":%s JOIN :%s", nick, chptr->chname);
		}

		lp2 = find_user_link(chptr->members, acptr);
		if (!lp2)
		{
			sendto_ops("WARNING! Possible desynch: missing user %s in SJOIN for %s", nick, chptr->chname);
		}

		*modebuf = '+';
		*parabuf = '\0';
		modebuf[1] = '\0';
		j = 1;

		if (modeflags & CHFL_CHANADMIN)
		{
			lp2->flags |= MODE_CHANADMIN;

			modebuf[j] = 'a';
			strcat(parabuf, nick);
			strcat(parabuf, " ");
			j++;
		}
		if (modeflags & CHFL_CHANOP)
		{
			lp2->flags |= MODE_CHANOP;

			modebuf[j] = 'o';
			strcat(parabuf, nick);
			strcat(parabuf, " ");
			j++;
		}
		if (modeflags & CHFL_HALFOP)
		{
			lp2->flags |= MODE_HALFOP;

			modebuf[j] = 'h';
			strcat(parabuf, nick);
			strcat(parabuf, " ");
			j++;
		}
		if (modeflags & CHFL_VOICE)
		{
			lp2->flags |= MODE_VOICE;

			modebuf[j] = 'v';
			strcat(parabuf, nick);
			strcat(parabuf, " ");
			j++;
		}

		if (j > 1)
		{
			modebuf[j] = '\0';

			sendto_serv_butone(cptr, ":%s MODE %s %s %s %lu", me.name, chptr->chname, modebuf, parabuf,
				chptr->creationtime);
			sendto_channel_butserv(chptr, acptr, ":%s MODE %s %s %s", me.name, chptr->chname, modebuf, parabuf);
		}
	}

	/*
	 * Set any channel modes that need setting
	 */
	if (!merge && !wipetheirs && !nomodes)
	{
		modebuf[0] = '\0';
		*parabuf = *parabuf2 = '\0';

		strcat(modebuf, parv[3]);

		if (!noparams)
		{
			for (j = 4; j <= (parc - 2); j++)
			{
				strcat(parabuf, parv[j]);
				strcat(parabuf, " ");
			}
		}

		strcpy(parabuf2, parabuf);

		ap = mp2parv(modebuf, parabuf);
		set_mode(chptr, cptr, ap->parc, ap->parv, &pcount, pvar, 0);

		sendto_serv_butone(cptr, ":%s MODE %s %s %s %lu", me.name, chptr->chname, modebuf, parabuf2, chptr->creationtime);
		sendto_channel_butserv(chptr, cptr, ":%s MODE %s %s %s", me.name, chptr->chname, modebuf, parabuf2);
	}

	if (merge && !nomodes)
	{
		*modebuf = *parabuf = *parabuf2 = *tmpbuf = '\0';
		keyerror = linkerror = 0;
		e = j = 1;

		if (noparams)
		{
			strcpy(modebuf, parv[3]);

			ap = mp2parv(modebuf, parabuf);
			set_mode(chptr, cptr, ap->parc, ap->parv, &pcount, pvar, 0);

			sendto_serv_butone(cptr, ":%s MODE %s %s %s %lu", me.name, chptr->chname, modebuf, parabuf,
				chptr->creationtime);
			sendto_channel_butserv(chptr, cptr, ":%s MODE %s %s %s", me.name, chptr->chname, modebuf, parabuf);
		}
		else
		{
			/* Copy channels original modes */
			bcopy(&chptr->mode, &oldmode, sizeof(Mode));

			/* Merge all modes, but do not send to channel */
			strcpy(modebuf, parv[3]);
			for (j = 4; j <= (parc - 2); j++)
			{
				strcat(parabuf, parv[j]);
				strcat(parabuf, " ");
			}

			ap = mp2parv(modebuf, parabuf);
			set_mode(chptr, cptr, ap->parc, ap->parv, &pcount, pvar, 0);

			/* Sort params */
			j = 1;
			strcpy(modebuf, "-");
			*parabuf = '\0';

			if (oldmode.limit && !chptr->mode.limit)
			{
				modebuf[j] = 'l';
				strcat(parabuf, (char *)my_itoa(oldmode.limit));
				strcat(parabuf, " ");
				j++;
			}

			if (oldmode.key[0] && !chptr->mode.key[0])
			{
				modebuf[j] = 'k';
				strcat(parabuf, oldmode.key);
				strcat(parabuf, " ");
				j++;
			}

			if (oldmode.link[0] && !chptr->mode.link[0])
			{
				modebuf[j] = 'L';
				strcat(parabuf, oldmode.link);
				strcat(parabuf, " ");
				j++;
			}

			if ((oldmode.fban || oldmode.msgs || oldmode.secs) && ((chptr->mode.fban == 0) && (chptr->mode.msgs == 0)
				&& (chptr->mode.secs == 0)))
			{
				(void)ircsprintf(tmpbuf, "%s%i:%i", ((oldmode.fban == 1) ? "*" : ""), oldmode.msgs, oldmode.secs);
				modebuf[j] = 'f';
				strcat(parabuf, tmpbuf);
				strcat(parabuf, " ");
				*tmpbuf = '\0';
				j++;
			}

			for (tab = cFlagTab; tab->mode; tab++)
			{
				if ((oldmode.mode & tab->mode) && !(chptr->mode.mode & tab->mode) && !tab->param)
				{
					modebuf[j] = tab->flag;
					j++;
				}
			}

			if (j > 1)
			{
				strcat(modebuf, "+");
			}
			else
			{
				modebuf[0] = '+';
				j = 1;
			}

			if (!oldmode.limit && chptr->mode.limit)
			{
				modebuf[j] = 'l';
				strcat(parabuf, (char *)my_itoa(chptr->mode.limit));
				strcat(parabuf, " ");
				j++;
			}

			if (!oldmode.key[0] && chptr->mode.key[0])
			{
				modebuf[j] = 'k';
				strcat(parabuf, chptr->mode.key);
				strcat(parabuf, " ");
				j++;
			}

			if (!oldmode.link[0] && chptr->mode.link[0])
			{
				modebuf[j] = 'L';
				strcat(parabuf, chptr->mode.link);
				strcat(parabuf, " ");
				j++;
			}

			if ((oldmode.fban == 0) && (oldmode.msgs == 0) && (oldmode.secs == 0) && (chptr->mode.fban || chptr->mode.msgs
				|| chptr->mode.secs))
			{
				(void)ircsprintf(tmpbuf, "%s%i:%i", ((chptr->mode.fban == 1) ? "*" : ""), chptr->mode.msgs, chptr->mode.secs);
				modebuf[j] = 'f';
				strcat(parabuf, tmpbuf);
				strcat(parabuf, " ");
				*tmpbuf = '\0';
				j++;
			}

			/* check which parameter to use - longest gets it */
			if (oldmode.limit && chptr->mode.limit && (oldmode.limit != chptr->mode.limit))
			{
				chptr->mode.limit = Max(oldmode.limit, chptr->mode.limit);

				if (oldmode.limit != chptr->mode.limit)
				{
					modebuf[j] = 'l';
					strcat(parabuf, (char *)my_itoa(chptr->mode.limit));
					strcat(parabuf, " ");
					j++;
				}
			}

			if (oldmode.key[0] && chptr->mode.key[0] && strcmp(oldmode.key, chptr->mode.key))
			{
				x = strlen(oldmode.key);
				y = strlen(chptr->mode.key);
				z = Max(x, y);

				if (x == y)
				{
					sendto_channel_butserv(chptr, cptr, ":%s NOTICE %s :*** WARNING! Possible desynch: unable to"
						" determine which key (+k) to use in SJOIN for %s - removing all keys", me.name,
						chptr->chname, chptr->chname);
					keyerror++;
				}
				else
				{
					if (z == x)
					{
						strcpy(chptr->mode.key, oldmode.key);
					}
					else
					{
						modebuf[j] = 'k';
						strcat(parabuf, chptr->mode.key);
						strcat(parabuf, " ");
						j++;
					}
				}
			}

			if (oldmode.link[0] && chptr->mode.link[0] && strcmp(oldmode.link, chptr->mode.link))
			{
				x = strlen(oldmode.link);
				y = strlen(chptr->mode.link);
				z = Max(x, y);

				if (x == y)
				{
					sendto_channel_butserv(chptr, cptr, ":%s NOTICE %s :*** WARNING! Possible desynch: unable to"
						" determine which link (+L) to use in SJOIN for %s - removing all links", me.name,
						chptr->chname, chptr->chname);
					linkerror++;
				}
				else
				{
					if (z == x)
					{
						strcpy(chptr->mode.link, oldmode.link);
					}
					else
					{
						modebuf[j] = 'L';
						strcat(parabuf, chptr->mode.link);
						strcat(parabuf, " ");
						j++;
					}
				}
			}

			if ((oldmode.fban != chptr->mode.fban) || (oldmode.msgs != chptr->mode.msgs) || (oldmode.secs != chptr->mode.secs))
			{
				chptr->mode.fban = Max(oldmode.fban, chptr->mode.fban);
				chptr->mode.msgs = Max(oldmode.msgs, chptr->mode.msgs);
				chptr->mode.secs = Max(oldmode.secs, chptr->mode.secs);

				if ((oldmode.fban != chptr->mode.fban) || (oldmode.msgs != chptr->mode.msgs) || (oldmode.secs != chptr->mode.secs))
				{
					(void)ircsprintf(tmpbuf, "%s%i:%i", ((chptr->mode.fban == 1) ? "*" : ""), chptr->mode.msgs,
						chptr->mode.secs);
					modebuf[j] = 'f';
					strcat(parabuf, tmpbuf);
					strcat(parabuf, " ");
					*tmpbuf = '\0';
					j++;
				}
			}

			strcat(modebuf, "\0");
			if (modebuf[1] != '\0')
			{
				sendto_serv_butone(cptr, ":%s MODE %s %s %s %lu", me.name, chptr->chname, modebuf, parabuf,
					chptr->creationtime);
				sendto_channel_butserv(chptr, cptr, ":%s MODE %s %s %s", me.name, chptr->chname, modebuf, parabuf);
			}

			if (keyerror || linkerror)
			{
				e = 1;
				*modebuf = *parabuf = '\0';;
				modebuf[0] = '-';
				modebuf[e] = '\0';

				if (keyerror)
				{
					modebuf[e] = 'k';
					strcat(parabuf, chptr->mode.key);
					strcat(parabuf, " ");
					e++;
				}
				if (linkerror)
				{
					modebuf[e] = 'L';
					strcat(parabuf, chptr->mode.link);
					strcat(parabuf, " ");
					e++;
				}
				modebuf[e] = '\0';

				ap = mp2parv(modebuf, parabuf);
				set_mode(chptr, cptr, ap->parc, ap->parv, &pcount, pvar, 0);

				sendto_serv_butone(cptr, ":%s MODE %s %s %s %lu", me.name, chptr->chname, modebuf, parabuf,
					chptr->creationtime);
				sendto_channel_butserv(chptr, cptr, ":%s MODE %s %s %s", me.name, chptr->chname, modebuf, parabuf);
			}
		}
	}

	if (oldts != -1)
	{
		if (oldts != chptr->creationtime)
		{
			sendto_channel_butserv(chptr, cptr, ":%s NOTICE %s :*** TS for %s changed from %lu to %lu", me.name, chptr->chname,
				chptr->chname, oldts, ts);
		}
	}

	*parabuf2 = '\0';
	if (!noparams)
	{
		for (j = 4; j <= (parc - 2); j++)
		{
			strcat(parabuf2, parv[j]);
			strcat(parabuf2, " ");
		}
	}

	sendto_serv_butone(cptr, ":%s SJOIN %s %s %s %s :%s", parv[0], parv[1], parv[2], parv[3],
		((*parabuf2 != '\0') ? parabuf2 : "<none>"), parv[parc - 1]);
	return 0;
}


/*
 * m_knock
 *
 * parv[0] - sender prefix
 * parv[1] - channel
 * parv[2] - reason (if specified)
 *
 * Recoded 17/08/00 by Th3Dud3.
 */
int m_knock(cptr, sptr, parc, parv)
	aClient		*cptr, *sptr;
	int		parc;
	char		*parv[];
{
	aChannel	*chptr;
	char		*reason;

	if (check_registered_user(sptr))
		return 0;

	if (parc < 2 || *parv[1] == '\0')
	{
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "KNOCK");
		return -1;
	}

	if (MyConnect(sptr))
		clean_channelname(parv[1]);
	if (check_channelmask(sptr, cptr, parv[1]))
		return 0;

	if (*parv[1] != '#' && *parv[1] != '&')
	{
		sendto_one(sptr, err_str(ERR_CANNOTKNOCK), me.name, parv[0], parv[1],
			"Remember to unclude a # or & prefix in channel name");
		return 0;
	}

	if (!(chptr = find_channel(parv[1], NullChn)))
	{
		sendto_one(sptr, err_str(ERR_CANNOTKNOCK), me.name, parv[0], parv[1],
			"Channel does not exist");
		return 0;
	}

	if (IsMember(sptr, chptr) == 1)
	{
		sendto_one(sptr, err_str(ERR_CANNOTKNOCK), me.name, parv[0], parv[1],
			"You're already on that channel");
		return 0;
	}

	if (chptr->mode.mode & MODE_NOKNOCKS)
	{
		sendto_one(sptr, err_str(ERR_CANNOTKNOCK), me.name, parv[0], parv[1],
			"No knock's are allowed (+K)");
		return 0;
	}

	if (!(chptr->mode.mode & MODE_INVITEONLY))
	{
		sendto_one(sptr, err_str(ERR_CANNOTKNOCK), me.name, parv[0], parv[1],
			"Channel is not invite only (+i)");
		return 0;
	}

	if (chptr->mode.mode & MODE_NOINVITES)
	{
		sendto_one(sptr, err_str(ERR_CANNOTKNOCK), me.name, parv[0], parv[1],
			"No invite's are allowed (+I)");
		return 0;
	}

	if (is_banned(sptr, chptr))
	{
		if (IsOper(sptr) && (chptr->mode.mode & MODE_OPERBAN))
		{
			sendto_one(sptr, err_str(ERR_CANNOTKNOCK), me.name, parv[0], parv[1],
				"You're banned (+b) and channel does not allow oper's to walk though bans (+B)");
		}
		else
		{
			sendto_one(sptr, err_str(ERR_CANNOTKNOCK), me.name, parv[0], parv[1],
				"You're banned (+b)");
		}
		return 0;
	}

	reason = (parv[2] ? parv[2] : "no reason specified");

	sendto_channelops_butone(NULL, &me, chptr, ":%s NOTICE @%s :*** Knock -- %s (%s@%s) wants to come in (%s)",
		me.name, parv[1], parv[0], sptr->user->username,
		IsMasked(sptr) ? sptr->user->maskedhost : sptr->user->host, reason);

	sendto_one(sptr, ":%s NOTICE %s :*** Knock -- Knocked on to %s (%s)", me.name, parv[0], parv[1], reason);
	return 0;
}

/*
 * m_mode -- written by binary (garryb@binary.islesfan.net)
 *      Completely rewrote it.  The old mode command was 820 lines of ICKY
 * coding, which is a complete waste, because I wrote it in 570 lines of
 * *decent* coding.  This is also easier to read, change, and fine-tune.  Plus,
 * everything isn't scattered; everything's grouped where it should be.
 *
 * Modified by [-Th3Dud3-].
 *
 * parv[0] - sender
 * parv[1] - channel
 */
int	m_mode(cptr, sptr, parc, parv)
	aClient				*cptr, *sptr;
	int				parc;
	char				*parv[];
{
	static		char		tmp[MODEBUFLEN];
	static		char		mode_buf[MODEBUFLEN], parabuf[MODEBUFLEN];
	long		unsigned	timestamp, sendts;
	aChannel			*chptr;
	Ban				*ban;
	Ban				*exban;
	aForward			*af;

	if (check_registered(sptr))
		return 0;

	/* Now, try to find the channel in question. */
	if (parc > 1)
	{
		if (*parv[1] == '#' || *parv[1] == '&')
		{
			chptr = find_channel(parv[1], NullChn);
			if (chptr == NullChn)
			{
				return m_umode(cptr, sptr, parc, parv);
			}
		}
		else
		{
			return m_umode(cptr, sptr, parc, parv);
		}
	}
	else
	{
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "MODE");
		return 0;
	}

	sptr->flags &= ~FLAGS_TS8;

	if (MyConnect(sptr))
		clean_channelname(parv[1]);
	if (check_channelmask(sptr, cptr, parv[1]))
		return 0;

	if (parc < 3)
	{
		*mode_buf = *parabuf = '\0';
		mode_buf[1] = '\0';

		channel_modes(sptr, mode_buf, parabuf, chptr);
		sendto_one(sptr, rpl_str(RPL_CHANNELMODEIS), me.name, parv[0], chptr->chname,
			mode_buf, parabuf);
		sendto_one(sptr, rpl_str(RPL_CREATIONTIME), me.name, parv[0], chptr->chname,
			chptr->creationtime);

		return 0;
	}

	if (IsPerson(sptr) && (parc < 4) && (((*parv[2] == 'b') && (parv[2][1] == '\0')) || ((parv[2][1] == 'b') && (parv[2][2] == '\0')
		&& ((*parv[2] == '+') || (*parv[2] == '-')))))
	{
		if (!IsMember(sptr, chptr))
			return 0;

		/* send ban list */
		for (ban = chptr->banlist; ban; ban = ban->next)
		{
			sendto_one(sptr, rpl_str(RPL_BANLIST), me.name, sptr->name, chptr->chname,
				ban->banstr, ban->who, ban->when);
		}
		sendto_one(sptr, rpl_str(RPL_ENDOFBANLIST), me.name, sptr->name, chptr->chname);
		return 0;
	}

	if (IsPerson(sptr) && (parc < 4) && (((*parv[2] == 'e') && (parv[2][1] == '\0')) || ((parv[2][1] == 'e') && (parv[2][2] == '\0')
		&& ((*parv[2] == '+') || (*parv[2] == '-')))))
	{
		if (!IsMember(cptr, chptr))
			return 0;

		/* send exban list */
		for (exban = chptr->exbanlist; exban; exban = exban->next)
		{
			sendto_one(sptr, rpl_str(RPL_EXBANLIST), me.name, sptr->name, chptr->chname,
				exban->banstr, exban->who, exban->when);
		}
		sendto_one(sptr, rpl_str(RPL_EXBANLISTEND), me.name, sptr->name, chptr->chname);
		return 0;
	}

	if (IsPerson(sptr) & (parc < 4) && (((*parv[2] == 'F') && (parv[2][1] == '\0')) || ((parv[2][1] == 'F') && (parv[2][2] == '\0')
		&& ((*parv[2] == '+') || (*parv[2] == '-')))))
	{
		if (!IsMember(cptr, chptr))
			return 0;

		/* send forward list */
		for (af = chptr->forwardlist; af; af = af->next)
		{
			sendto_one(sptr, rpl_str(RPL_FORWARDLIST), me.name, sptr->name, chptr->chname,
				af->hostname, af->channel, af->who, af->when);
		}
		sendto_one(sptr, rpl_str(RPL_FORWARDLISTEND), me.name, sptr->name, chptr->chname);
		return 0;
	}

	opermode = 0;
	if (IsPerson(sptr) && !IsULine(cptr, sptr) && !is_chan_op(sptr, chptr))
	{
		if (IsOper(sptr))
		{
			sendts = 0;
			opermode = 1;
			goto aftercheck;
		}
	}

	if (is_half_op(sptr, chptr) && (opermode == 0))
	{
		opermode = 2;
		goto aftercheck;
	}

	if (IsPerson(sptr) && !IsULine(cptr, sptr) && !is_chan_op(sptr, chptr) && !is_half_op(sptr, chptr)
		&& (cptr == sptr || !IsSAdmin(sptr) || !IsOper(sptr)))
	{
		if (cptr == sptr)
		{
			sendto_one(sptr, err_str(ERR_CHANOPRIVSNEEDED), me.name, parv[0], chptr->chname);
			return 0;
		}

		/* Tell the other server that the user is de-op/halfop'd. Op/HalfOp desync fix. */
		sendto_one(cptr, ":%s MODE %s -oh %s 0", me.name, chptr->chname, parv[0]);
		bounce_mode(chptr, cptr, parc - 2, parv + 2);
		return 0;
	}
	sendts = 0;

	if (IsServer(sptr) && (sendts = atoi(parv[parc - 1])) && !IsULine(cptr, sptr) &&
		chptr->creationtime && sendts > chptr->creationtime)
	{
		/* & denotes a bounce */
		if (!(*parv[2] == '&'))
		{
			sendto_ops("TS bounce for %s [ours: %lu theirs: %lu]", chptr->chname,
				chptr->creationtime, sendts);
			bounce_mode(chptr, cptr, parc - 2, parv + 2);
			return 0;
		}
		/* other servers will resync soon enough... */
	}

	if (IsServer(sptr) && !sendts && *parv[parc - 1] != '0')
		sendts = -1;

	if (IsServer(sptr) && sendts != -1)
	{
		/* server supplied timestamp, remove it now */
		parc--;
	}

	/* Now that the unprivileged has been sorted out, we can get on
	 * and do the mode.
	 */

      aftercheck:
	(void)do_mode(chptr, cptr, sptr, parc - 2, parv + 2, sendts);

	return 0;
}

/* bounce_mode -- written by binary
 *      User or server is NOT authorized to change the mode.  This takes care
 * of making the bounce string and bounce it.  Because of the 1 for the bounce
 * param (last param) of the calls to set_mode and make_mode_str, it will not
 * set the mode, but create the bounce string.
 */
void	bounce_mode(chptr, cptr, parc, parv)
	aChannel		*chptr;
	aClient			*cptr;
	int			parc;
	char			*parv[];
{
	char			pvar[MAXMODEPARAMS][MODEBUFLEN + 3];
	char			mode_buf[MODEBUFLEN], parabuf[MODEBUFLEN], oldL[LINKLEN + 1];
	int			pcount;
	u_int			oldm, oldl;

	oldm = chptr->mode.mode;
	oldl = chptr->mode.limit;
	strcpy(oldL, chptr->mode.link);
	strcat(oldL, " ");

	set_mode(chptr, cptr, parc, parv, &pcount, pvar, 1);
	make_mode_str(chptr, oldm, oldl, oldL, pcount, pvar, mode_buf, parabuf, 1);

	/* the '&' denotes a bounce, so servers wont bounce a bounce */
	if (chptr->creationtime)
	{
		sendto_one(cptr, ":%s MODE %s &%s %s %lu", me.name, chptr->chname, mode_buf, parabuf, chptr->creationtime);
	}
	else
	{
		sendto_one(cptr, ":%s MODE %s &%s %s", me.name, chptr->chname, mode_buf, parabuf);
	}
}

/* do_mode -- written by binary
 *      User or server is authorized to do the mode.  This takes care of
 * setting the mode and relaying it to other users and servers.
 */
void	do_mode(chptr, cptr, sptr, parc, parv, sendts)
	aChannel	*chptr;
	aClient		*cptr, *sptr;
	int		parc;
	char		*parv[];
	int		sendts;
{
	char		pvar[MAXMODEPARAMS][MODEBUFLEN + 3], oldL[LINKLEN + 1];
	int		pcount;
	u_int		oldm, oldl;
	char		tschange = 0, isbounce = 0;
	char		mode_buf[MODEBUFLEN], parabuf[MODEBUFLEN];

	if (**parv == '&')
		isbounce = 1;

	oldm = chptr->mode.mode;
	oldl = chptr->mode.limit;
	strcpy(oldL, chptr->mode.link);
	strcat(oldL, " ");

	set_mode(chptr, sptr, parc, parv, &pcount, pvar, 0);
        make_mode_str(chptr, oldm, oldl, oldL, pcount, pvar, mode_buf, parabuf, 0);

	if (IsServer(sptr))
	{
		if (sendts > 0)
		{
			if (!chptr->creationtime || sendts < chptr->creationtime)
			{
				/* new channel or our timestamp is wrong, so lets fix it.
				 * now works for double-bounce prevention
				 */
				tschange = 1;
				chptr->creationtime = sendts;
			}
			if ((sendts > chptr->creationtime) && chptr->creationtime)
			{
					/* their timestamp is wrong, but we'll let it pass */
					sendts = chptr->creationtime;
					sendto_one(cptr, ":%s MODE %s + %lu", me.name, chptr->chname, chptr->creationtime);
			}
		}
		if ((sendts == -1) && chptr->creationtime)
			sendts = chptr->creationtime;
	}
	if (*mode_buf == '\0' || (*(mode_buf + 1) == '\0' && (*mode_buf == '+' || *mode_buf == '-')))
	{
		/* relay bounce time changes */
		if (tschange || isbounce)
		{
			if (chptr->creationtime)
			{
				sendto_serv_butone(cptr, ":%s MODE %s %s+ %lu", me.name, chptr->chname,
					isbounce ? "&" : "", chptr->creationtime);
			}
			else
			{
				sendto_serv_butone(cptr, ":%s MODE %s %s+", me.name, chptr->chname,
					isbounce ? "&" : "", chptr->creationtime);
			}
		}
		return;
	}
	if (opermode == 1)
	{
		if (mode_buf[1])
		{
			sendto_umode(UMODE_OPER, "*** OperMode -- [IRCop: %s] - [Channel: %s] - [Mode: %s %s]",
				sptr->name, chptr->chname, mode_buf, parabuf);
		}
		sendts = 0;
	}

	if (*(mode_buf + 1) == '\0')
	{
		return;
	}

	sendto_channel_butserv(chptr, sptr, ":%s MODE %s %s %s", sptr->name, chptr->chname,
		mode_buf, parabuf);

	if (IsServer(sptr) && (sendts != -1))
	{
		sendto_serv_butone(&me, ":%s MODE %s %s%s %s %lu", sptr->name, chptr->chname,
			isbounce ? "&" : "", mode_buf, parabuf, sendts);
	}
	else
	{
		/* tell them that it's not a timestamp, just incase the last
		 * param is a number.
		 */
		sendto_serv_butone(&me, ":%s MODE %s %s%s %s NoTS", sptr->name, chptr->chname,
			isbounce ? "&" : "", mode_buf, parabuf);
	}
}

/* make_mode_str -- written by binary
 *      Reconstructs the mode string, to make it look clean.  mode_buf will
 *  contain the +x-y stuff, and the parabuf will contain the parameters.
 *  If bounce is set to 1, it will make the string it needs for a bounce.
 */
void	make_mode_str(chptr, oldm, oldl, oldL, pcount, pvar, mode_buf, parabuf, bounce)
	aChannel	*chptr;
	u_int		oldm, oldl;
	int		pcount;
	char		pvar[MAXMODEPARAMS][MODEBUFLEN+3], *mode_buf, *parabuf, bounce, oldL[LINKLEN + 1];
{
	char		*x = mode_buf, tmpbuf[MODEBUFLEN], *tmpstr;
	int		what, cnt, z, linkchg = 0;
	aCtab		*tab = &cFlagTab[0];

	what = 0;

	*tmpbuf = '\0';
	*mode_buf = '\0';
	*parabuf = '\0';

	/* + param-less mode */
	tab = &cFlagTab[0];
	while (tab->mode != 0x0)
	{
		if (chptr->mode.mode & tab->mode)
		{
			if (!(oldm & tab->mode))
			{
				if (what != MODE_ADD)
				{
					*x++ = bounce ? '-' : '+';
					what = MODE_ADD;
				}
				*x++ = tab->flag;
			}
		}
		tab++;
	}
	*x = '\0';

	/* - param-less mode */
	tab = &cFlagTab[0];
	while (tab->mode != 0x0)
	{
		if (!(chptr->mode.mode & tab->mode))
		{
			if (oldm & tab->mode)
			{
				if (what != MODE_DEL)
				{
					*x++ = bounce ? '+' : '-';
					what = MODE_DEL;
				}
				*x++ = tab->flag;
			}
		}
		tab++;
	}
	*x = '\0';

	/* reconstruct the bekohv chain */
	for (cnt = 0; cnt < pcount; cnt++)
	{
		if ((*(pvar[cnt]) == '+') && what != MODE_ADD)
		{
			*x++ = bounce ? '-' : '+';
			what = MODE_ADD;
		}
		if ((*(pvar[cnt]) == '-') && what != MODE_DEL)
		{
			*x++ = bounce ? '+' : '-';
			what = MODE_DEL;
		}
		*x++ = *(pvar[cnt] + 1);
		tmpstr = &pvar[cnt][2];
		strncat(parabuf, tmpstr, MODEBUFLEN - 1);
		parabuf[MODEBUFLEN - 1] = '\0';
		z = strlen(parabuf);
		parabuf[z] = ' ';
		parabuf[z + 1] = '\0';
	}

	/* user limit */
	if (chptr->mode.limit != oldl)
	{
		if ((!bounce && chptr->mode.limit == 0) || (bounce && chptr->mode.limit != 0))
		{
			if (what != MODE_DEL)
			{
				*x++ = '-';
				what = MODE_DEL;
			}

			if (chptr->mode.link[0])
			{
				if (!bounce)
				{
					strcpy(chptr->mode.link, "");
				}
				*x++ = 'L';
				linkchg++;
			}
			if (bounce)
			{
				chptr->mode.limit = 0;
			}
			*x++ = 'l';

			if (linkchg > 0)
			{
				(void)ircsprintf(parabuf, "%s%s%d", parabuf, oldL, oldl);
			}
			else
			{
				(void)ircsprintf(parabuf, "%s%d", parabuf, oldl);
			}
		}
		else
		{
			if (what != MODE_ADD)
			{
				*x++ = '+';
				what = MODE_ADD;
			}
			*x++ = 'l';
			if (bounce)
			{
				chptr->mode.limit = oldl;
			}
			(void)ircsprintf(parabuf, "%s%d", parabuf, chptr->mode.limit);
		}
	}
	if (bounce)
	{
		chptr->mode.mode = oldm;
	}
	z = strlen(parabuf);

	if (parabuf[z - 1] == ' ')
	{
		parabuf[z - 1] = '\0';
	}
	*x = '\0';

	if (*mode_buf == '\0')
	{
		/* dont send empty lines */
		*mode_buf = '+';
		mode_buf++;
		*mode_buf = '\0';
	}
	return;
}

/* do_mode_char
 *  processes one mode character      
 *  returns 1 if it ate up a param, otherwise 0
 *      written by binary
 */
int	do_mode_char(chptr, modetype, modechar, param, what, cptr, pcount, pvar, bounce)
	aChannel		*chptr;
	u_int			modetype;
	char			modechar, *param;
	u_int			what;
	aClient			*cptr;
	u_int			*pcount;
	char			pvar[MAXMODEPARAMS][MODEBUFLEN+3], bounce;
{
	char			modeflag = ' ', *modename;
	int			retval = 0;
	Link			*member;
	aClient			*who;
	unsigned	int	tmp = 0;
	char			tmpbuf[256], *tmpstr;
	int			chasing, x;
	aCtab			*tab = &cFlagTab[0];
	char			*xp;
	int			xxx, yyy, zzz, hascolon;
	char			*xx, *yy;

	chasing = 0;

	if ((opermode == 2) && !is_chan_op(cptr, chptr) && !IsULine(cptr, cptr))
	{
		if (check_halfop_mode(modetype) == FALSE)
		{
			while (tab->mode != 0x0)
			{
				if (tab->mode == modetype)
				{
					sendto_one(cptr, ":%s NOTICE %s :*** Cannot change mode for channel %s - you are only a halfop.",
						me.name, cptr->name, chptr->chname);
				}
				tab++;
			}
			return (0);
		}
	}

	switch (modetype)
	{
		/*
		 * Modes that DO NOT require parameters but do require permissions.
		 */
		case MODE_RGSTR:
			if (!IsServer(cptr) && !IsULine(cptr, cptr))
				break;
		case MODE_NETADMINSONLY:
			if (!IsServer(cptr) && !IsULine(cptr, cptr) && !IsNetAdmin(cptr) && !IsCoNetAdmin(cptr) && !IsTechAdmin(cptr))
			{
				if (IsPerson(cptr))
				{
					sendto_one(cptr, ":%s NOTICE %s :*** Only Network Administrators can set that mode.", me.name, cptr->name);
				}
				break;
			}
			if (what == MODE_ADD)
			{
				if (modetype == MODE_NETADMINSONLY)
				{
					if (chptr->mode.mode & MODE_ADMINSONLY)
					{
						chptr->mode.mode &= ~MODE_ADMINSONLY;
					}
					if (chptr->mode.mode & MODE_IRCOPSONLY)
					{
						chptr->mode.mode &= ~MODE_IRCOPSONLY;
					}
				}
			}
		case MODE_ADMINSONLY:
			if (!IsServer(cptr) && !IsULine(cptr, cptr) && !(IsNetAdmin(cptr) || IsCoNetAdmin(cptr) || IsTechAdmin(cptr)
				|| IsAdmin(cptr) || IsSCoAdmin(cptr)))
			{
				if (IsPerson(cptr))
				{
					sendto_one(cptr, ":%s NOTICE %s :*** Only Administrators can set that mode.", me.name, cptr->name);
				}
				break;
			}
			if ((modetype == MODE_ADMINSONLY) && (what == MODE_ADD))
			{
				if (chptr->mode.mode & MODE_ADMINSONLY)
				{
					if (IsPerson(cptr))
					{
						sendto_one(cptr, ":%s NOTICE %s :*** Channel %s is currently Network"
							" Administrators only - cannot set +A.", me.name, cptr->name,
							chptr->chname);
					}
					break;
				}
			}
			if (what == MODE_ADD)
			{
				if (modetype == MODE_ADMINSONLY)
				{
					if (chptr->mode.mode & MODE_IRCOPSONLY)
					{
						chptr->mode.mode &= ~MODE_IRCOPSONLY;
					}
				}
			}
		case MODE_IRCOPSONLY:
			if (!IsServer(cptr) && !IsULine(cptr, cptr) && !IsOper(cptr))
			{
				if (IsPerson(cptr))
				{
					sendto_one(cptr, ":%s NOTICE %s :*** Only IRC Operators can set that mode.", me.name, cptr->name);
				}
				break;
			}
			if ((modetype == MODE_IRCOPSONLY) && (what == MODE_ADD))
			{
				if (chptr->mode.mode & MODE_NETADMINSONLY)
				{
					if (IsPerson(cptr))
					{
						sendto_one(cptr, ":%s NOTICE %s :*** Channel %s is currently"
							" Network Administrators only - cannot set +O.", me.name,
							cptr->name, chptr->chname);
					}
					break;
				}
				if (chptr->mode.mode & MODE_ADMINSONLY)
				{
					if (IsPerson(cptr))
					{
						sendto_one(cptr, ":%s NOTICE %s :*** Channel %s is currently"
							" Administrators only - cannot set +O.", me.name,
							cptr->name, chptr->chname);
					}
					break;
				}
			}
		case MODE_OPERBAN:
			if (!IsServer(cptr) && !IsULine(cptr, cptr) && !IsNetAdmin(cptr) && !IsCoNetAdmin(cptr))
			{
				if (IsPerson(cptr))
				{
					sendto_one(cptr, ":%s NOTICE %s :*** Only Network Administrators can set that mode.", me.name, cptr->name);
				}
				break;
			}
		case MODE_NOKNOCKS:
			if (modetype == MODE_NOKNOCKS)
			{
				if ((what == MODE_ADD) && !(chptr->mode.mode & MODE_INVITEONLY))
				{
					if (IsPerson(cptr))
					{
						sendto_one(cptr, ":%s NOTICE %s :*** Cannot set channel %s +K - channel is not +i (invite only).",
							me.name, cptr->name, chptr->chname);
					}
					break;
				}
			}
		case MODE_INVITEONLY:
			if ((what == MODE_DEL) && (modetype == MODE_INVITEONLY)
				&& (chptr->mode.mode & MODE_NOKNOCKS))
			{
				chptr->mode.mode &= ~MODE_NOKNOCKS;
			}
		/*
		 * Modes that DO NOT require any parameters or permissions.
		 */
		case MODE_TOPICLIMIT:
		case MODE_SECRET:
		case MODE_MODERATED:
		case MODE_NOINVITES:
		case MODE_PEACE:
		case MODE_NOPRIVMSGS:
		case MODE_PRIVATE:
		case MODE_RGSTRONLY:
		case MODE_STRIPBADWORDS:
		case MODE_NOCTCP:
		case MODE_NONICKCH:
		case MODE_STRIPCOLORS:
		case MODE_NOCOLORS:
		case MODE_STRIPFORMAT:
			if ((modetype == MODE_SECRET) && (chptr->mode.mode & MODE_PRIVATE))
			{
				chptr->mode.mode &= ~MODE_PRIVATE;
			}
			if ((modetype == MODE_PRIVATE) && (chptr->mode.mode & MODE_SECRET))
			{
				chptr->mode.mode &= ~MODE_SECRET;
			}
			if ((modetype == MODE_STRIPCOLORS) && (chptr->mode.mode & MODE_NOCOLORS))
			{
				chptr->mode.mode &= ~MODE_NOCOLORS;
			}
			if ((modetype == MODE_NOCOLORS) && (chptr->mode.mode & MODE_STRIPCOLORS))
			{
				chptr->mode.mode &= ~MODE_STRIPCOLORS;
			}

			retval = 0;
			if (what == MODE_ADD)
			{
				chptr->mode.mode |= modetype;
			}
			else
			{
				chptr->mode.mode &= ~modetype;
			}
			break;

		/*
		 * Modes that DO require parameters and do require permissions.
		 */
		case MODE_CHANADMIN:
			if (!IsServer(cptr) && !IsULine(cptr, cptr) && !IsNetAdmin(cptr)
				&& !IsCoNetAdmin(cptr) && !IsTechAdmin(cptr) && !is_chanadmin(cptr, chptr))
			{
				sendto_one(cptr, err_str(ERR_ONLYSERVERSCANCHANGE), me.name, cptr->name,
					chptr->chname);
				break;
			}

		/*
		 * Modes that DO require parameters but do not require permissions.
		 */
		case MODE_CHANOP:
		case MODE_HALFOP:
		case MODE_VOICE:
			switch (modetype)
			{
				case MODE_CHANADMIN:
					modename = "de-admin";
					break;
				case MODE_CHANOP:
					modename = "de-op";
					break;
				case MODE_HALFOP:
					modename = "de-halfop";
					break;
				case MODE_VOICE:
					modename = "de-voice";
					break;
				default:
					modename = "de-?";
			}

			if (!param || *pcount >= MAXMODEPARAMS)
			{
				retval = 0;
				break;
			}
			retval = 1;

			if (!(who = find_chasing(cptr, param, &chasing)))
				break;
			if (!(member = find_user_link(chptr->members, who)))
			{
				sendto_one(cptr, err_str(ERR_USERNOTINCHANNEL), me.name, cptr->name,
					who->name, chptr->chname);
				break;
			}

			/*
			 * This will stop chanadmins from being de-admined, de-oped, de-halfoped or
			 * de-voiced accept by a server, services client (uline), senior admin, or
			 * another chanadmin user. -TD
			 */
			if ((what == MODE_DEL) && is_chanadmin(who, chptr) && !IsULine(cptr, cptr) && !IsServer(cptr))
			{
				if (!IsNetAdmin(cptr) && !IsCoNetAdmin(cptr) && !IsTechAdmin(cptr) && !is_chanadmin(cptr, chptr)
					&& (who != cptr))
				{
					if (MyClient(cptr))
					{
						sendto_one(cptr, ":%s NOTICE %s :*** Cannot %s %s (Channel Administrator)",
							me.name, cptr->name, modename, who->name);
					}
					break;
				}
			}

			tmp = member->flags;
			if (what == MODE_ADD)
				member->flags |= modetype;
			else
				member->flags &= ~modetype;

			if ((tmp == member->flags) && (bounce || !IsULine(cptr, cptr)))
				break;
			if (bounce)
				member->flags = tmp;

			if (modetype == MODE_CHANADMIN)
				modeflag = 'a';
			if (modetype == MODE_CHANOP)
				modeflag = 'o';
			if (modetype == MODE_HALFOP)
				modeflag = 'h';
			if (modetype == MODE_VOICE)
				modeflag = 'v';

			(void)ircsprintf(pvar[*pcount], "%c%c%s", ((what == MODE_ADD) ? '+' : '-'),
				modeflag, who->name);
			(*pcount)++;
			break;
		case MODE_LIMIT:
			if (what == MODE_ADD)
			{
				if (!param)
				{
					retval = 0;
					break;
				}
				retval = 1;
				tmp = atoi(param);

				if (chptr->mode.limit == tmp)
					break;

				chptr->mode.limit = tmp;
			}
			else
			{
				retval = 0;
				if (!(chptr->mode.limit))
					break;
				chptr->mode.limit = 0;
			}
			break;
		case MODE_KEY:
			if (!param || *pcount >= MAXMODEPARAMS)
			{
				retval = 0;
				break;
			}
			retval = 1;

			for (x = 0; x < *pcount; x++)
			{
				if (pvar[x][1] == 'k')
				{
					retval = 0;
					break;
				}
			}
			if (retval == 0)
				break;
			if (what == MODE_ADD)
			{
				if (!bounce)
				{
					strncpyzt(chptr->mode.key, param, sizeof(chptr->mode.key));
					tmpstr = param;
				}
			}
			else
			{
				if (!*chptr->mode.key)
					break;

				strncpyzt(tmpbuf, chptr->mode.key, sizeof(tmpbuf));
				tmpstr = tmpbuf;

				if (!bounce)
				{
					strcpy(chptr->mode.key, "");
				}
			}
			retval = 1;

			(void)ircsprintf(pvar[*pcount], "%ck%s", ((what == MODE_ADD) ? '+' : '-'), tmpstr);
			(*pcount)++;
			break;
		case MODE_BAN:
			if (!param || *pcount >= MAXMODEPARAMS)
			{
				retval = 0;
				break;
			}
			retval = 1;
			tmpstr = pretty_mask(param);

			if (!bounce && ((what == MODE_ADD && add_banid(cptr, chptr, tmpstr)) || (what == MODE_DEL && del_banid(chptr, tmpstr))))
				break;

			(void)ircsprintf(pvar[*pcount], "%cb%s", ((what == MODE_ADD) ? '+' : '-'), tmpstr);
			(*pcount)++;
			break;
		case MODE_EXBAN:
			if (!param || *pcount >= MAXMODEPARAMS)
			{
				retval = 0;
				break;
			}
			retval = 1;
			tmpstr = pretty_mask(param);

			if (!bounce && ((what == MODE_ADD && add_exbanid(cptr, chptr, tmpstr)) || (what == MODE_DEL && del_exbanid(chptr, tmpstr))))
					break;

			(void)ircsprintf(pvar[*pcount], "%ce%s", ((what == MODE_ADD) ? '+' : '-'), tmpstr);
			(*pcount)++;
			break;
		case MODE_LINK:
			if (IsULine(cptr, cptr) || IsServer(cptr))
			{
				goto linkok;
			}
			if (!IsNetAdmin(cptr) && !IsCoNetAdmin(cptr) && !IsTechAdmin(cptr) && !is_chanadmin(cptr, chptr))
			{
				sendto_one(cptr, ":%s NOTICE %s :*** Channel mode +L may only be set by channel administrators (+a)",
					me.name, cptr->name);
				break;
			}

				linkok:
			if (!(chptr->mode.limit) && (what == MODE_ADD))
			{
				if (MyClient(cptr))
				{
					sendto_one(cptr, ":%s NOTICE %s :*** Channel mode +l <max> is required for channel mode +L",
						me.name, cptr->name);
					break;
				}
			}

			retval = 1;
			for (x = 0; x < *pcount; x++)
			{
				if (pvar[x][1] == 'L')
				{
					retval = 0;
					break;
				}
			}
			if (retval == 0)
				break;

			if (what == MODE_ADD)
			{
				if (!param || *pcount >= MAXMODEPARAMS)
				{
					retval = 0;
					break;
				}

				if (!stricmp(param, chptr->chname))
				{
					if (MyClient(cptr))
					{
						sendto_one(cptr, ":%s NOTICE %s :*** You cannot link channel %s to itself", me.name,
							cptr->name, chptr->chname);
						break;
					}
				}

				if (check_restrictions(param) == 1)
				{
					sendto_one(cptr,":%s NOTICE %s :*** Cannot use %s in link - channel is restricted",
						me.name, cptr->name, param);
					break;
				}

				if (strchr(param, ','))
				{
					if (MyClient(cptr))
					{
						sendto_one(cptr, ":%s NOTICE %s :*** You cannot link to more chan one channel",
							me.name, cptr->name);
						break;
					}
				}
				if (!IsChannelName(param))
				{
					if (MyClient(cptr))
					{
						sendto_one(cptr, err_str(ERR_NOSUCHCHANNEL), me.name, cptr->name, param);
						break;
					}
				}
				if (!bounce)
				{
					strncpyzt(chptr->mode.link, param, sizeof(chptr->mode.link));
				}
				tmpstr = param;
			}
			else
			{
				if (!(*chptr->mode.link))
				{
					break;
				}
				strncpyzt(tmpbuf, chptr->mode.link, sizeof(tmpbuf));
				tmpstr = tmpbuf;

				if (!bounce)
				{
					strcpy(chptr->mode.link, "");
				}
			}
			retval = 1;
			(void)ircsprintf(pvar[*pcount], "%cL%s", ((what == MODE_ADD) ? '+' : '-'), tmpstr);
			(*pcount)++;
			break;
		case MODE_FORWARD:
			if (what == MODE_ADD)
			{
				if (!bounce)
				{
					if (!param || *pcount >= MAXMODEPARAMS)
					{
						retval = 0;
						break;
					}

					hascolon = 0;
					for (xp = param; *xp; xp++)
					{
						if (*xp == ':')
						{
							hascolon++;
						}
					}

					if (hascolon != 1)
					{
						break;
					}

					xp = index(param, ':');
					*xp = '\0';
					xx = pretty_mask(param);
					xp++;
					yy = xp;
					xp--;
					*xp = ':';

					if (!IsChannelName(yy))
					{
						sendto_one(cptr, err_str(ERR_NOSUCHCHANNEL), me.name, cptr->name, yy);
						break;
					}

					if (!strcmp(yy, chptr->chname))
					{
						sendto_one(cptr, ":%s NOTICE %s :*** Cannot set %s +F %s", me.name, cptr->name,
							chptr->chname, yy);
						break;
					}

					if (check_restrictions(yy) == 1)
					{
						sendto_one(cptr,":%s NOTICE %s :*** Cannot use %s in forwarding ban - channel"
							" is restricted", me.name, cptr->name, yy);
						break;
					}

					if (find_forwardid(chptr, xx, yy) == 1)
					{
						break;
					}

					(void)ircsprintf(tmpbuf, "%s:%s", xx, yy);
					add_forwardid(cptr, chptr, xx, yy);
				}

				tmpstr = tmpbuf;
				retval = 1;
			}
			else
			{
				if (!bounce)
				{
					if (!param || *pcount >= MAXMODEPARAMS)
					{
						retval = 0;
						break;
					}

					hascolon = 0;
					for (xp = param; *xp; xp++)
					{
						if (*xp == ':')
						{
							hascolon++;
						}
					}

					if (hascolon != 1)
					{
						break;
					}

					xp = index(param, ':');
					*xp = '\0';
					xx = pretty_mask(param);
					xp++;
					yy = xp;
					xp--;
					*xp = ':';

					if (!IsChannelName(yy))
					{
						sendto_one(cptr, err_str(ERR_NOSUCHCHANNEL), me.name, cptr->name, yy);
						break;
					}

					if (find_forwardid(chptr, xx, yy) == 0)
					{
						break;
					}

					(void)ircsprintf(tmpbuf, "%s:%s", xx, yy);
					del_forwardid(chptr, xx, yy);
				}

				tmpstr = tmpbuf;
				retval = 0;
			}

			(void)ircsprintf(pvar[*pcount], "%cF%s", ((what == MODE_ADD) ? '+' : '-'), tmpstr);
			(*pcount)++;
			break;

		case MODE_FLOODLIMIT:
			retval = 1;
			for (x = 0; x < *pcount; x++)
			{
				if (pvar[x][1] == 'f')
				{
					retval = 0;
					break;
				}
			}
			if (retval == 0)
				break;

			if (what == MODE_ADD)
			{
				if (!bounce)
				{
					if (!param || *pcount >= MAXMODEPARAMS)
					{
						retval = 0;
						break;
					}

					if (strlen(param) < 3)
						break;

					hascolon = 0;
					for (xp = param; *xp; xp++)
					{
						if (*xp == ':')
						{
							hascolon++;
						}

						if (((*xp < '0') || (*xp > '9')) && (*xp != ':') && (*xp != '*'))
						{
							sendto_one(cptr, ":%s NOTICE %s :*** Flood Limit may only contain numbers 1 to 9"
								" and :", me.name, cptr->name);
							break;
						}

						if ((*xp == '*') && (*param != '*'))
						{
							break;
						}
					}

					if (hascolon != 1)
					{
						break;
					}

					if (*param == '*')
					{
						xxx = 1;
					}
					else
					{
						xxx = 0;
					}
					
					xp = index(param, ':');
					*xp = '\0';
					yyy = atoi(((*param == '*') ? (param + 1) : param));
					xp++;
					zzz = atoi(xp);
					xp--;
					*xp = ':';

					if ((yyy == 0) || (zzz == 0))
					{
						sendto_one(cptr, ":%s NOTICE %s :*** Cannot set a flood limit of 0", me.name, cptr->name);
						break;
					}

					if ((chptr->mode.fban == xxx) && (chptr->mode.msgs == yyy) && (chptr->mode.secs == zzz))
						break;

					if (yyy < 2)
					{
						sendto_one(cptr, ":%s NOTICE %s :*** It's not a good idea to set a flood limit below 2 lines",
							me.name, cptr->name);
						break;
					}
					if (zzz < 2)
					{
						sendto_one(cptr, ":%s NOTICE %s :*** It's not a good idea to set a flood limit below 2"
							" seconds", me.name, cptr->name);
						break;
					}
					chptr->mode.fban = xxx;
					chptr->mode.msgs = yyy;
					chptr->mode.secs = zzz;
				}
				(void)ircsprintf(tmpbuf, "%s%i:%i", ((chptr->mode.fban == 1) ? "*" : ""), chptr->mode.msgs, chptr->mode.secs);
				tmpstr = tmpbuf;
				retval = 1;
			}
			else
			{
				if (!chptr->mode.msgs || !chptr->mode.secs)
				{
					break;
				}

				(void)ircsprintf(tmpbuf, "%s%i:%i", ((chptr->mode.fban == 1) ? "*" : ""), chptr->mode.msgs, chptr->mode.secs);
				tmpstr = tmpbuf;

				if (!bounce)
				{
					chptr->mode.msgs = chptr->mode.secs = chptr->mode.fban = 0;
				}
				retval = 0;
			}
			(void)ircsprintf(pvar[*pcount], "%cf%s", ((what == MODE_ADD) ? '+' : '-'), tmpstr);
			(*pcount)++;
			break;

	}
	return(retval);
}

/* set_mode
 *      written by binary
 */
void	set_mode(chptr, cptr, parc, parv, pcount, pvar, bounce)
	aChannel	*chptr;
	aClient		*cptr;
	int		parc;
	char		*parv[];
	u_int		*pcount;
	char		pvar[MAXMODEPARAMS][MODEBUFLEN+3], bounce;
{
	Reg1	char	*curchr;
	u_int		what = MODE_ADD; 
	int		modetype, paracount, found = 0;
	aCtab		*tab = &cFlagTab[0];
	aCtab		foundat;
                        
	paracount = 1;
	*pcount = 0;

	for (curchr = parv[0]; *curchr; curchr++)
	{
		switch (*curchr)
		{
			case '+':
				what = MODE_ADD;
				break;
			case '-':
				what = MODE_DEL;
				break;
			default:
				found = 0;
				tab = &cFlagTab[0];

				while ((tab->mode != 0x0) && (found == 0))
				{
					if (tab->flag == *curchr)
					{
						found = 1;
						foundat = *tab;
					}
					tab++;
				}

				if (found == 1)
					modetype = foundat.mode;
				if (found == 0)
					modetype = 0;

				if (modetype == 0)
				{
					if (!MyClient(cptr))
						break;

					sendto_one(cptr, err_str(ERR_UNKNOWNMODE), me.name, cptr->name,
						*curchr);
					break;
				}
				if (parv[paracount] && strlen(parv[paracount]) >= MODEBUFLEN)
				{
					parv[paracount][MODEBUFLEN - 1] = '\0';
				}

				paracount += do_mode_char(chptr, modetype, *curchr, parv[paracount],
					what, cptr, pcount, pvar, bounce);
				break;
		}
	}
}


/*
 * check_for_chan_flood
 *
 * Used in conjunction with channelmode +f
 */
int	check_for_chan_flood(cptr, sptr, chptr)
	aClient		*cptr, *sptr;
	aChannel	*chptr;
{
	Link		*lp;

	if (!MyClient(sptr) || IsULine(cptr, sptr) || IsServer(cptr))
		return 0;

	if (!(lp = find_user_link(chptr->members, sptr)))
		return 0;
	if ((chptr->mode.msgs < 1) || (chptr->mode.secs < 1))
		return 0;

	if ((TStime() - (lp->flood->lastmsg)) >= chptr->mode.secs)
	{
		lp->flood->lastmsg = TStime();
		lp->flood->nmsg = 1;
	}

	lp->flood->nmsg++;
	lp->flood->lastmsg = TStime();

	if ((lp->flood->nmsg) > chptr->mode.msgs)
	{
		char	comment[1024], mask[1024];

		if (chptr->mode.fban == 1)
		{
			(void)ircsprintf(mask, "*!*@%s", (IsMasked(sptr) ? sptr->user->maskedhost : sptr->user->host));

			add_banid(&me, chptr, mask);
			sendto_channel_butserv(chptr, &me, ":%s MODE %s +b %s", me.name, chptr->chname, mask);
			sendto_serv_butone(&me, ":%s MODE %s +b %s", me.name, chptr->chname, mask);
		}

		(void)ircsprintf(comment, "flooding (limit is %i lines within %i seconds)", chptr->mode.msgs, chptr->mode.secs);

		sendto_channel_butserv(chptr, &me, ":%s KICK %s %s :%s", me.name, chptr->chname, sptr->name, comment);
		sendto_serv_butone(cptr, ":%s KICK %s %s :%s", me.name, chptr->chname, sptr->name, comment);
		remove_user_from_channel(sptr, chptr);

		return 1;
	}
}


int DoesOp(modebuf)
    char *modebuf;
{
    modebuf--;			/* Is it possible that a mode starts with o and not +o ? */
    while (*++modebuf)
	if (*modebuf == 'a' || *modebuf == 'o' || *modebuf == 'h' || *modebuf == 'v')
	    return (1);
    return 0;
}

int sendmodeto_one(cptr, from, name, mode, param, creationtime)
    Reg2 aClient *cptr;
    char *from, *name, *mode, *param;
    time_t creationtime;
{
    if ((IsServer(cptr) && DoesOp(mode) && creationtime) ||
	IsULine(cptr, cptr))
	sendto_one(cptr, ":%s %s %s %s %s %lu", from,
		(IsToken(cptr) ? TOK_MODE : MSG_MODE), name, mode, param,
		   creationtime);
    else
	sendto_one(cptr, ":%s %s %s %s %s", from,
	       (IsToken(cptr) ? TOK_MODE : MSG_MODE), name, mode, param);
    return 0;
}

char *pretty_mask(mask)
    char *mask;
{
    Reg1 char *cp;
    Reg2 char *user;
    Reg3 char *host;

    if ((user = index((cp = mask), '!')))
	*user++ = '\0';
    if ((host = rindex(user ? user : cp, '@'))) {
	*host++ = '\0';
	if (!user)
	    return make_nick_user_host(NULL, cp, host);
    } else if (!user && index(cp, '.'))
	return make_nick_user_host(NULL, NULL, cp);
    return make_nick_user_host(cp, user, host);
}

/*
 * can_join
 *
 * Detects wether user may or may not join channel.
 * Recoded 17/08/00 by Th3Dud3.
 */
static int can_join(cptr, sptr, chptr, key, parv)
	aClient			*cptr, *sptr;
	Reg2	aChannel	*chptr;
	char			*key;
	char			*parv[];
{
	Reg1	Link		*lp;
	aForward		*af;
	aChannel		*afc, *linkptr;

	if (IsULine(sptr, sptr) || IsServer(sptr))
		return 0;

	if (*chptr->mode.key && (BadPtr(key) || mycmp(chptr->mode.key, key)) && !lp)
	{
		if (IsOper(sptr))
		{
			sendto_umode(UMODE_OPER, "*** OperWalk -- [Type: key] - [IRCop: %s] - [Channel: %s]", sptr->name, chptr->chname);
			return 0;
		}
		return (ERR_BADCHANNELKEY);
	}

	for (lp = sptr->user->invited; lp; lp = lp->next)
	{
		if (lp->value.chptr == chptr)
			break;
	}

	if ((chptr->mode.mode & MODE_INVITEONLY) && !lp)
	{
		if (IsOper(sptr))
		{
			sendto_umode(UMODE_OPER, "*** OperWalk -- [Type: invite] - [IRCop: %s] - [Channel: %s]", sptr->name, chptr->chname);
			return 0;
		}
		return (ERR_INVITEONLYCHAN);
	}

	if ((chptr->mode.limit && chptr->users >= chptr->mode.limit) && !lp)
	{
		if (IsOper(sptr))
		{
			sendto_umode(UMODE_OPER, "*** OperWalk -- [Type: link] - [IRCop: %s] - [Channel: %s]", sptr->name, chptr->chname);
			return 0;
		}

		if (*chptr->mode.link != '\0')
		{
			sendto_one(sptr, ":%s NOTICE %s :*** Channel %s has become full, so you are being automatically transferred to"
				" linked channel %s", me.name, sptr->name, chptr->chname, chptr->mode.link);

			linkptr = find_channel(chptr->mode.link, (aChannel *)NULL);

			if (linkptr)
			{
				if (!strcmp(chptr->chname, linkptr->mode.link))
				{
					sendto_one(sptr, ":%s NOTICE %s :*** Cannot join channel %s - looped-link!", me.name, cptr->name,
						chptr->chname);
					return (ERR_CHANNELISFULL);
				}
			}

			parv[0] = sptr->name;
			parv[1] = (chptr->mode.link);
			channel_link(cptr, sptr, 2, parv);
		}
		return (ERR_CHANNELISFULL);
	}

	if (is_forwarded(sptr, chptr) && !lp)
	{
		if (IsOper(sptr))
		{
			sendto_umode(UMODE_OPER, "*** OperWalk -- [Type: forward] - [IRCop: %s] - [Channel: %s]", sptr->name, chptr->chname);
			return 0;
		}

		af = is_forwarded(sptr, chptr);
		afc = find_channel((af->channel), NullChn);

		sendto_one(sptr, ":%s NOTICE %s :*** You are banned from channel %s, so you are being automatically forwarded"
			" to channel %s", me.name, sptr->name, chptr->chname, af->channel);

		if (afc)
		{
			if (is_forwarded(sptr, afc))
			{
				sendto_one(sptr, ":%s NOTICE %s :*** Cannot join channel %s - looped-forward!", me.name, sptr->name,
					afc->chname);
				return (ERR_BANNEDFROMCHAN);
			}
		}

		parv[0] = sptr->name;
		parv[1] = (af->channel);
		channel_link(cptr, sptr, 2, parv);

		return (ERR_BANNEDFROMCHAN);
	}

	if (is_banned(sptr, chptr) && is_banexception(sptr, chptr) && !lp)
	{
		return 0;
	}
	else if (is_banned(sptr, chptr) && !is_banexception(sptr, chptr) && !lp)
	{
		if (IsOper(sptr) && !(chptr->mode.mode & MODE_OPERBAN))
		{
			sendto_umode(UMODE_OPER, "*** OperWalk -- [Type: ban] - [IRCop: %s] - [Channel: %s]", sptr->name, chptr->chname);
			return 0;
		}
		return (ERR_BANNEDFROMCHAN);
	}

	if ((chptr->mode.mode & MODE_RGSTRONLY) && !IsARegNick(sptr) && !IsOper(sptr) && !lp)
	{
		return (ERR_NEEDREGGEDNICK);
	}

	if ((chptr->mode.mode & MODE_NETADMINSONLY) && (IsNetAdmin(sptr) || IsCoNetAdmin(sptr) || IsTechAdmin(sptr)))
	{
		goto joinok;
	}
	else if ((chptr->mode.mode & MODE_NETADMINSONLY) && (!IsNetAdmin(sptr) && !IsCoNetAdmin(sptr) && !IsTechAdmin(sptr)))
	{
		return (ERR_NETADMINSONLY);
	}

	if ((chptr->mode.mode & MODE_ADMINSONLY) && !(IsNetAdmin(sptr) || IsCoNetAdmin(sptr) || IsTechAdmin(sptr) || IsAdmin(sptr)
		|| IsSCoAdmin(sptr)))
	{
		goto joinok;
	}
	else if ((chptr->mode.mode & MODE_ADMINSONLY) && !(IsNetAdmin(sptr) || IsCoNetAdmin(sptr) || IsTechAdmin(sptr) || IsAdmin(sptr)
		|| IsSCoAdmin(sptr)))
	{
		return (ERR_ADMINSONLY);
	}

	if ((chptr->mode.mode & MODE_IRCOPSONLY) && IsOper(sptr))
	{
		goto joinok;
	}
	else if ((chptr->mode.mode & MODE_IRCOPSONLY) && !IsOper(sptr))
	{
		return (ERR_IRCOPSONLY);
	}

		joinok:
	if (is_banned(sptr, chptr) && (chptr->mode.mode & MODE_OPERBAN) && !IsNetAdmin(sptr) && !IsCoNetAdmin(sptr) && !lp)
	{
		return (ERR_BANNEDFROMCHAN);
	}
	return 0;
}

/*
** Remove bells and commas from channel name
*/
void clean_channelname(cn)
    char *cn;
{
    Reg1 u_char *ch = (u_char *) cn;


    for (; *ch; ch++)
	/* Don't allow any control chars, the space, the comma,
	 * or the "non-breaking space" in channel names.
	 * Might later be changed to a system where the list of
	 * allowed/non-allowed chars for channels was a define
	 * or some such.
	 *   --Wizzu
	 */
	if (*ch < 33 || *ch == ',' || *ch == 160) {
	    *ch = '\0';
	    return;
	}
}

/*
** Return -1 if mask is present and doesnt match our server name.
*/
static int check_channelmask(sptr, cptr, chname)
    aClient *sptr, *cptr;
    char *chname;
{
    Reg1 char *s;

    if (*chname == '&' && IsServer(cptr))
	return -1;
    s = rindex(chname, ':');
    if (!s)
	return 0;

    s++;
    if (match(s, me.name) || (IsServer(cptr) && match(s, cptr->name))) {
	if (MyClient(sptr))
	    sendto_one(sptr, err_str(ERR_BADCHANMASK),
		       me.name, sptr->name, chname);
	return -1;
    }
    return 0;
}

/*
**  Get Channel block for i (and allocate a new channel
**  block, if it didn't exists before).
*/
static aChannel *get_channel(cptr, chname, flag)
    aClient *cptr;
    char *chname;
    int flag;
{
    Reg1 aChannel *chptr;
    int len;

    if (BadPtr(chname))
	return NULL;

    len = strlen(chname);
    if (MyClient(cptr) && len > CHANNELLEN) {
	len = CHANNELLEN;
	*(chname + CHANNELLEN) = '\0';
    }
    if ((chptr = find_channel(chname, (aChannel *) NULL)))
	return (chptr);
    if (flag == CREATE) {
	chptr = (aChannel *) MyMalloc(sizeof(aChannel) + len);
	bzero((char *) chptr, sizeof(aChannel));
	strncpyzt(chptr->chname, chname, len + 1);
	if (channel)
	    channel->prevch = chptr;
	chptr->prevch = NULL;
	chptr->nextch = channel;
	chptr->creationtime = MyClient(cptr) ? TStime() : (time_t) 0;
	channel = chptr;
	(void) add_to_channel_hash_table(chname, chptr);
	IRCstats.channels++;
    }
    return chptr;
}

/*
 * Slight changes in routine, now working somewhat symmetrical:
 *   First try to remove the client & channel pair to avoid duplicates
 *   Second check client & channel invite-list lengths and remove tail
 *   Finally add new invite-links to both client and channel
 * Should U-lined clients have higher limits?   -Donwulff
 */

static void add_invite(cptr, chptr)
    aClient *cptr;
    aChannel *chptr;
{
    Reg1 Link *inv, *tmp;

    del_invite(cptr, chptr);
    /*
     * delete last link in chain if the list is max length
     */
    if (list_length(cptr->user->invited) >= MAXCHANNELSPERUSER) {
/*      This forgets the channel side of invitation     -Vesa
   inv = cptr->user->invited;
   cptr->user->invited = inv->next;
   free_link(inv);
 */
	for (tmp = cptr->user->invited; tmp->next; tmp = tmp->next);
	del_invite(cptr, tmp->value.chptr);

    }
    /* We get pissy over too many invites per channel as well now,
     * since otherwise mass-inviters could take up some major
     * resources -Donwulff
     */
    if (list_length(chptr->invites) >= MAXCHANNELSPERUSER) {
	for (tmp = chptr->invites; tmp->next; tmp = tmp->next);
	del_invite(tmp->value.cptr, chptr);
    }
    /*
     * add client to the beginning of the channel invite list
     */
    inv = make_link();
    inv->value.cptr = cptr;
    inv->next = chptr->invites;
    chptr->invites = inv;
    /*
     * add channel to the beginning of the client invite list
     */
    inv = make_link();
    inv->value.chptr = chptr;
    inv->next = cptr->user->invited;
    cptr->user->invited = inv;
}

/*
 * Delete Invite block from channel invite list and client invite list
 */
void del_invite(cptr, chptr)
    aClient *cptr;
    aChannel *chptr;
{
    Reg1 Link **inv, *tmp;

    for (inv = &(chptr->invites); (tmp = *inv); inv = &tmp->next)
	if (tmp->value.cptr == cptr) {
	    *inv = tmp->next;
	    free_link(tmp);
	    break;
	}
    for (inv = &(cptr->user->invited); (tmp = *inv); inv = &tmp->next)
	if (tmp->value.chptr == chptr) {
	    *inv = tmp->next;
	    free_link(tmp);
	    break;
	}
}

/*
   **  Subtract one user from channel i (and free channel
   **  block, if channel became empty).
 */
static void sub1_from_channel(chptr)
    Reg1 aChannel *chptr;
{
    Ban *ban;
    Link *lp;

    if (--chptr->users <= 0) {
	/*
	 * Now, find all invite links from channel structure
	 */
	while ((lp = chptr->invites))
	    del_invite(lp->value.cptr, chptr);

	while (chptr->banlist) {
	    ban = chptr->banlist;
	    chptr->banlist = ban->next;
	    MyFree(ban->banstr);
	    MyFree(ban->who);
	    free_ban(ban);
	}
	if (chptr->prevch)
	    chptr->prevch->nextch = chptr->nextch;
	else
	    channel = chptr->nextch;
	if (chptr->nextch)
	    chptr->nextch->prevch = chptr->prevch;
	(void) del_from_channel_hash_table(chptr->chname, chptr);
	IRCstats.channels--;
	MyFree((char *) chptr);
    }
}

/*
 * send_user_joins - synchs links when !SupportSJOIN
 */
void	send_user_joins(cptr, user)
	aClient		*cptr, *user;
{
	Link		*lp;
	aChannel	*chptr;
	int		cnt = 0, len = 0, clen;
	char		*mask;

	(void)ircsprintf(buf, ":%s %s ", user->name, (IsToken(cptr) ? TOK_JOIN : MSG_JOIN));
	len = strlen(buf);

	for (lp = user->user->channel; lp; lp = lp->next)
	{
		chptr = lp->value.chptr;
		if ((mask = index(chptr->chname, ':')))
		{
			if (match(++mask, cptr->name))
			{
				continue;
			}
		}
		if (*chptr->chname == '&')
		{
			continue;
		}
		clen = strlen(chptr->chname);

		if (clen + 1 + len > BUFSIZE - 3)
		{
			if (cnt)
			{
				buf[len - 1] = '\0';
				sendto_one(cptr, "%s", buf);
			}
			(void)ircsprintf(buf, ":%s %s ", user->name, (IsToken(cptr) ? TOK_JOIN : MSG_JOIN));
			len = strlen(buf);
			cnt = 0;
		}

		(void)strcpy(buf + len, chptr->chname);
		cnt++;

		if (lp->next)
		{
			len++;
			(void)strcat(buf, ",");
		}
	}

	if (*buf && cnt)
	{
		sendto_one(cptr, "%s", buf);
	}

	return;
}

/*
 * send_channel_modes - synchs channel modes when !SupportSJOIN
 */
void	send_channel_modes(cptr, chptr)
	aClient			*cptr;
	aChannel		*chptr;
{
	if (*chptr->chname != '#')
		return;

	*modebuf = *parabuf = '\0';
	channel_modes(cptr, modebuf, parabuf, chptr);

	if (chptr->creationtime)
	{
		sendto_one(cptr, ":%s %s %s %s %s %lu", me.name, (IsToken(cptr) ? TOK_MODE : MSG_MODE), chptr->chname, modebuf,
			parabuf, chptr->creationtime);
	}

	*modebuf = '+';
	*parabuf = '\0';
	modebuf[1] = '\0';
	(void)send_mode_list(cptr, chptr->chname, chptr->creationtime, chptr->members, CHFL_CHANADMIN, 'a');
	if (modebuf[1] || *parabuf)
	{
		sendmodeto_one(cptr, me.name, chptr->chname, modebuf, parabuf, chptr->creationtime);
	}

	*modebuf = '+';
	*parabuf = '\0';
	modebuf[1] = '\0';
	(void)send_mode_list(cptr, chptr->chname, chptr->creationtime, chptr->members, CHFL_CHANOP, 'o');
	if (modebuf[1] || *parabuf)
	{
		sendmodeto_one(cptr, me.name, chptr->chname, modebuf, parabuf, chptr->creationtime);
	}

	*modebuf = '+';
	*parabuf = '\0';
	modebuf[1] = '\0';
	(void)send_mode_list(cptr, chptr->chname, chptr->creationtime, chptr->members, CHFL_HALFOP, 'h');
	if (modebuf[1] || *parabuf)
	{
		sendmodeto_one(cptr, me.name, chptr->chname, modebuf, parabuf, chptr->creationtime);
	}

	*modebuf = '+';
	*parabuf = '\0';
	modebuf[1] = '\0';
	(void)send_mode_list(cptr, chptr->chname, chptr->creationtime, chptr->members, CHFL_VOICE, 'v');
	if (modebuf[1] || *parabuf)
	{
		sendmodeto_one(cptr, me.name, chptr->chname, modebuf, parabuf, chptr->creationtime);
	}

	*modebuf = '+';
	*parabuf = '\0';
	modebuf[1] = '\0';
	(void)send_ban_list(cptr, chptr->chname, chptr->creationtime, chptr);
	if (modebuf[1] || *parabuf)
	{
		sendmodeto_one(cptr, me.name, chptr->chname, modebuf, parabuf, chptr->creationtime);
	}
}

/*
 * channel_link
 */
int	channel_link(cptr, sptr, parc, parv)
	aClient		*cptr, *sptr;
	int		parc;
	char		*parv[];
{
	static	char	jbuf[BUFSIZE];
	Link		*lp;
	aChannel	*chptr;
	char		*name, *key = NULL;
	int		i, flags = 0, zombie = 0;
	char		*p = NULL, *p2 = NULL;

	if ((parc < 2) || (*parv[1] == '\0'))
	{
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "JOIN");
		return 0;
	}
	*jbuf = '\0';

	bouncedtimes++;
	if (bouncedtimes > MAXBOUNCE)
	{
		sendto_one(sptr, ":%s NOTICE %s :*** Cannot join %s - link too unstable", me.name, sptr->name, parv[1]);
		return;
	}

        for (i = 0, name = strtoken(&p, parv[1], ","); name; name = strtoken(&p, NULL, ","))
        {
                if (strlen(name) > CHANNELLEN)
                        name[CHANNELLEN] = '\0';
                if (MyConnect(sptr)) 
                        clean_channelname(name);
                if (check_channelmask(sptr, cptr, name) == -1)
                        continue;
                if ((*name == '&') && !MyConnect(sptr))
                        continue;
                if ((*name == '0') && !atoi(name))
                {
                        (void)strcpy(jbuf, "0"); 
                        i = 1;
                        continue;
                }
                else if (!IsChannelName(name))
                {
                        if (MyClient(sptr))
                        {
                                sendto_one(sptr, err_str(ERR_NOSUCHCHANNEL), me.name, parv[0], name);
                        }
                        continue;
                }
                
                if (*jbuf)
                {
                        (void)strcat(jbuf, ",");
                }
                (void)strncat(jbuf, name, (sizeof(jbuf) - i - 1));
                i += (strlen(name + 1));
        }
        (void)strcpy(parv[1], jbuf);

        p = NULL;
        if (parv[2])
        {
                key = strtoken(&p2, parv[2], ",");
        }
        parv[2] = NULL;
                        
        for (name = strtoken(&p, jbuf, ","); name; key = ((key) ? strtoken(&p2, NULL, ",") : NULL), name = strtoken(&p, NULL, ","))
        {
                if ((*name == '0') && !atoi(name))
                {
                        while ((lp = sptr->user->channel))
                        {
                                chptr = lp->value.chptr;
                                sendto_channel_butserv(chptr, sptr, PartFmt, parv[0], chptr->chname);
                        }
                        sendto_serv_butone(cptr, ":%s JOIN 0", parv[0]);
                        continue;
                }
         
                if (MyConnect(sptr))
                {
                        if (!IsModelessChannel(name))
                        {
                                flags = (ChannelExists(name) ? CHFL_DEOPPED : (CHFL_CHANADMIN | CHFL_CHANOP));
                        }
                        else
                        {
                                flags = CHFL_DEOPPED;
                        }
         
                        if (sptr->user->joined >= MAXCHANNELSPERUSER)
                        {
                                sendto_one(sptr, err_str(ERR_TOOMANYCHANNELS), me.name, parv[0], name);
                                return 0;
                        }

                        if (DYNAMIC_CHANNELS == FALSE)
                        {
                                if (!ChannelExists(name) && !IsOper(sptr))
                                {   
                                        sendto_one(sptr, err_str(ERR_NODYNCHANS), me.name, parv[0], parv[1], ircnet_channel);
                                        return 0;
                                }
                        }
		}

                chptr = get_channel(sptr, name, CREATE);
                if (chptr && (lp = find_user_link(chptr->members, sptr)))
                {
                        if (lp->flags & CHFL_ZOMBIE) 
                        {
                                zombie = 1;
                                flags = (lp->flags & (CHFL_DEOPPED | CHFL_SERVOPOK));
                                remove_user_from_channel(sptr, chptr);
                                chptr = get_channel(sptr, name, CREATE);
                        }
                        else
                        {
                                continue;
                        }
		}

                if (!zombie)
                {
                        if (!MyConnect(sptr))   
                                flags = CHFL_DEOPPED;
                        if (sptr->flags & FLAGS_TS8)   
                                flags |= CHFL_SERVOPOK;
                }
                        
                if (!chptr || (MyConnect(sptr) && (i = can_join(cptr, sptr, chptr, key, parv))))
                {
                        sendto_one(sptr, err_str(i), me.name, parv[0], name);
       	                continue;
                }

                add_user_to_channel(chptr, sptr, flags);
                sendto_channel_butserv(chptr, sptr, ":%s JOIN :%s", parv[0], name);
                sendto_match_servs(chptr, cptr, ":%s JOIN :%s", parv[0], name);
                        
                if (MyClient(sptr))
                {
                        if (chptr->creationtime == 0) 
                        {
                                chptr->creationtime = TStime();
                                sendto_match_servs(chptr, cptr, ":%s MODE %s + %lu", me.name, name, chptr->creationtime);
                        }
                        del_invite(sptr, chptr);

                        if (flags & CHFL_CHANOP) {
                                sendto_match_servs(chptr, cptr, ":%s MODE %s +o %s %lu", me.name, name, parv[0], chptr->creationtime);
                        }
                                
                        if (chptr->topic[0] |= '\0')
                        {
                                sendto_one(sptr, rpl_str(RPL_TOPIC), me.name, parv[0], name, chptr->topic);
                                sendto_one(sptr, rpl_str(RPL_TOPICWHOTIME), me.name, parv[0], name, chptr->topic_nick, chptr->topic_time);
                        }
                         
                        parv[1] = name;
                        cptr->flags |= FLAGS_JOINING;
                        (void)m_names(cptr, sptr, 2, parv);
                        cptr->flags &= ~FLAGS_JOINING;
		}
        }
	bouncedtimes = 0;
        return 0;
}

/*
 * m_join
 *   parv[0] = sender prefix
 *   parv[1] = channel
 *   parv[2] = channel password (key)
 *
 * revamped - TD (December 2000)
 */
int	m_join(cptr, sptr, parc, parv)
	aClient		*cptr, *sptr;
	int		parc;
	char		*parv[];
{
	static	char	jbuf[BUFSIZE];
	Link		*lp;
	aChannel	*chptr;
	char		*name, *key = NULL;
	int		i, flags = 0, zombie = 0;
	char		*p = NULL, *p2 = NULL;

	bouncedtimes = 0;

	if ((parc < 2) || (*parv[1] == '\0'))
	{
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "JOIN");
		return 0;
	}
	*jbuf = '\0';

	for (i = 0, name = strtoken(&p, parv[1], ","); name; name = strtoken(&p, NULL, ","))
	{
		if (strlen(name) > CHANNELLEN)
			name[CHANNELLEN] = '\0';
		if (MyConnect(sptr))
			clean_channelname(name);
		if (check_channelmask(sptr, cptr, name) == -1)
			continue;
		if ((*name == '&') && !MyConnect(sptr))
			continue;
		if ((*name == '+'))
			continue;
		if ((*name == '0') && !atoi(name))
		{
			(void)strcpy(jbuf, "0");
			i = 1;
			continue;
		}
		else if (!IsChannelName(name))
		{
			if (MyClient(sptr))
			{
				sendto_one(sptr, err_str(ERR_NOSUCHCHANNEL), me.name, parv[0], name);
			}
			continue;
		}

		if (*jbuf)
		{
			(void)strcat(jbuf, ",");
		}
		(void)strncat(jbuf, name, (sizeof(jbuf) - i - 1));
		i += (strlen(name + 1));
	}
	(void)strcpy(parv[1], jbuf);

	p = NULL;
	if (parv[2])
	{
		key = strtoken(&p2, parv[2], ",");
	}
	parv[2] = NULL;

	for (name = strtoken(&p, jbuf, ","); name; key = ((key) ? strtoken(&p2, NULL, ",") : NULL), name = strtoken(&p, NULL, ","))
	{
		if ((*name == '0') && !atoi(name))
		{
			while ((lp = sptr->user->channel))
			{
				chptr = lp->value.chptr;
				sendto_channel_butserv(chptr, sptr, PartFmt, parv[0], chptr->chname);
			}
			sendto_serv_butone(cptr, ":%s JOIN 0", parv[0]);
			continue;
		}

		if (MyConnect(sptr))
		{
			if (!IsModelessChannel(name))
			{
				flags = (ChannelExists(name) ? CHFL_DEOPPED : (CHFL_CHANADMIN | CHFL_CHANOP));
			}
			else
			{
				flags = CHFL_DEOPPED;
			}

			if (sptr->user->joined >= MAXCHANNELSPERUSER)
			{
				sendto_one(sptr, err_str(ERR_TOOMANYCHANNELS), me.name, parv[0], name);
				return 0;
			}

			if (DYNAMIC_CHANNELS == FALSE)
			{
				if (!ChannelExists(name) && !IsOper(sptr))
				{
					sendto_one(sptr, err_str(ERR_NODYNCHANS), me.name, parv[0], parv[1], ircnet_channel);
					return 0;
				}
			}
			if ((check_restrictions(name) == 1) && !IsOper(sptr) && !IsULine(sptr, sptr))
			{
				sendto_one(sptr, ":%s NOTICE %s :*** Unable to join %s - channel is restricted", me.name,
					parv[0], name);
				return 0;
			}
		}

		chptr = get_channel(sptr, name, CREATE);
		if (chptr && (lp = find_user_link(chptr->members, sptr)))
		{
			if (lp->flags & CHFL_ZOMBIE)
			{
				zombie = 1;
				flags = (lp->flags & (CHFL_DEOPPED | CHFL_SERVOPOK));
				remove_user_from_channel(sptr, chptr);
				chptr = get_channel(sptr, name, CREATE);
			}
			else
			{
				continue;
			}
		}

		if (!zombie)
		{
			if (!MyConnect(sptr))
				flags = CHFL_DEOPPED;
			if (sptr->flags & FLAGS_TS8)
				flags |= CHFL_SERVOPOK;
		}

		if (!chptr || (MyConnect(sptr) && (i = can_join(cptr, sptr, chptr, key, parv))))
		{
			sendto_one(sptr, err_str(i), me.name, parv[0], name);
			continue;
		}

		add_user_to_channel(chptr, sptr, flags);
		sendto_channel_butserv(chptr, sptr, ":%s JOIN :%s", parv[0], name);
		sendto_match_servs(chptr, cptr, ":%s JOIN :%s", parv[0], name);

		if (MyClient(sptr))
		{
			if (chptr->creationtime == 0)
			{
				chptr->creationtime = TStime();
				sendto_match_servs(chptr, cptr, ":%s MODE %s + %lu", me.name, name, chptr->creationtime);
			}
			del_invite(sptr, chptr);

			if (flags & CHFL_CHANOP) {
				sendto_match_servs(chptr, cptr, ":%s MODE %s +o %s %lu", me.name, name, parv[0], chptr->creationtime);
			}

			if (chptr->topic[0] |= '\0')
			{
				sendto_one(sptr, rpl_str(RPL_TOPIC), me.name, parv[0], name, chptr->topic);
				sendto_one(sptr, rpl_str(RPL_TOPICWHOTIME), me.name, parv[0], name, chptr->topic_nick, chptr->topic_time);
			}

			parv[1] = name;
			cptr->flags |= FLAGS_JOINING;
			(void)m_names(cptr, sptr, 2, parv);
			cptr->flags &= ~FLAGS_JOINING;
		}
	}
	bouncedtimes = 0;
	return 0;
}

/*
 * m_part
 *
 *	parv[0] = sender prefix
 *	parv[1] = channel
 *	if (parc > 2)
 *		parv[2] = comment
 */
int	m_part(cptr, sptr, parc, parv)
	aClient		*cptr, *sptr;
	int		parc;
	char		*parv[];
{
	aChannel	*chptr;
	Link		*lp;
	char		*p = NULL, *name, *comment = NULL;

	sptr->flags &= ~FLAGS_TS8;

	if (parc < 2 || parv[1][0] == '\0')
	{
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "PART");
		return 0;
	}

	for (; (name = strtoken(&p, parv[1], ",")); parv[1] = NULL)
	{
		chptr = get_channel(sptr, name, 0);
		if (!chptr)
		{
			sendto_one(sptr, err_str(ERR_NOSUCHCHANNEL), me.name, parv[0], name);
			continue;
		}
		if (check_channelmask(sptr, cptr, name))
		{
			continue;
		}

		lp = find_user_link(chptr->members, sptr);
		if (!lp)
		{
			if (MyClient(sptr))
			{
				sendto_one(sptr, err_str(ERR_NOTONCHANNEL), me.name, parv[0], name);
			}
			continue;
		}

		comment = parc < 3 ? NULL : parv[2];

		if (comment == NULL)
		{
			sendto_match_servs(chptr, cptr, PartFmt, parv[0], name);
		}
		else
		{
			sendto_match_servs(chptr, cptr, PartFmt2, parv[0], name, comment);
		}

		if (!(lp->flags & CHFL_ZOMBIE))
		{
			if (comment == NULL)
			{
				sendto_channel_butserv(chptr, sptr, PartFmt, parv[0], name);
			}
			else
			{
				sendto_channel_butserv(chptr, sptr, PartFmt2, parv[0], name, comment);
			}
		}
		else if (MyClient(sptr))
		{
			if (comment == NULL)
			{
				sendto_one(sptr, PartFmt, parv[0], name);
			}
			else
			{
				sendto_one(sptr, PartFmt2, parv[0], name, comment);
			}
		}
		remove_user_from_channel(sptr, chptr);
	}
	return 0;
}

/*
 * m_kick
 *	parv[0] = sender prefix
 *	parv[1] = channel
 *	parv[2] = client to kick
 *	parv[3] = kick comment
 */
int	m_kick(cptr, sptr, parc, parv)
	aClient		*cptr, *sptr;
	int		parc;
	char		*parv[];
{
	aClient		*who;
	aChannel	*chptr;
	int		chasing = 0;
	char		*comment, *name, *p = NULL, *user, *p2 = NULL;
	Link		*lp, *lp2;

	if (check_registered(sptr))
		return 0;

	sptr->flags &= ~FLAGS_TS8;

	if (parc < 3 || *parv[1] == '\0')
	{
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS),
			   me.name, parv[0], "KICK");
		return 0;
	}

	comment = (BadPtr(parv[3]) ? "No reason specified" : parv[3]);

	if (strlen(comment) > (size_t) TOPICLEN)
		comment[TOPICLEN] = '\0';

	*nickbuf = *buf = '\0';

	for (; (name = strtoken(&p, parv[1], ",")); parv[1] = NULL)
	{
		chptr = get_channel(sptr, name, !CREATE);
		if (!chptr)
		{
			sendto_one(sptr, err_str(ERR_NOSUCHCHANNEL), me.name, parv[0], name);
			continue;
		}
		if (check_channelmask(sptr, cptr, name))
			continue;
		if (!IsServer(cptr) && !IsOper(sptr) && !IsULine(cptr,sptr) && !is_chan_op(sptr, chptr)
			&& !is_half_op(sptr, chptr))
		{
			sendto_one(sptr, err_str(ERR_CHANOPRIVSNEEDED), me.name, parv[0], chptr->chname);
			continue;
		}
		if ((chptr->mode.mode & MODE_PEACE) && !IsULine(sptr, sptr) && !IsOper(sptr) && !IsServer(sptr))
		{
			sendto_one(sptr, ":%s NOTICE %s :*** Cannot kick %s - channel is +j (peace mode)", me.name,
				parv[0], parv[2]);
			continue;
		}

	        lp2 = find_user_link(chptr->members, sptr);

		for (; (user = strtoken(&p2, parv[2], ",")); parv[2] = NULL)
		{
			if (!(who = find_chasing(sptr, user, &chasing)))
				continue; /* No such user left! */

			if ((lp = find_user_link(chptr->members, who)) && !(lp->flags & CHFL_ZOMBIE))
	                {
				if (is_chanadmin(who, chptr) || IsServices(who))
				{
					if (IsNetAdmin(sptr) || IsCoNetAdmin(sptr) || IsTechAdmin(sptr))
					{
						if (!is_chan_op(sptr, chptr) && !is_half_op(sptr, chptr))
						{
							sendto_umode(UMODE_OPER, "*** OperKick -- [IRCop: %s] - [Channel: %s] - [Kicked: %s] - [Reason: %s]",
								parv[0], parv[1], parv[2], comment);
						}
						goto attack;
					}
					else
					{
						sendto_one(sptr, ":%s NOTICE %s :*** Cannot kick %s (Channel Administrator)",
							me.name, sptr->name, who->name);
						continue;
					}
				}
				if (IsOper(sptr))
				{
					if (!is_chan_op(sptr, chptr) && !is_half_op(sptr, chptr) && !IsULine(cptr, sptr))
					{

						sendto_umode(UMODE_OPER, "*** OperKick -- [IRCop: %s] - [Channel: %s] -"
							" [Kicked: %s] - [Reason: %s]", parv[0], parv[1], parv[2], comment);
						goto attack;
					}
				}
				if (is_half_op(sptr, chptr))
				{
					if (IsOper(sptr))
					{
						sendto_umode(UMODE_OPER, "*** OperKick -- [IRCop: %s] - [Channel: %s] -"
							" [Kicked: %s] - [Reason: %s]", parv[0], parv[1], parv[2], comment);
						goto attack;
					}
					else
					{
						if (is_chan_op(who, chptr) && !is_chan_op(sptr, chptr))
						{
							sendto_one(sptr, ":%s NOTICE %s :*** Cannot kick user %s"
								" - you are only a halfop.", me.name, sptr->name,
								who->name);
							continue;
						}
						else
						{
							goto attack;
						}
					}
				}
			}
			else if (who != sptr)
			{
				if (IsULine(cptr,sptr) || IsServices(sptr) || IsNetAdmin(sptr) || IsCoNetAdmin(sptr) || IsTechAdmin(sptr))
					goto attack;
			}

				attack:
			if (lp)
			{
				sendto_channel_butserv(chptr, sptr, ":%s KICK %s %s :%s",
					parv[0], name, who->name, comment);
				sendto_match_servs(chptr, sptr, ":%s KICK %s %s :%s",
					parv[0], name, who->name, comment);
					/* zombies, away with you! */
				if (lp)
				{
					lp->flags |= CHFL_ZOMBIE;
					remove_user_from_channel(who, chptr);
				}
			}
			else if (MyClient(sptr))
			{
				sendto_one(sptr, err_str(ERR_USERNOTINCHANNEL), me.name, parv[0],
					user, name);
			}
			if (!IsServer(cptr) || !IsULine(cptr, sptr))
				break;
		}
	if (!IsServer(cptr) || !IsULine(cptr, sptr))
		break;
	}
	return 0;
}

int count_channels(sptr)
    aClient *sptr;
{
    Reg1 aChannel *chptr;
    Reg2 int count = 0;

    for (chptr = channel; chptr; chptr = chptr->nextch)
	count++;
    return (count);
}

/*
 * m_topic
 *
 *	parv[0] = sender prefix
 *	parv[1] = topic text
 *	For servers using TS (Lefler):
 *		parv[0] = sender prefix
 *		parv[1] = channel list
 *		parv[2] = topic nickname
 *		parv[3] = topic time
 *		parv[4] = topic text
 */

int	m_topic(cptr, sptr, parc, parv)
	aClient		*cptr, *sptr;
	int		parc;
	char		*parv[];
{
	aChannel	*chptr = NullChn;
	char		*topic = NULL, *name, *p = NULL, *tnick = NULL;
	time_t		ttime = 0;

	if (parc < 2)
	{
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "TOPIC");
		return 0;
	}

	for (; (name = strtoken(&p, parv[1], ",")); parv[1] = NULL)
	{
		if (parc > 1 && IsChannelName(name))
		{
			chptr = find_channel(name, NullChn);
			if (!chptr)
			{
				sendto_one(sptr, err_str(ERR_NOSUCHCHANNEL), me.name, parv[0], name);
				continue;
			}

			if (parc > 2)
			{
				topic = parv[2];
			}
			if (parc > 4)
			{
				tnick = parv[2];
				ttime = atol(parv[3]);
				topic = parv[4];
			}
		}

		if (!chptr) {
			sendto_one(sptr, rpl_str(RPL_NOTOPIC), me.name, parv[0], name);
			continue;
		}

		if (check_channelmask(sptr, cptr, name))
		{
			continue;
		}

		if (!topic)
		{
			if (IsMember(sptr, chptr))
			{
				if (chptr->topic[0] == '\0')
				{
					sendto_one(sptr, rpl_str(RPL_NOTOPIC), me.name, parv[0], chptr->chname);
				} else {
					sendto_one(sptr, rpl_str(RPL_TOPIC), me.name, parv[0], chptr->chname, chptr->topic);
					sendto_one(sptr, rpl_str(RPL_TOPICWHOTIME), me.name, parv[0], chptr->chname,
						chptr->topic_nick, chptr->topic_time);
				}
			} else {
				if (!SecretChannel(chptr))
				{
					sendto_one(sptr, rpl_str(RPL_TOPIC), me.name, parv[0], chptr->chname, chptr->topic);
				} else {
					sendto_one(sptr, err_str(ERR_NOTONCHANNEL), me.name, parv[0], name);
				}
			}
		}
		else if (ttime && topic && (IsServer(sptr) || IsULine(cptr, sptr)))
		{
			if (!chptr->topic_time || ttime < chptr->topic_time)
			{
				strncpyzt(chptr->topic, topic, sizeof(chptr->topic));
				strcpy(chptr->topic_nick, tnick);
				chptr->topic_time = ttime;

				sendto_match_servs(chptr, cptr, ":%s TOPIC %s %s %lu :%s", parv[0], chptr->chname,
					chptr->topic_nick, chptr->topic_time, chptr->topic);
				sendto_channel_butserv(chptr, sptr, ":%s TOPIC %s :%s (%s)", parv[0], chptr->chname,
					chptr->topic, chptr->topic_nick);
			}
		}
		else if ((is_chan_op(sptr, chptr) || is_half_op(sptr, chptr) || IsOper(sptr) || (IsULine(cptr, sptr) && topic))
			|| ((chptr->mode.mode & MODE_TOPICLIMIT) == 0 && IsMember(sptr, chptr)))
		{

			if (IsOper(sptr) && !is_chan_op(sptr, chptr) && !IsULine(cptr, sptr) && (chptr->mode.mode & MODE_TOPICLIMIT))
			{
				sendto_umode(UMODE_OPER, "*** OperTopic -- [IRCop: %s] - [Channel: %s] - [New Topic: %s]",
					sptr->name, chptr->chname, topic);
			}

			strncpyzt(chptr->topic, topic, sizeof(chptr->topic));
			strcpy(chptr->topic_nick, sptr->name);

			if (ttime && IsServer(cptr))
			{
				chptr->topic_time = ttime;
			} else {
				chptr->topic_time = TStime();
			}

			sendto_match_servs(chptr, cptr, ":%s TOPIC %s %s %lu :%s", parv[0], chptr->chname, parv[0],
				chptr->topic_time, chptr->topic);
			sendto_channel_butserv(chptr, sptr, ":%s TOPIC %s :%s", parv[0], chptr->chname, chptr->topic);
		}
		else if (!IsMember(sptr, chptr))
		{
			sendto_one(sptr, err_str(ERR_NOTONCHANNEL), me.name, parv[0], name);
		} else {
			sendto_one(sptr, err_str(ERR_CHANOPRIVSNEEDED), me.name, parv[0], name);
		}
	}
	return 0;
}

/*
 * m_invite
 *	parv[0] - sender prefix
 *	parv[1] - user to invite
 *	parv[2] - channel number
 */
int  m_invite(cptr, sptr, parc, parv)
	aClient *cptr, *sptr;
	int  parc;
	char *parv[];
{
	aClient *acptr;
	aChannel *chptr;

	if (parc < 3 || *parv[1] == '\0')
	{
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "INVITE");
		return -1;
	}

	if (!(acptr = find_person(parv[1], (aClient *)NULL)))
	{
		sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, parv[0], parv[1]);
		return 0;
	}

	if (MyConnect(sptr))
		clean_channelname(parv[2]);
	if (check_channelmask(sptr, cptr, parv[2]))
		return 0;
	if (!(chptr = find_channel(parv[2], NullChn)))
	{
		sendto_prefix_one(acptr, sptr, ":%s INVITE %s :%s", parv[0], parv[1], parv[2]);
		return 0;
	}
	if (chptr->mode.mode & MODE_NOINVITES)
		if (!IsULine(cptr, sptr))
		{
			sendto_one(sptr, ":%s NOTICE %s :*** Cannot invite %s to %s - channel is +I (no invites)",
				me.name, parv[0], parv[1], parv[2]);
			return -1;
		}

	if (chptr->mode.mode & MODE_OPERBAN)
		if (IsOper(sptr) && is_banned(acptr, chptr))
		{
			sendto_one(sptr, ":%s NOTICE %s :*** Cannot invite %s to %s - user is banned and channel is +B"
				" (does not allow opers to walk though bans)", me.name, parv[0], parv[0], parv[2]);
			return -1;
		}

	if (chptr && !IsMember(sptr, chptr) && !IsULine(cptr, sptr))
	{
		sendto_one(sptr, err_str(ERR_NOTONCHANNEL), me.name, parv[0], parv[2]);
		return -1;
	}

	if (IsMember(acptr, chptr))
	{
		sendto_one(sptr, err_str(ERR_USERONCHANNEL), me.name, parv[0], parv[1], parv[2]);
		return 0;
	}

	if (chptr && (chptr->mode.mode & MODE_INVITEONLY))
	{
		if (!is_chan_op(sptr, chptr) && !IsULine(cptr, sptr) && !is_half_op(sptr, chptr))
		{
			sendto_one(sptr, err_str(ERR_CHANOPRIVSNEEDED), me.name, parv[0], chptr->chname);
			return -1;
		}
		else if (!IsMember(sptr, chptr) && !IsULine(cptr, sptr))
		{
			sendto_one(sptr, err_str(ERR_CHANOPRIVSNEEDED), me.name, parv[0],
				((chptr) ? (chptr->chname) : parv[2]));
			return -1;
		}
	}

	if (MyConnect(sptr))
	{
		if (check_for_target_limit(sptr, acptr, acptr->name))
			return 0;

		sendto_one(sptr, rpl_str(RPL_INVITING), me.name, parv[0], acptr->name, ((chptr) ? (chptr->chname) : parv[2]));
		if (acptr->user->away)
			sendto_one(sptr, rpl_str(RPL_AWAY), me.name, parv[0], acptr->name, acptr->user->away);
	}
	/* Note: is_banned() here will cause some extra CPU load, 
	 *       and we're really only relying on the existence
	 *       of the limit because we could momentarily have
	 *       less people on channel.
	 */
	if (MyConnect(acptr))
		if (chptr && sptr->user
		    && (is_banned(acptr, chptr)
		    || (chptr->mode.mode & MODE_INVITEONLY)
		    || chptr->mode.limit) && (is_chan_op(sptr, chptr)
		    || IsULine(cptr, sptr)))
		{
			add_invite(acptr, chptr);
			sendto_channelops_butone(NULL, &me, chptr, ":%s NOTICE @%s :%s invited %s into the channel.",
			    me.name, chptr->chname, sptr->name, acptr->name);
		}
	sendto_prefix_one(acptr, sptr, ":%s INVITE %s :%s", parv[0], acptr->name, ((chptr) ? (chptr->chname) : parv[2]));
	return 0;
}

static int number_of_zombies(chptr)
    aChannel *chptr;
{
    return 0;
}

/*
 * send_list
 *
 * The function which sends
 * The function which sends the actual /list output back to the user.
 * Operates by stepping through the hashtable, sending the entries back if
 * they match the criteria.
 * cptr = Local client to send the output back to.
 * numsend = Number (roughly) of lines to send back. Once this number has
 * been exceeded, send_list will finish with the current hash bucket,
 * and record that number as the number to start next time send_list
 * is called for this user. So, this function will almost always send
 * back more lines than specified by numsend (though not by much,
 * assuming CHANNELHASHSIZE is was well picked). So be conservative
 * if altering numsend };> -Rak
 */
void send_list(cptr, numsend)
    aClient *cptr;
    int numsend;
{
    int hashptr, done = 0;
    aChannel *chptr;
/*    Link  *tmpl; Compiler says this is unused */
    char tempbuff[LINKLEN + 1 + KEYLEN + 8 + 3 + 1 + 3], modestuff[LINKLEN + 1 + TOPICLEN + 3 + KEYLEN + 8 + 3 + 1];

#define l cptr->lopt		/* lazy shortcut */

    for (hashptr = l->starthash; hashptr < CHANNELHASHSIZE; hashptr++) {
	for (chptr = hash_get_chan_bucket(hashptr);
	     chptr; chptr = chptr->hnextch) {
	    if (SecretChannel(chptr) && !IsMember(cptr, chptr) && !IsSAdmin(cptr))
		continue;
	    if (!l->showall && ((chptr->users <= l->usermin) ||
	       ((l->usermax == -1) ? 0 : (chptr->users >= l->usermax)) ||
			((chptr->creationtime || 1) <= l->chantimemin) ||
				(chptr->topic_time < l->topictimemin) ||
				(chptr->creationtime >= l->chantimemax) ||
				(chptr->topic_time > l->topictimemax)))
		continue;
	    /* For now, just extend to topics as well. Use patterns starting
	     * with # to stick to searching channel names only. -Donwulff
	     */
	    if (l->nolist &&
		(find_str_match_link(&(l->nolist), chptr->chname) ||
		 find_str_match_link(&(l->nolist), chptr->topic)))
		continue;
	    if (l->yeslist &&
		(!find_str_match_link(&(l->yeslist), chptr->chname) &&
		 !find_str_match_link(&(l->yeslist), chptr->topic)))
		continue;

	    if (IsSAdmin(cptr))
	    {
		*modebuf = *parabuf = '\0';
		modebuf[1] = '\0';
		channel_modes(cptr, modebuf, parabuf, chptr);

		if (*parabuf)
		{
		    (void)ircsprintf(tempbuff, " [%s %s]", modebuf, parabuf);
		}
		else
		{
		    (void)ircsprintf(tempbuff, " [%s]", modebuf);
		}
		(void)ircsprintf(modestuff, "%-2s %s", tempbuff, chptr->topic);

		sendto_one(cptr, rpl_str(RPL_LIST), me.name, cptr->name, chptr->chname, chptr->users - number_of_zombies(chptr),
			modestuff);
	    }
	    else
	    {
		sendto_one(cptr, rpl_str(RPL_LIST), me.name, cptr->name, ShowChannel(cptr, chptr) ? chptr->chname : "*",
			chptr->users - number_of_zombies(chptr), ShowChannel(cptr, chptr) ? chptr->topic : "");
	    }

	    if (--numsend == 0)	/* Send to the end of the list and return */
		done = 1;
	}

	if (done && (++hashptr < CHANNELHASHSIZE)) {
	    l->starthash = hashptr;
	    return;
	}
    }

    sendto_one(cptr, rpl_str(RPL_LISTEND), me.name, cptr->name);
    cptr->flags2 &= ~FLAGS2_HTC;

    free_str_list(l->yeslist);
    free_str_list(l->nolist);
    MyFree(l);
    l = NULL;

    /* List finished, penalize by 10 seconds -Donwulff */
    if (!IsPrivileged(cptr))
	cptr->since += 10;

    return;
}


/*
 * m_list
 *  parv[0] = sender prefix
 *  parv[1,2,3...] = Channels or list options.
 */
int m_list(cptr, sptr, parc, parv)
    aClient *cptr, *sptr;
    int parc;
    char *parv[];
{
    aChannel *chptr;
    char *name, *p = NULL;
    LOpts *lopt;
    short int showall = 0;
    Link *yeslist = NULL, *nolist = NULL, *listptr;
    short usermin = 0, usermax = -1;
    time_t currenttime = TStime();
    time_t chantimemin = 0, topictimemin = 0;
    time_t chantimemax, topictimemax;

    char tempbuff[LINKLEN + 1 + KEYLEN + 8 + 3 + 1 + 3], modestuff[LINKLEN + 1 + TOPICLEN + 3 + KEYLEN + 8 + 3 + 1];

    static char *usage[] =
    {
	"   Usage: /raw LIST options (on mirc) or /quote LIST options (ircII)",
	"",
      "If you don't include any options, the default is to send you the",
     "entire unfiltered list of channels. Below are the options you can",
	"use, and what channels LIST will return when you use them.",
	">number  List channels with more than <number> people.",
	"<number  List channels with less than <number> people.",
	"C>number List channels created between now and <number> minutes ago.",
     "C<number List channels created earlier than <number> minutes ago.",
    "T>number List channels whose topics are older than <number> minutes",
     "         (Ie, they have not changed in the last <number> minutes.",
	"T<number List channels whose topics are not older than <number> minutes.",
	"*mask*   List channels that match *mask*",
	"!*mask*  List channels that do not match *mask*",
	NULL
    };

    cptr->flags2 |= FLAGS2_HTC; /* To prevent a dead socket */

/* Not needed re change to parse.c
   None of that unregistered LIST stuff.  -- Barubary
   if (check_registered(sptr)) return 0;
 */
    /*
     * I'm making the assumption it won't take over a day to transmit
     * the list... -Rak
     */
    chantimemax = topictimemax = currenttime + 86400;


    if ((parc == 2) && (!strcasecmp(parv[1], "?"))) {
	char **ptr = usage;

	for (; *ptr; ptr++)
	    sendto_one(sptr, rpl_str(RPL_LISTSYNTAX), me.name,
		       cptr->name, *ptr);
        
	return 0;
    }
    /*
     * A list is already in process, for now we just interrupt the
     * current listing, perhaps later we can allow stacked ones...
     *  -Donwulff (Not that it's hard or anything, but I don't see
     *             much use for it, beyond flooding)
     */

    if (cptr->lopt) {
	free_str_list(cptr->lopt->yeslist);
	free_str_list(cptr->lopt->nolist);
	MyFree(cptr->lopt);
	cptr->lopt = NULL;
	sendto_one(sptr, rpl_str(RPL_LISTEND), me.name, cptr->name);
        cptr->flags2 &= ~FLAGS2_HTC;

	/* Interrupted list, penalize 10 seconds */
	if (!IsPrivileged(sptr))
	    sptr->since += 10;
    	return 0;
    }
    sendto_one(sptr, rpl_str(RPL_LISTSTART), me.name, cptr->name);

    /* LIST with no arguements */
    if (parc < 2 || BadPtr(parv[1])) {
	lopt = (LOpts *) MyMalloc(sizeof(LOpts));
	if (!lopt)
	    return 0;

	/*
	 * Changed to default to ignoring channels with only
	 * 1 person on, to decrease floods... -Donwulff
	 */
	bzero(lopt, sizeof(LOpts));	/* To be sure! */
	lopt->next = (LOpts *) lopt->yeslist = lopt->nolist = (Link *) NULL;
	lopt->usermin = 0;	/* Default */
	lopt->usermax = -1;
	lopt->chantimemax = lopt->topictimemax = currenttime + 86400;
	cptr->lopt = lopt;
	if (IsSendable(cptr))
	    send_list(cptr, 64);
	return 0;
    }
    /*
     * General idea: We don't need parv[0], since we can get that
     * information from cptr.name. So, let's parse each element of
     * parv[], setting pointer parv to the element being parsed.
     */
    while (--parc) {
	parv += 1;
	if (BadPtr(parv))	/* Sanity check! */
	    continue;

	name = strtoken(&p, *parv, ",");

	while (name) {
	    switch (*name) {
	    case '>':
		showall = 1;
		usermin = strtol(++name, (char **) 0, 10);
		break;

	    case '<':
		showall = 1;
		usermax = strtol(++name, (char **) 0, 10);
		break;

	    case 't':
	    case 'T':
		showall = 1;
		switch (*++name) {
		case '>':
		    topictimemax = currenttime - 60 *
			strtol(++name, (char **) 0, 10);
		    break;

		case '<':
		    topictimemin = currenttime - 60 *
			strtol(++name, (char **) 0, 10);
		    break;

		case '\0':
		    topictimemin = 1;
		    break;

		default:
		    sendto_one(sptr, err_str(ERR_LISTSYNTAX),
			       me.name, cptr->name);
		    free_str_list(yeslist);
		    free_str_list(nolist);
		    sendto_one(sptr, rpl_str(RPL_LISTEND),
			       me.name, cptr->name);
                    cptr->flags2 &= ~FLAGS2_HTC;
                    return 0;
		}
		break;

	    case 'c':
	    case 'C':
		showall = 1;
		switch (*++name) {
		case '>':
		    chantimemin = currenttime - 60 *
			strtol(++name, (char **) 0, 10);
		    break;

		case '<':
		    chantimemax = currenttime - 60 *
			strtol(++name, (char **) 0, 10);
		    break;

		default:
		    sendto_one(sptr, err_str(ERR_LISTSYNTAX),
			       me.name, cptr->name);
		    free_str_list(yeslist);
		    free_str_list(nolist);
		    sendto_one(sptr, rpl_str(RPL_LISTEND),
			       me.name, cptr->name);
                    cptr->flags2 &= ~FLAGS2_HTC;

		    return 0;
		}
		break;

	    default:		/* A channel or channel mask */

		/*
		 * new syntax: !channelmask will tell ircd to ignore
		 * any channels matching that mask, and then
		 * channelmask will tell ircd to send us a list of
		 * channels only masking channelmask. Note: Specifying
		 * a channel without wildcards will return that
		 * channel even if any of the !channelmask masks
		 * matches it.
		 */

		if (*name == '!') {
		    showall = 1;
		    listptr = make_link();
		    listptr->next = nolist;
		    DupString(listptr->value.cp, name + 1);
		    nolist = listptr;
		} else if (strchr(name, '*') || strchr(name, '?')) {
		    showall = 1;
		    listptr = make_link();
		    listptr->next = yeslist;
		    DupString(listptr->value.cp, name);
		    yeslist = listptr;
		} else {
		    chptr = find_channel(name, NullChn);

		    if (IsSAdmin(sptr) && chptr)
		    {
			*modebuf = *parabuf = '\0';
			modebuf[1] = '\0';
			channel_modes(sptr, modebuf, parabuf, chptr);

			if (*parabuf)
			{
			    (void)ircsprintf(tempbuff, " [%s %s]", modebuf, parabuf);
			}
			else
			{
			    (void)ircsprintf(tempbuff, " [%s]", modebuf);
			}
			(void)ircsprintf(modestuff, "%-2s %s", tempbuff, chptr->topic);

			sendto_one(cptr, rpl_str(RPL_LIST), me.name, cptr->name, ShowChannel(sptr, chptr) ? name : "*",
				chptr->users - number_of_zombies(chptr), modestuff);
		    }
		    else
		    {
			if (chptr && ShowChannel(sptr, chptr))
			{
				sendto_one(sptr, rpl_str(RPL_LIST), me.name, cptr->name, ShowChannel(sptr, chptr) ? name : "*",
					chptr->users - number_of_zombies(chptr), chptr->topic);
			}
		    }
		}

	    }			/* switch (*name) */
	    name = strtoken(&p, NULL, ",");
	}			/* while(name) */
    }				/* while(--parc) */

    if (!showall || (chantimemin > currenttime) ||
	(topictimemin > currenttime)) {
	free_str_list(yeslist);
	free_str_list(nolist);
	sendto_one(sptr, rpl_str(RPL_LISTEND), me.name, cptr->name);
        cptr->flags2 &= ~FLAGS2_HTC;

	return 0;
    }
    lopt = (LOpts *) MyMalloc(sizeof(LOpts));

    lopt->showall = 0;
    lopt->next = NULL;
    lopt->yeslist = yeslist;
    lopt->nolist = nolist;
    lopt->starthash = 0;
    lopt->usermin = usermin;
    lopt->usermax = usermax;
    lopt->currenttime = currenttime;
    lopt->chantimemin = chantimemin;
    lopt->chantimemax = chantimemax;
    lopt->topictimemin = topictimemin;
    lopt->topictimemax = topictimemax;

    cptr->lopt = lopt;
    send_list(cptr, 64);

    return 0;
}

/************************************************************************
 * m_names() - Added by Jto 27 Apr 1989
 ************************************************************************/
/*
** m_names
**   parv[0] = sender prefix
**   parv[1] = channel
*/
int m_names(cptr, sptr, parc, parv)
    aClient *cptr, *sptr;
    int parc;
    char *parv[];
{
    Reg1 aChannel *chptr;
    Reg2 aClient *c2ptr;
    Reg3 Link *lp;
    aChannel *ch2ptr = NULL;
    int idx, flag, len, mlen, x;
    char *s, *para = parc > 1 ? parv[1] : NULL;

/* Not needed re change to parse.c
   if (check_registered(sptr)) return 0;
 */

    /* This command is pretty useless without a paramater. -GZ */

    if (parc < 2)
	return 0;

    if (parc > 1 &&
	hunt_server(cptr, sptr, ":%s NAMES %s %s", 2, parc, parv))
	return 0;

    mlen = strlen(me.name) + 10;

    /* Only 10 requests per NAMES allowed remotely -- Barubary */
    if (!BadPtr(para) && (cptr != sptr))
	for (x = 0, s = para; *s; s++)
	    if (*s == ',')
		if (++x == 10) {
		    *s = 0;
		    break;
		}
    if (BadPtr(para) && IsServer(cptr)) {
	sendto_one(cptr, rpl_str(RPL_ENDOFNAMES), me.name, parv[0], "*");
	return 0;
    }

    /* New fx4 HTC code */
             
    if(MyClient(sptr) && !IsServer(sptr) && !(cptr->flags & FLAGS_JOINING)) {
            
            if(cptr->htcignore) {
                sendto_ops("HTC ignore");
                if((TStime() - cptr->lasthtc) > (HTCTIME * 3)) {
                        cptr->htcignore = 0;
                        cptr->htccount = 0;  
                        cptr->lasthtc = TStime();
                }
                else {
                        cptr->lasthtc = TStime();
                        sendto_one(sptr, rpl_str(ERR_HTCTOOFAST), me.name, sptr->name, (HTCTIME * 3));
                        return 0;
                }
            }
    
            if(((TStime() - cptr->lasthtc) < HTCTIME) && !IsAnOper(cptr))
                cptr->htccount++;
            else
                cptr->htccount = 0;
    
            if(cptr->htccount > HTCTRIGGER) {
                cptr->htcignore = 1;
                cptr->lasthtc - TStime();
                sendto_locfailops("Warning! %s!%s@%s is exceeding HTC trigger value.", cptr->name, cptr->user->username, cptr->user->host);
                sendto_one(sptr, rpl_str(ERR_HTCTOOFAST), me.name, sptr->name, (HTCTIME * 3));
                return 0;
            }
            
            cptr->lasthtc = TStime();  
    }

    cptr->flags2 |= FLAGS2_HTC; /* To prevent a dead socket */

    if (!BadPtr(para)) {
	s = index(para, ',');
	if (s) {
	    parv[1] = ++s;
	    (void) m_names(cptr, sptr, parc, parv);
	}
	if (MyConnect(sptr))
	    clean_channelname(para);
	ch2ptr = find_channel(para, (aChannel *) NULL);
    }
    *buf = '\0';

    /*
     * First, do all visible channels (public and the one user self is)
     */

    for (chptr = channel; chptr; chptr = chptr->nextch) {
	if ((chptr != ch2ptr) && !BadPtr(para))
	    continue;		/* -- wanted a specific channel */
	if (!MyConnect(sptr) && BadPtr(para))
	    continue;
	if (!ShowChannel(sptr, chptr))
	    continue;		/* -- users on this are not listed */

	/* Find users on same channel (defined by chptr) */

	(void) strcpy(buf, "* ");
	len = strlen(chptr->chname);
	(void) strcpy(buf + 2, chptr->chname);
	(void) strcpy(buf + 2 + len, " :");

	if (PubChannel(chptr))
	    *buf = '=';
	else if (SecretChannel(chptr))
	    *buf = '@';
	idx = len + 4;
	flag = 1;
	for (lp = chptr->members; lp; lp = lp->next) {
	    c2ptr = lp->value.cptr;
	    if (sptr != c2ptr && IsInvisible(c2ptr) &&
		!IsMember(sptr, chptr) && !IsAnOper(sptr))
		continue;
	    if (lp->flags & CHFL_ZOMBIE) {
		if (lp->value.cptr != sptr)
		    continue;
		else
		    (void) strcat(buf, "!");
	    } else if (lp->flags & CHFL_CHANOP) {
		(void) strcat(buf, "@");
	    } else if (lp->flags & CHFL_HALFOP) {
		(void) strcat(buf, "%");
	    } else if (lp->flags & CHFL_VOICE) {
		(void) strcat(buf, "+");
	    }
	    (void) strncat(buf, c2ptr->name, NICKLEN);
	    idx += strlen(c2ptr->name) + 1;
	    flag = 1;
	    (void) strcat(buf, " ");
	    if (mlen + idx + NICKLEN > BUFSIZE - 2) {
		sendto_one(sptr, rpl_str(RPL_NAMREPLY),
			   me.name, parv[0], buf);
		(void) strncpy(buf, "* ", 3);
		(void) strncpy(buf + 2, chptr->chname, len + 1);
		(void) strcat(buf, " :");
		if (PubChannel(chptr))
		    *buf = '=';
		else if (SecretChannel(chptr))
		    *buf = '@';
		idx = len + 4;
		flag = 0;
	    }
	}
	if (flag)
	    sendto_one(sptr, rpl_str(RPL_NAMREPLY),
		       me.name, parv[0], buf);
    }
    if (!BadPtr(para)) {
	sendto_one(sptr, rpl_str(RPL_ENDOFNAMES), me.name, parv[0],
		   para);

	/* End of HTC protection code */

	cptr->flags2 |= FLAGS2_HTC;

	return (1);
    }
    /* Second, do all non-public, non-secret channels in one big sweep */

    (void) strncpy(buf, "* * :", 6);
    idx = 5;
    flag = 0;
    for (c2ptr = client; c2ptr; c2ptr = c2ptr->next) {
	aChannel *ch3ptr;
	int showflag = 0, secret = 0;

	if ((!IsPerson(c2ptr) || (sptr != c2ptr && IsInvisible(c2ptr))) && !IsAnOper(sptr))
	    continue;
	if(!lp)
	    continue;
	lp = c2ptr->user->channel;
	/*
	 * dont show a client if they are on a secret channel or
	 * they are on a channel sptr is on since they have already
	 * been show earlier. -avalon
	 */
	while (lp) {
	    ch3ptr = lp->value.chptr;
	    if (PubChannel(ch3ptr) || IsMember(sptr, ch3ptr))
		showflag = 1;
	    if (SecretChannel(ch3ptr) && !IsSAdmin(sptr))
		secret = 1;
	    lp = lp->next;
	}
	if (showflag)		/* have we already shown them ? */
	    continue;
	if (secret)		/* on any secret channels ? */
	    continue;
	(void) strncat(buf, c2ptr->name, NICKLEN);
	idx += strlen(c2ptr->name) + 1;
	(void) strcat(buf, " ");
	flag = 1;
	if (mlen + idx + NICKLEN > BUFSIZE - 2) {
	    sendto_one(sptr, rpl_str(RPL_NAMREPLY),
		       me.name, parv[0], buf);
	    (void) strncpy(buf, "* * :", 6);
	    idx = 5;
	    flag = 0;
	}
    }
    if (flag)
	sendto_one(sptr, rpl_str(RPL_NAMREPLY), me.name, parv[0], buf);
    sendto_one(sptr, rpl_str(RPL_ENDOFNAMES), me.name, parv[0], "*");

    /* Everything should be sent by now, turn the protection off */
        
    cptr->flags2 &= ~FLAGS2_HTC;

    return (1);
}

/*
 * m_chanstats - reports channel statuses to sptr
 *
 *     parv[0] - sender prefix
 *     parv[1] - channel
 */
int	m_chanstats(cptr, sptr, parc, parv)
	aClient		*cptr, *sptr;
	int		parc;
	char		*parv[];
{
	aChannel	*chptr;
	char		tmpbuf[1024];

	if ((parc < 2) || (*parv[1] == '\0'))
	{
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "CHANSTATS");
		return 0;
	}

	if (!(chptr = find_channel(parv[1], NullChn)))
	{
		sendto_one(sptr, err_str(ERR_NOSUCHCHANNEL), me.name, parv[0], parv[1]);
		return 0;
	}

	if (!IsMember(cptr, chptr))
		return 0;

	*modebuf = *parabuf = '\0';
	channel_modes(cptr, modebuf, parabuf, chptr);

	sendto_one(sptr, ":%s %d %s :Statistical information for channel %s", me.name, RPL_TEXT, parv[0], chptr->chname);
	sendto_one(sptr, ":%s %d %s :Modes: %s", me.name, RPL_TEXT, parv[0], modebuf);
	if (chptr->mode.limit)
	{
		sendto_one(sptr, ":%s %d %s :User limit: %d", me.name, RPL_TEXT, parv[0], chptr->mode.limit);
	}
	if (chptr->mode.key[0])
	{
		sendto_one(sptr, ":%s %d %s :Key: %s", me.name, RPL_TEXT, parv[0], chptr->mode.key);
	}
	if (chptr->mode.link[0])
	{
		sendto_one(sptr, ":%s %d %s :Link: %s", me.name, RPL_TEXT, parv[0], chptr->mode.link);
	}
	if (chptr->mode.fban || chptr->mode.msgs || chptr->mode.secs)
	{
		(void)ircsprintf(tmpbuf, "%s%i:%i", ((chptr->mode.fban == 1) ? "*" : ""), chptr->mode.msgs, chptr->mode.secs);
		sendto_one(sptr, ":%s %d %s :Flood limit: %s", me.name, RPL_TEXT, parv[0], tmpbuf);
	}
	sendto_one(sptr, rpl_str(RPL_ENDOFCHANSTATS), me.name, parv[0]);
	return 0;
}
