/****************************************************************************
 * IRC - Internet Relay Chat, src/s_gline.c
 * Copyright (C) 2000 Alasdair McWilliam. RageIRCd, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "userload.h"
#include "config.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include "h.h"
#include "dynconf2.h"
#include "gline.h"

#ifndef GPATH
#define GPATH "gline.log"
#endif
#define GLINE_LOGFILE GPATH

#define AllocCpy(x, y) x = (char *)MyMalloc(strlen(y) + 1); strcpy(x, y)

void	list_glines PROTO((aClient *));
void	add_gline PROTO((char *, char *, char *, time_t, time_t, char *));
void	check_expire_gline PROTO((void));

aGline	*expire_gline PROTO((aGline *));
aGline	*glines = NULL;

/*
 * m_gline - /quote gline add/del username hostname [duration in seconds] [:reason]
 *
 * parv[0] - sender prefix
 * parv[1] - add/del
 * parv[2] - user@host
 * parv[3] - duration (seconds)
 * parv[4] - reason
 */
int	m_gline(cptr, sptr, parc, parv)
	aClient		*cptr, *sptr;
	int		parc;
	char		*parv[];
{
	char		*set_by, *username, *hostname, *reason, *p, *tmpmask, *x;
	char		gmt[256];
	time_t		secs, expire, set_at;
	aGline		*ag;
	FILE		*log;
	int		perm, i;

	if (!IsOper(sptr) && !IsServer(sptr))
	{
		sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
		return 0;
	}

	if (parc < 2)
	{
		sendto_one(sptr, ":%s NOTICE %s :*** GLine -- Syntax: /gline <add/del> <user@host> [duration in seconds] [:reason]",
			me.name, parv[0]);
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "GLINE");
		return 0;
	}

	(void)check_expire_gline();
	if (parc > 2)
	{
		tmpmask = parv[2];

		/* Check to see if the mask is a user@mask */
		x = strchr(tmpmask, '@');
		if (!x)
		{
			sendto_one(sptr, ":%s NOTICE %s :*** GLine -- Please use a user@mask", me.name, parv[0]);
			return 0;
		}
	}

	username = strtok(tmpmask, "@");
	if (!username)
	{
		sendto_one(sptr, ":%s NOTICE %s :*** GLine -- Please specify a username in the user@host", me.name, parv[0]);
		return 0;
	}

	hostname = strtok(NULL, "@");
	if (!hostname)
	{
		sendto_one(sptr, ":%s NOTICE %s :*** GLine -- Please specify a hostname in the user@host", me.name, parv[0]);
		return 0;
	}

	if (!strcasecmp(parv[1], "ADD"))
	{
		if (parc < 5)
		{
			sendto_one(sptr, ":%s NOTICE %s :*** GLine -- Syntax: /gline add <user@host> <duration in seconds> :<reason>",
				me.name, parv[0]);
			sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "GLINE");
			return 0;
		}

		set_by = parv[0];
		reason = parv[4];

		if (strlen(reason) > TOPICLEN)
		{
			reason[TOPICLEN + 1] = '\0';
		}

		if (strlen(username) > (USERLEN+1))
		{
			sendto_one(sptr, ":%s NOTICE %s :*** GLine -- Username is too long (max %i chars)", me.name, parv[0],
				(USERLEN+1));
			return 0;
		}

		if (strlen(hostname) > (HOSTLEN+1))
		{
			sendto_one(sptr, ":%s NOTICE %s :*** GLine -- Hostname is too long (max %i chars)", me.name, parv[0],
				(HOSTLEN+1));
			return 0;
		}

		p = hostname;
		if (p)
		{
			p++;
			i = 0;
			while (*p)
			{
				if ((*p != '*') && (*p != '.'))
				{
					i++;
				}
				p++;
			}

			if (i < 4)
			{
				sendto_one(sptr, ":%s NOTICE %s :*** GLine -- Hostname is too inaccurate", me.name, parv[0]);
				return 0;
			}
		}

		if (find_gline(username, hostname) == 1)
		{
			if (!IsServer(sptr) && MyClient(sptr))
			{
				sendto_one(sptr, ":%s NOTICE %s :*** GLine -- %s@%s is already on the GLine list", me.name, parv[0],
					username, hostname);
			}
			return 0;
		}

		secs = atol(parv[3]);
		perm = 0;

		if (secs < 0)
		{
			if (!IsServer(sptr) && MyClient(sptr))
			{
				sendto_one(sptr, ":%s NOTICE %s :*** GLine -- You cannot GLine somone for a negative amount of time (%i)",
					me.name, parv[0], secs);
			}
			return 0;
		}
		else if (secs == 0)
		{
			perm = 1;
			expire = secs;
		}
		else
		{
			perm = 0;
			expire = secs + TStime();
		}

#ifdef __OpenBSD__
		strncpy(gmt, asctime(gmtime((clock_t *)&expire)), sizeof(gmt));
#else
		strncpy(gmt, asctime(gmtime((time_t *)&expire)), sizeof(gmt));
#endif

		gmt[strlen(gmt) - 1] = '\0';
		set_at = TStime();

		if (!IsServer(sptr))
		{
			if (perm == 1)
			{
				sendto_umode(UMODE_OPER, "*** GLine -- %s added GLine for %s@%s which will not expire (%s)",
					set_by, username, hostname, reason);
			}
			else
			{
				sendto_umode(UMODE_OPER, "*** GLine -- %s added GLine for %s@%s which will expire on %s GMT (%s)",
					set_by, username, hostname, gmt, reason);
			}
		}

		if (LOG_GLINES == TRUE)
		{
			log = fopen(GLINE_LOGFILE, "a");
			(void)fprintf(log, "| Added:      %s\n", myctime(TStime()));
			(void)fprintf(log, "| Mask:       %s@%s\n", username, hostname);
			(void)fprintf(log, "| Set by:     %s\n", set_by);
			(void)fprintf(log, "| Expires:    %s\n", (perm == 1) ? "permanent" : gmt);
			(void)fprintf(log, "| Reason:     %s\n", reason);
			(void)fprintf(log, "|------------------------------------------------------------------------------\n");
			fclose(log);
		}

		sendto_serv_butone(cptr, ":%s GLINE ADD %s@%s %lu :%s", parv[0], username, hostname, expire, reason);
		(void)add_gline(set_by, username, hostname, expire, set_at, reason);
		(void)sweep_gline();
		return 1;
	}
	else if (!strcasecmp(parv[1], "DEL"))
	{
		if (parc < 3)
		{
			sendto_one(sptr, ":%s NOTICE %s :*** GLine -- Syntax: /gline del <user@host>", me.name, parv[0]);
			sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "GLINE");
			return 0;
		}
		if (find_gline(username, hostname) == 0)
		{
			if (!IsServer(sptr) && MyClient(sptr))
			{
				sendto_one(sptr, ":%s NOTICE %s :*** GLine -- %s@%s is not on the GLine list", me.name, parv[0], username,
					hostname);
			}
			return 0;
		}

		for (ag = glines; ag; ag = ag->next)
		{
			sendto_umode(UMODE_OPER, "*** GLine -- %s deleted GLine on %s@%s (set by %s %li seconds ago)",
				parv[0], ag->username, ag->hostname, ag->set_by, (TStime() - ag->set_at));

			if (LOG_GLINES == TRUE)
			{
				log = fopen(GLINE_LOGFILE, "a");
				(void)fprintf(log, "| Deleted:     %s\n", myctime(TStime()));
				(void)fprintf(log, "| Mask:        %s@%s\n", username, hostname);
				(void)fprintf(log, "| Added by:    %s\n", ag->set_by);
				(void)fprintf(log, "| Set:         %li seconds ago\n", (TStime() - ag->set_at));
				(void)fprintf(log, "| Deleted by:  %s\n", parv[0]);
				(void)fprintf(log, "|------------------------------------------------------------------------------\n");
				fclose(log);
			}

			del_gline(username, hostname);
			sendto_serv_butone(cptr, ":%s GLINE DEL %s@%s", parv[0], username, hostname);
			return 0;
		}
	}
	else
	{
		sendto_one(sptr, ":%s NOTICE %s :*** GLine -- Syntax: /gline <add/del> <user@host> [duration in seconds] [:reason]",
			me.name, parv[0]);
		return 0;
	}
}

/*
 * find_gline - if paramteres match a GLine, returns 1; otherwise, returns 0;
 */
int	find_gline(username, hostname)
	char	*username, *hostname;
{
	aGline		*ag;

	(void)check_expire_gline();

	for (ag = glines; ag; ag = ag->next)
	{
		if (!match(ag->username, username) && !match(ag->hostname, hostname))
		{
			return 1;
		}
	}
	return 0;
}

/*
 * sweep_gline - sweeps local server for users matching G:Line(s)
 */
int	sweep_gline()
{
	aClient		*acptr;
	int		i;

	(void)check_expire_gline();

	for (i = 0; i <= highest_fd; i++)
	{
		if (!(acptr = local[i]) || !IsClient(acptr)) {
			continue;
		}
		find_glined_user(acptr);
	}
	return 1;
}

/*
 * find_glined_user - checks if user matches gline
 */
int	find_glined_user(cptr)
	aClient		*cptr;
{
	aGline		*ag;
	int		found = 0;

	(void)check_expire_gline();

	for (ag = glines; ag; ag = ag->next) {
		found = 0;
		if (!match(ag->username, cptr->user->username)) {
			if (!match(ag->hostname, cptr->user->host) || !match(ag->hostname, (char *)inet_ntoa(cptr->ip))) {
				found = 1;
			}
			if (!match(ag->hostname, me.name)) {
				found = 0;
			}
		}
		if (found == 1) {
			break;
		}
	}

	if (found == 1) {
		ircstp->is_ref++;
		sendto_one(cptr, ":%s NOTICE %s :*** You have been banned from this network by %s until %s", me.name,
			cptr->name, ag->set_by, myctime(ag->expire));
		sendto_one(cptr, ":%s NOTICE %s :*** Reason: %s", me.name, cptr->name, ag->reason);
		sendto_one(cptr, ":%s NOTICE %s :*** For more information, contact %s.", me.name,
			cptr->name, ircnet_kline);
		sendto_ops("GLine active for %s (%s@%s) [%s]", cptr->name, cptr->user->username, cptr->user->host, ag->reason);
		return exit_client(cptr, cptr, &me, "G-Lined");
	}
	return 0;
}

/*
 * add_gline - creates a GLine string with parameters given by m_gline
 */
void	add_gline(set_by, username, hostname, expire, set_at, reason)
	char		*username, *hostname;
	char		*set_by, *reason;
	time_t		expire, set_at;
{
	aGline		*ag;

	ag = (aGline *)MyMalloc(sizeof(aGline));

	AllocCpy(ag->set_by, set_by);
	AllocCpy(ag->username, username);
	AllocCpy(ag->hostname, hostname);
	AllocCpy(ag->reason, reason);
	ag->expire = expire;
	ag->set_at = set_at;

	ag->next = glines;
	ag->prev = NULL;

	if (glines)
	{
		glines->prev = ag;
	}
	glines = ag;
}

/*
 * del_gline - delete a GLine string matching parameters given by m_gline
 */
int	del_gline(username, hostname)
	char		*username, *hostname;
{
	aGline		*ag;

	for (ag = glines; ag; ag = ag->next)
	{
		if (!match(ag->username, username) && !match(ag->hostname, hostname))
		{
			MyFree((char *)ag->set_by);
			MyFree((char *)ag->username);
			MyFree((char *)ag->hostname);
			MyFree((char *)ag->reason);

			if (ag->prev)
			{
				ag->prev->next = ag->next;
			}
			else
			{
				glines = ag->next;
			}
			if (ag->next)
			{
				ag->next->prev = ag->prev;
			}
			MyFree((aGline *) ag);

			return 1;
		}
	}
	return 0;
}

/*
 * check_expire_gline - checks for glines that are expired
 */
void	check_expire_gline(void)
{
	aGline		*ag, x;
	time_t		nowtime = TStime();

	for (ag = glines; ag; ag = ag->next)
	{
		if ((ag->expire <= nowtime) && (ag->expire != 0))
		{
			expire_gline(ag);
		}
	}
}

/*
 * expire_gline - delete an expired gline
 */
aGline	*expire_gline(ag)
	aGline		*ag;
{
	FILE		*log;

	sendto_umode(UMODE_OPER, "*** GLine -- GLine for %s@%s (set by %s %li seconds ago) has now expired",
		ag->username, ag->hostname, ag->set_by, (TStime() - ag->set_at), ag->reason);

	if (LOG_GLINES == TRUE)
	{
		log = fopen(GLINE_LOGFILE, "a");
		(void)fprintf(log, "| Expired:     %s\n", myctime(TStime()));
		(void)fprintf(log, "| Mask:        %s@%s\n", ag->username, ag->hostname);
		(void)fprintf(log, "| Added by:    %s\n", ag->set_by);
		(void)fprintf(log, "| Set:         %li seconds ago\n", (TStime() - ag->set_at));
		(void)fprintf(log, "|------------------------------------------------------------------------------\n");
		fclose(log);
	}
	del_gline(ag->username, ag->hostname);
	return 0;
}

/*
 * sync_glines
 */
void	sync_glines(cptr)
	aClient		*cptr;
{
	aGline		*ag;
	time_t		expire;

	(void)check_expire_gline();

	for (ag = glines; ag; ag = ag->next)
	{
		expire = ag->expire;
		if (expire < 0)
		{
			expire = -expire;
		}
		else if (expire > 0)
		{
			expire = expire - TStime();
		}
		sendto_one(cptr, ":%s GLINE ADD %s@%s %lu :%s", me.name, ag->username, ag->hostname,
			(ag->expire == 0) ? ag->expire : ag->expire - TStime(), ag->reason);
	}
}

/*
 * list_glines - generates a list of GLines
 */
void	list_glines(sptr)
	aClient		*sptr;
{
	aGline		*ag;

	(void)check_expire_gline();

	for (ag = glines; ag; ag = ag->next)
	{
		sendto_one(sptr, ":%s %i %s :G %s@%s %s (%s)", me.name, RPL_TEXT, sptr->name, ag->username, ag->hostname,
			ag->reason, ag->set_by);
	}
}
