/****************************************************************************
 * IRC - Internet Relay Chat, src/s_motd.c
 * Copyright (C) 2001 Alasdair McWilliam. RageIRCd, Inc.
 *
 * Contains code from Bahamut - Copyright (C) The Bahamut Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "config.h"
#include "struct.h"
#include "dynconf2.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "userload.h"
#include <sys/stat.h>
#include <fcntl.h>
#include <utmp.h>
#include <time.h>
#include "h.h"

char	motd_last_changed_date[34];
char	rules_last_changed_date[34];
char	news_last_changed_date[34];
char	omotd_last_changed_date[34];
char	smotd_last_changed_date[34];

aMotd		*motd;
aMotd		*rules;
aMotd		*news;
aMotd		*omotd;
aMotd		*smotd;

struct	tm	*motd_tm;
struct	tm	*rules_tm;
struct	tm	*news_tm;
struct	tm	*omotd_tm;
struct	tm	*smotd_tm;

/*
 * bootup initialization
 */
int	init_motd(mode)
	int	mode;
{
	if (mode == 0)
	{
		motd_tm = NULL;
		rules_tm = NULL;
		news_tm = NULL;
		omotd_tm = NULL;
		smotd_tm = NULL;

		motd = (aMotd *)NULL;
		rules = (aMotd *)NULL;
		news = (aMotd *)NULL;
		omotd = (aMotd *)NULL;
		smotd = (aMotd *)NULL;
	}
	load_motd();
	load_rules();
	load_news();
	load_omotd();
	load_smotd();
}

/*
 * file loaders
 */
int	load_motd()
{
	aMotd		*temp, *last;
	struct	stat	sb;
	char		buffer[MOTDLINELEN], *tmp;
	int		fd;

	while (motd)
	{
		temp = motd->next;
		MyFree(motd);
		motd = temp;
	}

	fd = open(MPATH, O_RDONLY);
	if (fd == -1)
	{
		return;
	}
	fstat(fd, &sb);
	motd_tm = localtime(&sb.st_mtime);
	last = (aMotd *)NULL;

	while (dgets(fd, buffer, MOTDLINELEN - 1) > 0)
	{
		if ((tmp = (char *)strchr(buffer, '\n')))
		{
			*tmp = '\0';
		}
		if ((tmp = (char *)strchr(buffer, '\r')))
		{
			*tmp = '\0';
		}

		temp = (aMotd *)MyMalloc(sizeof(aMotd));

		strncpyzt(temp->line, buffer, MOTDLINELEN);
		temp->next = (aMotd *)NULL;

		if (!motd)
		{
			motd = temp;
		}
		else
		{
			last->next = temp;
		}
		last = temp;
	}

	close(fd);

	if (motd_tm)
	{
		(void)ircsprintf(motd_last_changed_date, "%d:%02d, %d/%d/%d", motd_tm->tm_hour, motd_tm->tm_min,
			motd_tm->tm_mday, motd_tm->tm_mon + 1, 1900 + motd_tm->tm_year);
	}
}

int	load_rules()
{
	aMotd		*temp, *last;
	struct	stat	sb;
	char		buffer[MOTDLINELEN], *tmp;
	int		fd;

	while (rules)
	{
		temp = rules->next;
		MyFree(rules);
		rules = temp;
	}

	fd = open(RPATH, O_RDONLY);
	if (fd == -1)
	{
		return;
	}

	fstat(fd, &sb);
	rules_tm = localtime(&sb.st_mtime);
	last = (aMotd *)NULL;

	while (dgets(fd, buffer, MOTDLINELEN - 1) > 0)
	{
		if ((tmp = (char *)strchr(buffer, '\n')))
		{
			*tmp = '\0';
		}
		if ((tmp = (char *)strchr(buffer, '\r')))
		{
			*tmp = '\0';
		}

		temp = (aMotd *)MyMalloc(sizeof(aMotd));

		strncpyzt(temp->line, buffer, MOTDLINELEN);
		temp->next = (aMotd *)NULL;

		if (!rules)
		{
			rules = temp;
		}
		else
		{
			last->next = temp;
		}
		last = temp;
	}

	close(fd);

	if (rules_tm)
	{
		(void)ircsprintf(rules_last_changed_date, "%d:%02d, %d/%d/%d", rules_tm->tm_hour, rules_tm->tm_min,
			rules_tm->tm_mday, rules_tm->tm_mon + 1, 1900 + rules_tm->tm_year);
	}
}

int	load_news()
{
	aMotd		*temp, *last;
	struct	stat	sb;
	char		buffer[MOTDLINELEN], *tmp;
	int		fd;

	while (news)
	{
		temp = news->next;
		MyFree(news);
		news = temp;
	}

	fd = open(XPATH, O_RDONLY);
	if (fd == -1)
	{
		return;
	}

	fstat(fd, &sb);
	news_tm = localtime(&sb.st_mtime);
	last = (aMotd *)NULL;

	while (dgets(fd, buffer, MOTDLINELEN - 1) > 0)
	{
		if ((tmp = (char *)strchr(buffer, '\n')))
		{
			*tmp = '\0';
		}
		if ((tmp = (char *)strchr(buffer, '\r')))
		{
			*tmp = '\0';
		}

		temp = (aMotd *)MyMalloc(sizeof(aMotd));

		strncpyzt(temp->line, buffer, MOTDLINELEN);
		temp->next = (aMotd *)NULL;

		if (!news)
		{
			news = temp;
		}
		else
		{
			last->next = temp;
		}
		last = temp;
	}

	close(fd);

	if (news_tm)
	{
		(void)ircsprintf(news_last_changed_date, "%d:%02d, %d/%d/%d", news_tm->tm_hour, news_tm->tm_min,
			news_tm->tm_mday, news_tm->tm_mon + 1, 1900 + news_tm->tm_year);
	}
}

int	load_omotd()
{
	aMotd		*temp, *last;
	struct	stat	sb;
	char		buffer[MOTDLINELEN], *tmp;
	int		fd;

	while (omotd)
	{
		temp = omotd->next;
		MyFree(omotd);
		omotd = temp;
	}

	fd = open(OPATH, O_RDONLY);
	if (fd == -1)
	{
		return;
	}

	fstat(fd, &sb);
	omotd_tm = localtime(&sb.st_mtime);
	last = (aMotd *)NULL;

	while (dgets(fd, buffer, MOTDLINELEN - 1) > 0)
	{
		if ((tmp = (char *)strchr(buffer, '\n')))
		{
			*tmp = '\0';
		}
		if ((tmp = (char *)strchr(buffer, '\r')))
		{
			*tmp = '\0';
		}

		temp = (aMotd *)MyMalloc(sizeof(aMotd));

		strncpyzt(temp->line, buffer, MOTDLINELEN);
		temp->next = (aMotd *)NULL;

		if (!omotd)
		{
			omotd = temp;
		}
		else
		{
			last->next = temp;
		}
		last = temp;
	}

	close(fd);

	if (omotd_tm)
	{
		(void)ircsprintf(omotd_last_changed_date, "%d:%02d, %d/%d/%d", omotd_tm->tm_hour, omotd_tm->tm_min,
			omotd_tm->tm_mday, omotd_tm->tm_mon + 1, 1900 + omotd_tm->tm_year);
	}
}

int	load_smotd()
{
	aMotd		*temp, *last;
	struct	stat	sb;
	char		buffer[MOTDLINELEN], *tmp;
	int		fd, lines = 0;

	while (smotd)
	{
		temp = smotd->next;
		MyFree(smotd);
		smotd = temp;
	}

	fd = open(HPATH, O_RDONLY);
	if (fd == -1)
	{
		return;
	}

	fstat(fd, &sb);
	smotd_tm = localtime(&sb.st_mtime);
	last = (aMotd *)NULL;

	while (dgets(fd, buffer, MOTDLINELEN - 1) > 0)
	{
		if (lines >= 5)
		{
			break;
		}
		if ((tmp = (char *)strchr(buffer, '\n')))
		{
			*tmp = '\0';
		}
		if ((tmp = (char *)strchr(buffer, '\r')))
		{
			*tmp = '\0';
		}

		temp = (aMotd *)MyMalloc(sizeof(aMotd));

		strncpyzt(temp->line, buffer, MOTDLINELEN);
		temp->next = (aMotd *)NULL;

		if (!smotd)
		{
			smotd = temp;
		}
		else
		{
			last->next = temp;
		}
		last = temp;
		lines++;
	}

	close(fd);

	if (smotd_tm)
	{
		(void)ircsprintf(smotd_last_changed_date, "%d:%02d, %d/%d/%d", smotd_tm->tm_hour, smotd_tm->tm_min,
			smotd_tm->tm_mday, smotd_tm->tm_mon + 1, 1900 + smotd_tm->tm_year);
	}
}

/*
 * user functions
 */
int	m_motd(cptr, sptr, parc, parv)
	aClient		*cptr, *sptr;
	int		parc;
	char		*parv[];
{
	if (hunt_server(cptr, sptr, ":%s MOTD :%s", 1, parc, parv) != HUNTED_ISME)
	{
		return 0;
	}

	if (!IsOper(sptr))
	{
		sendto_ops("/MOTD requested by %s (%s@%s)", sptr->name, sptr->user->username,
			IsMasked(sptr) ? sptr->user->maskedhost : sptr->user->host);
	}

	send_motd(cptr, sptr, parc, parv);
	return 0;
}

int	m_rules(cptr, sptr, parc, parv)
	aClient		*cptr, *sptr;
	int		parc;
	char		*parv[];
{
	if (hunt_server(cptr, sptr, ":%s RULES :%s", 1, parc, parv) != HUNTED_ISME)
	{
		return 0;
	}

	if (!IsOper(sptr))
	{
		sendto_ops("/RULES requested by %s (%s@%s)", sptr->name, sptr->user->username,
			IsMasked(sptr) ? sptr->user->maskedhost : sptr->user->host);
	}

	send_rules(cptr, sptr, parc, parv);
	return 0;
}

int	m_news(cptr, sptr, parc, parv)
	aClient		*cptr, *sptr;
	int		parc;
	char		*parv[];
{
	if (hunt_server(cptr, sptr, ":%s NEWS :%s", 1, parc, parv) != HUNTED_ISME)
	{
		return 0;
	}

	if (!IsOper(sptr))
	{
		sendto_ops("/NEWS requested by %s (%s@%s)", sptr->name, sptr->user->username,
			IsMasked(sptr) ? sptr->user->maskedhost : sptr->user->host);
	}

	send_news(cptr, sptr, parc, parv);
	return 0;
}

int	m_opermotd(cptr, sptr, parc, parv)
	aClient		*cptr, *sptr;
	int		parc;
	char		*parv[];
{
	if (hunt_server(cptr, sptr, ":%s OPERMOTD :%s", 1, parc, parv) != HUNTED_ISME)
	{
		return 0;
	}

	if (!IsAnOper(sptr))
	{
		sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
		return 0;
	}

	send_omotd(cptr, sptr, parc, parv);
	return 0;
}

/*
 * send functions
 */
int	send_motd(cptr, sptr, parc, parv)
	aClient		*cptr, *sptr;
	int		parc;
	char		*parv[];
{
	aMotd		*temp;
	struct	tm	*tm;

	tm = motd_tm;

	if (motd == (aMotd *)NULL)
	{
		sendto_one(sptr, err_str(ERR_NOMOTD), me.name, parv[0], "MOTD");
		return 0;
	}

	if (strlen(motd_last_changed_date) > 1)
	{
		sendto_one(sptr, ":%s NOTICE %s :*** The MOTD was last changed at %s", me.name, parv[0], motd_last_changed_date);
	}
	sendto_one(sptr, rpl_str(RPL_MOTDSTART), me.name, parv[0], me.name, "Message of the Day");

	if (tm)
	{
		sendto_one(sptr, ":%s %d %s :- %d:%02d, %d/%d/%d", me.name, RPL_MOTD, parv[0],
			tm->tm_hour, tm->tm_min, tm->tm_mday, tm->tm_mon + 1, 1900 + tm->tm_year);
	}

	temp = motd;
	while (temp)
	{
		sendto_one(sptr, rpl_str(RPL_MOTD), me.name, parv[0], temp->line);
		temp = temp->next;
	}
	sendto_one(sptr, rpl_str(RPL_ENDOFMOTD), me.name, parv[0], "MOTD");
	return 0;
}

int	send_rules(cptr, sptr, parc, parv)
	aClient		*cptr, *sptr;
	int		parc;
	char		*parv[];
{
	aMotd		*temp;
	struct	tm	*tm;

	tm = rules_tm;

	if (rules == (aMotd *)NULL)
	{
		sendto_one(sptr, err_str(ERR_NOMOTD), me.name, parv[0], "RULES");
		return 0;
	}

	if (strlen(rules_last_changed_date) > 1)
	{
		sendto_one(sptr, ":%s NOTICE %s :*** The RULES were last changed at %s", me.name, parv[0], rules_last_changed_date);
	}
	sendto_one(sptr, rpl_str(RPL_MOTDSTART), me.name, parv[0], me.name, "Server Rules");

	if (tm)
	{
		sendto_one(sptr, ":%s %d %s :- %d:%02d, %d/%d/%d", me.name, RPL_MOTD, parv[0],
			tm->tm_hour, tm->tm_min, tm->tm_mday, tm->tm_mon + 1, 1900 + tm->tm_year);
	}

	temp = rules;
	while (temp)
	{
		sendto_one(sptr, rpl_str(RPL_MOTD), me.name, parv[0], temp->line);
		temp = temp->next;
	}
	sendto_one(sptr, rpl_str(RPL_ENDOFMOTD), me.name, parv[0], "RULES");
	return 0;
}

int	send_news(cptr, sptr, parc, parv)
	aClient		*cptr, *sptr;
	int		parc;
	char		*parv[];
{
	aMotd		*temp;
	struct	tm	*tm;

	tm = news_tm;

	if (news == (aMotd *)NULL)
	{
		sendto_one(sptr, err_str(ERR_NOMOTD), me.name, parv[0], "NEWS");
		return 0;
	}

	if (strlen(news_last_changed_date) > 1)
	{
		sendto_one(sptr, ":%s NOTICE %s :*** The NEWS was last changed at %s", me.name, parv[0], news_last_changed_date);
	}
	sendto_one(sptr, rpl_str(RPL_MOTDSTART), me.name, parv[0], me.name, "Server News");

	if (tm)
	{
		sendto_one(sptr, ":%s %d %s :- %d:%02d, %d/%d/%d", me.name, RPL_MOTD, parv[0],
			tm->tm_hour, tm->tm_min, tm->tm_mday, tm->tm_mon + 1, 1900 + tm->tm_year);
	}

	temp = news;
	while (temp)
	{
		sendto_one(sptr, rpl_str(RPL_MOTD), me.name, parv[0], temp->line);
		temp = temp->next;
	}
	sendto_one(sptr, rpl_str(RPL_ENDOFMOTD), me.name, parv[0], "NEWS");
	return 0;
}

int	send_omotd(cptr, sptr, parc, parv)
	aClient		*cptr, *sptr;
	int		parc;
	char		*parv[];
{
	aMotd		*temp;
	struct	tm	*tm;

	tm = omotd_tm;

	if (omotd == (aMotd *)NULL)
	{
		sendto_one(sptr, err_str(ERR_NOMOTD), me.name, parv[0], "OPERMOTD");
		return 0;
	}

	if (strlen(omotd_last_changed_date) > 1)
	{
		sendto_one(sptr, ":%s NOTICE %s :*** The OPERMOTD was last changed at %s", me.name, parv[0], omotd_last_changed_date);
	}
	sendto_one(sptr, rpl_str(RPL_MOTDSTART), me.name, parv[0], me.name, "Operator Message of the Day");

	if (tm)
	{
		sendto_one(sptr, ":%s %d %s :- %d:%02d, %d/%d/%d", me.name, RPL_MOTD, parv[0],
			tm->tm_hour, tm->tm_min, tm->tm_mday, tm->tm_mon + 1, 1900 + tm->tm_year);
	}

	temp = omotd;
	while (temp)
	{
		sendto_one(sptr, rpl_str(RPL_MOTD), me.name, parv[0], temp->line);
		temp = temp->next;
	}
	sendto_one(sptr, rpl_str(RPL_ENDOFMOTD), me.name, parv[0], "OPERMOTD");
	return 0;
}

int	send_smotd(cptr, sptr, parc, parv)
	aClient		*cptr, *sptr;
	int		parc;
	char		*parv[];
{
	aMotd		*temp;
	struct	tm	*tm;

	tm = smotd_tm;

	if (smotd == (aMotd *)NULL)
	{
		sendto_one(sptr, rpl_str(RPL_MOTDSTART), me.name, parv[0], me.name, "Short Message of the Day");
		sendto_one(sptr, rpl_str(RPL_MOTD), me.name, parv[0], "*** This is the Short MOTD ***");
		sendto_one(sptr, rpl_str(RPL_ENDOFMOTD), me.name, parv[0], "MOTD");
		return 0;
	}

	if (strlen(smotd_last_changed_date) > 1)
	{
		sendto_one(sptr, ":%s NOTICE %s :*** The Short MOTD was last changed at %s", me.name, parv[0], smotd_last_changed_date);
	}
	sendto_one(sptr, rpl_str(RPL_MOTDSTART), me.name, parv[0], me.name, "Short Message of the Day");

	if (tm)
	{
		sendto_one(sptr, ":%s %d %s :- %d:%02d, %d/%d/%d", me.name, RPL_MOTD, parv[0], tm->tm_hour, tm->tm_min,
			tm->tm_mday, tm->tm_mon + 1, 1900 + tm->tm_year);
	}

	temp = smotd;
	while (temp)
	{
		sendto_one(sptr, rpl_str(RPL_MOTD), me.name, parv[0], temp->line);
		temp = temp->next;
	}
	sendto_one(sptr, rpl_str(RPL_ENDOFMOTD), me.name, parv[0], "MOTD");
	return 0;
}

