/************************************************************************
 * IRC - Internet Relay Chat, src/s_sync.c
 * Copyright (C) 2000 Alasdair McWilliam. RageIRCd, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "ircstats.h"
#include "patchlevel.h"
#include "struct.h"
#include "dynconf2.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "userload.h"
#include <sys/stat.h>
#include <fcntl.h>
#include <utmp.h>
#include <time.h>
#include "h.h"
#include "sjoin.h"

extern	ircstats	IRCstats;

/*
 * m_sync - checks that an incomming server is suitable for link.
 *          (originally based on NETINFO from UnrealIRCd)
 *
 * parv[0] - sender prefix
 * parv[1] - series name
 * parv[2] - irc network
 * parv[3] - max global user count
 * parv[4] - TS
 * parv[5] - service?
 *
 * revamped. -TD (December 2000)
 *
 * This code is eventually going to be removed when the new
 * Server-II protocol is implemented. -TD (April 2001)
 */
int	m_sync(cptr, sptr, parc, parv)
	aClient	*cptr, *sptr;
	int	parc;
	char	*parv[];
{
	int	services, globalmax, synctime;
	int	caps, support_snick, support_sjoin, support_nickv2;
	time_t	z;

	if (!IsServer(cptr) || IsPerson(cptr) || (parc < 6))
		return 0;

	services = atoi(parv[5]);
	if (services == 1)
	{
		goto services;
	}

	if (match(SERIES, parv[1]))
	{
		sendto_ops("Dropping link %s - IRCd is incompatible", cptr->name);
		sendto_one(cptr, "ERROR :IRCd is incompatible");

		exit_client(cptr, cptr, cptr, "IRCd is incompatible");
		return 0;
	}

	if (strcasecmp(ircnet, parv[2]))
	{
		sendto_ops("Dropping link %s - IRC network mismatch", cptr->name);
		sendto_one(cptr, "ERROR :IRC network mismatch");

		exit_client(cptr, cptr, cptr, "IRC network mismatch");
		return 0;
	}

	/* update max global user count */
	globalmax = atoi(parv[3]);
	if (globalmax > IRCstats.global_max)
	{
		IRCstats.global_max = globalmax;
		save_stats();

		sendto_ops("Max global user count is now %i (set by link %s)", IRCstats.global_max, cptr->name);
	}

		services:
	/* check timestamps */
	/*synctime = atoi(parv[4]);
	z = TStime();
	if ((z - synctime) < 0)
	{
		sendto_ops("Possible negative TS split at link %s (%i - %i = %i)", cptr->name, (z), (synctime), (z - synctime));
			nuked this, it has no purpose..theres nothing you can do about it..live with it.
								- TD (June 2001)
	} */

	/* check capabilities */
	/* If SYNC is not removed when the Server-II protocol is implemented,
         * this code will be removed and replaced with the StarChat bitflag
         * protocol, which allows capabilities to be recognised *BEFORE*
         * any data has been sent to the new server.
         *                                             - TD (April 2001)
	 */
	caps = support_snick = support_nickv2 = support_sjoin = 0;
	if (SupportSNICK(cptr))
	{
		/* SNICK compliant */
		support_snick = 1;
		caps++;
	}
	else if (SupportNICKv2(cptr))
	{
		/* non-SNICK & NICKv2 compliant */
		support_nickv2 = 1;
		caps++;
	}
	if (SupportSJOIN(cptr))
	{
		/* SJOIN compliant */
		support_sjoin = 1;
		caps++;
	}

	if (caps > 0)
	{
		sendto_ops("Link %s is now synched using RTS-1 protocol [capabilities:%s%s]", cptr->name,
			((support_snick == 1) ? " SNICK" : ((support_nickv2 == 1) ? " NICKv2" : "")),
			((support_sjoin == 1) ? " SJOIN" : ""));
	}
	else
	{
		sendto_ops("Link %s is now synched", cptr->name);
	}
	SetSynced(cptr);
}

/*
 * m_snick - used to synch users and their usermodes
 *
 *	parv[0] = sender prefix
 *	parv[1] = nickname
 *	parv[2] = hopcount
 *	parv[3] = lastnick
 *	parv[4] = username
 *	parv[5] = hostname
 *	parv[6] = server
 *	parv[7] = service stamp
 *	parv[8] = user modes
 *	parv[9] = masked hostname (or '*' if client is not masked)
 *	parv[10] = real name
 *
 * Original SNICK code from StarChat IRCd v5.10.fx4. Recoded to use the RageIRCd's
 * RTS-1 protocol which, is used in the new SJOIN function.
 *
 * RTS-1 is based on the StarChat STS1 protocol, and the EFNet TS3 protocol.
 *
 *				- TD (April 2001)
 */
#define ParseMode(x)	if (what == MODE_ADD) { sptr->umodes |= (x); } \
				else if (what == MODE_DEL) { sptr->umodes &= ~(x); } \
				else { break; }

int	m_snick(cptr, sptr, parc, parv)
	aClient		*cptr, *sptr;
	int		parc;
	char		*parv[];
{
	aClient		*acptr, *snptr, *serv = NULL;
	aConfItem	*aconf;
	aSqlineItem	*asqline;
	anUser		*user;
	int		ourts, theirts;
	int		keepours, keeptheirs, keepnone, same;
	int		makestruct = 1;
	char		*p;
	int		what;

	if (!IsServer(sptr) || parc < 11)
	{
		return 0;
	}

	if (!(serv = find_server(parv[6], NULL)) || serv->from != cptr->from)
	{
		sendto_one(cptr, ":%s KILL %s :%s (Fatal synchroinzing error in SNICK)", me.name, parv[1], me.name);
		return 0;
	}

	if ((acptr = find_client(parv[1], NULL)))
	{
		ourts = theirts = keepours = keeptheirs = keepnone = same = 0;

		ourts = acptr->lastnick;
		theirts = atoi(parv[3]);

		if (ourts < theirts)
		{
			keepours = 1;
			makestruct = 0;
		}
		else if (ourts > theirts)
		{
			keeptheirs = makestruct = 1;
		}
		else
		{
			keepnone = 1;
			makestruct = 0;
		}

		if (acptr->user)
		{
			if (!match(acptr->user->host, parv[5]))
			{
				if (keepnone == 0)
				{
					same = 1;

					if (keepours == 1)
					{
						keepours = 0;
						keeptheirs = makestruct = 1;
					}
					else
					{
						keepours = 1;
						keeptheirs = makestruct = 0;
					}	
				}
			}
		}

		if (keeptheirs == 1)
		{
			ircstp->is_kill++;

			if (same == 1)
			{
				sendto_serv_butone(cptr, ":%s KILL %s :%s (Nick collision (old nick killed))",
					me.name, acptr->name, me.name);
				acptr->flags |= FLAGS_KILLED;
				exit_client(NULL, acptr, &me, "(Nick collision (old nick killed))");
			}
			else
			{
				sendto_serv_butone(cptr, ":%s KILL %s :%s (Nick collision (new nick killed))",
					me.name, acptr->name, me.name);
				acptr->flags |= FLAGS_KILLED;
				exit_client(NULL, acptr, &me, "(Nick collision (new nick killed))");
			}
		}
		else if (keepnone == 1)
		{
			ircstp->is_kill++;
			sendto_serv_butone(cptr, ":%s KILL %s :%s (Nick collision (all nicks killed))",
				me.name, acptr->name, me.name);
			acptr->flags |= FLAGS_KILLED;
			exit_client(NULL, acptr, &me, "(Nick collision (all nicks killed))");
		}
	}

	if (makestruct == 1)
	{
		sptr = make_client(cptr, serv);
		add_client_to_list(sptr);

		strncpyzt(sptr->name, parv[1], sizeof(sptr->name));

		sptr->hopcount = atoi(parv[2]);
		sptr->lastnick = atoi(parv[3]);

		(void)add_to_client_hash_table(parv[1], sptr);
		user = make_user(sptr);

		strncpyzt(user->server, parv[6], sizeof(user->server));
		strncpyzt(user->host, parv[5], sizeof(user->host));
		if (*parv[9] != '*')
		{
			if (user->maskedhost)
			{
				MyFree(user->maskedhost);
			}
			user->maskedhost = MyMalloc(strlen(parv[9]) + 1);
			(void)ircsprintf(user->maskedhost, "%s", parv[9]);
		}
		user->servicestamp = atol(parv[7]);
		strncpyzt(sptr->info, parv[10], sizeof(sptr->info));
		strncpyzt(user->username, parv[4], USERLEN + 1);

		SetClient(sptr);

		what = 0;
		for (p = parv[8]; *p; p++)
		{
			switch (*p)
			{
				case '+':
					what = MODE_ADD;
					break;
				case '-':
					what = MODE_DEL;
					break;
				case 'H':
					ParseMode(UMODE_SRA);
					break;
				case 'd':
					ParseMode(UMODE_DCCBLOCK);
					break;
				case 'S':
					ParseMode(UMODE_SERVICES);
					break;
				case 'V':
					ParseMode(UMODE_STRIPBADWORDS);
					break;
				case 'N':
					ParseMode(UMODE_NETADMIN);
					break;
				case 'C':
					ParseMode(UMODE_CONETADMIN);
					break;
				case 'T':
					ParseMode(UMODE_TECHADMIN);
					break;
				case 'Q':
					ParseMode(UMODE_ABUSEADMIN);
					break;
				case 'a':
					ParseMode(UMODE_SADMIN);
					break;
				case 'A':
					ParseMode(UMODE_ADMIN);
					break;
				case 'J':
					ParseMode(UMODE_SCOADMIN);
					break;
				case 'L':
					ParseMode(UMODE_GLOBCON);
					break;
				case 'o':
					ParseMode(UMODE_OPER);
					break;
				case 'i':
					ParseMode(UMODE_INVISIBLE);
					break;
				case 'g':
					ParseMode(UMODE_FAILOP);
					break;
				case 'k':
					ParseMode(UMODE_KILLS);
					break;
				case 'c':
					ParseMode(UMODE_CLIENT);
					break;
				case 'z':
					ParseMode(UMODE_MASK);
					break;
				case 'w':
					ParseMode(UMODE_WALLOPS);
					break;
				case 'b':
					ParseMode(UMODE_OPERCHAT);
					break;
				case 'h':
					ParseMode(UMODE_HELPOP);
					break;
				case 'B':
					ParseMode(UMODE_BOT);
					break;
				case 'W':
					ParseMode(UMODE_WHOIS);
					break;
				default:
					/* ignore unknown and local modes */
					break;
			}
		}

		IRCstats.clients++;
		if (IRCstats.clients > IRCstats.global_max)
		{
			IRCstats.global_max = IRCstats.clients;
		}
		if (IRCstats.me_clients > IRCstats.me_max)
		{
			IRCstats.me_max = IRCstats.me_clients;
		}

		if (IsInvisible(sptr))
		{
			IRCstats.invisible++;
		}
		if (IsOper(sptr))
		{
			IRCstats.operators++;
		}

		if ((aconf = find_conf_name(parv[1], CONF_QUARANTINED_NICK)) || (asqline = find_sqline_match(parv[1])))
		{
			sendto_ops("%s nick %s from %s@%s on %s", (aconf) ? "Q:Lined" : "SQ:Lined", parv[1],
				sptr->user->username, IsMasked(sptr) ? sptr->user->maskedhost : sptr->user->host,
				sptr->user->server);
		}

		if (find_conf_host(cptr->confs, sptr->name, CONF_UWORLD) || (sptr->user && find_conf_host(cptr->confs,
			sptr->user->server, CONF_UWORLD)))
		{
			sptr->flags |= FLAGS_ULINE;
		}

		hash_check_notify(sptr, RPL_LOGON);
	}

	if ((snptr = find_client(parv[1], NULL)))
	{
		sendto_serv_butone_snick(sptr, snptr->name, snptr->hopcount + 1, snptr->lastnick, snptr->user->username,
			snptr->user->host, snptr->user->server, snptr->user->servicestamp, get_mode_str(snptr),
			(IsMasked(snptr) ? snptr->user->maskedhost : "*"), snptr->info);
	}
	return 0;
}
